/*=========================================================================

Program:   ORFEO Toolbox
Language:  C++
Date:      $Date$
Version:   $Revision$


Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


This software is distributed WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbViewerEducationController_cxx
#define __otbViewerEducationController_cxx

#include "otbViewerEducationController.h"
#include "otbMsgReporter.h"
#include <FL/fl_ask.H>

namespace otb
{

ViewerEducationController::ViewerEducationController() : m_View(), m_WidgetsController(),
                                                         m_ResizingHandler(), m_ChangeScaleHandler(),
                                                         m_ChangeRegionHandler(), 
                                                         m_ChangeScaledRegionHandler(),
                                                         m_PixelDescriptionHandler(),
                                                         m_LastLayerAdded("")
{
  // Build the widgets controller
  m_WidgetsController           = WidgetsControllerType::New();
  m_PixelSpectrumController = PixelSpectrumControllerType::New();

  // Initialize the widgets controller and the action handlers
  this->InitializeWidgetsController();
}

ViewerEducationController::~ViewerEducationController()
{}

void ViewerEducationController::InitializeWidgetsController()
{
  // Clear all action handler added
  m_WidgetsController->ClearAllActionHandlers();

  // Build the action handlers
  m_ResizingHandler           = ResizingHandlerType::New();
  m_ChangeScaleHandler        = ChangeScaleHandlerType::New();
  m_ChangeRegionHandler       = ChangeRegionHandlerType::New();
  m_ChangeScaledRegionHandler = ChangeScaledRegionHandlerType::New();
  m_PixelDescriptionHandler   = PixelDescriptionActionHandlerType::New();
  m_PixelDescriptionHandler->SetUsePlaceName(false);

  // Add the action handlers to the widgets controller
  m_WidgetsController->AddActionHandler(m_ResizingHandler);
  m_WidgetsController->AddActionHandler(m_ChangeScaleHandler);
  m_WidgetsController->AddActionHandler(m_ChangeRegionHandler);
  m_WidgetsController->AddActionHandler(m_ChangeScaledRegionHandler);
  m_WidgetsController->AddActionHandler(m_PixelDescriptionHandler);
}

void ViewerEducationController::SetModel(ViewerEducationModel* model)
{
  m_Model = model;
  m_PixelSpectrumController->SetModel(m_Model->GetPixelSpectrumModel());
  
  // Register the model to the action handlers
  m_ResizingHandler->SetModel(m_Model->GetVisualizationModel());
  m_ChangeScaleHandler->SetModel(m_Model->GetVisualizationModel());
  m_ChangeRegionHandler->SetModel(m_Model->GetVisualizationModel());
  m_ChangeScaledRegionHandler->SetModel(m_Model->GetVisualizationModel());
  m_PixelDescriptionHandler->SetModel(m_Model->GetPixelDescriptionModel());
}

void ViewerEducationController::Save(const std::string & projectFile)
{
  try
    {
    // const std::string ext = itksys::SystemTools::GetFilenameExtension(projectFile);
    // //TODO : il faut plutot creer une pop up pour dire que l'extension n'est pas valide ...
    // if (ext != ".vep")
    // {
    //    itkGenericExceptionMacro("L'extension du fichier "<< projectFile <<" est incorrecte ! ")
    // }
    m_Model->Save(projectFile);
    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
}

void ViewerEducationController::ExportQGIS(const std::string & qgsFile, const std::string & qgsFile_utf8)
{
  try
    {
    // const std::string ext = itksys::SystemTools::GetFilenameExtension(qgsFile);
    // //TODO : il faut plutot creer une pop up pour dire que l'extension n'est pas valide ...
    // if (ext != ".qgs")
    // {
    //    itkGenericExceptionMacro("L'extension du fichier "<< qgsFile <<" est incorrecte ! ")
    // }
    m_Model->ExportQGIS(qgsFile, qgsFile_utf8);
    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
}

void ViewerEducationController::ExportKMZ(const std::string & kmzFile)
{
  try
    {
    // const std::string ext = itksys::SystemTools::GetFilenameExtension(kmzFile);
    // //TODO : il faut plutot creer une pop up pour dire que l'extension n'est pas valide ...
    // if (ext != ".kmz")
    // {
    //    itkGenericExceptionMacro("L'extension du fichier "<< kmzFile <<" est incorrecte ! ")
    // }
    m_Model->ExportKMZ(kmzFile);
    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
}

void ViewerEducationController::SetView(ViewerEducationView * view)
{
  m_View = view;
  m_PixelSpectrumController->SetView(m_View->GetPixelSpectrumView());
  m_ResizingHandler->SetView(m_View->GetImageView());
  m_ChangeScaleHandler->SetView(m_View->GetImageView());
  m_ChangeRegionHandler->SetView(m_View->GetImageView());
  m_ChangeScaledRegionHandler->SetView(m_View->GetImageView());
  m_PixelDescriptionHandler->SetView(m_View->GetImageView());
}

void ViewerEducationController::SetTrueColorComposition(const std::string & layerId)
{
   try
    {
    m_Model->SetTrueColorComposition(layerId);

    // Build a controller to the current visualization mode ImageView
    this->AddControllerToMap(layerId);

    // Link to the main view 
    this->LinkToMainDisplay(layerId);

    // Register this as the last layer added
    m_LastLayerAdded = layerId;
    
    // Link this layer to the other layers already added
    this->LinkSubViewers();
    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    } 
}

void ViewerEducationController::SetGrayColorComposition(const std::string & layerId,
                                                        const unsigned int &channel,
                                                        const std::string & desc)
{
   try
    {
    m_Model->SetGrayColorComposition(layerId, channel, desc);

    // Build a controller to the current visualization mode ImageView
    this->AddControllerToMap(layerId);

    // Link to the main view 
    this->LinkToMainDisplay(layerId);

    // Register this as the last layer added
    m_LastLayerAdded = layerId;
    
    // Link this layer to the other layers already added
    this->LinkSubViewers();
      
    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
}

void ViewerEducationController::KMeans(const std::string& layerId, unsigned int nbClasses)
{
  try
    {
    // Tell the model to build the current visualization mode
    // RenderingModel and to add it to RenderingMap
    m_Model->KMeans(layerId, nbClasses);
    
    // Build a controller to the current visualization mode ImageView
    this->AddControllerToMap(layerId);

    // Link to the main view 
    this->LinkToMainDisplay(layerId);
    
    // Register this as the last layer added
    m_LastLayerAdded = layerId;
    
    // Link this layer to the other layers already added
    this->LinkSubViewers();

    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
}

void ViewerEducationController::NDVI(const std::string & layerId)
{
  try
    {
    // Tell the model to build the current visualization mode
    // RenderingModel and to add it to RenderingMap
    m_Model->NDVI(layerId);
    
    // Build a controller to the current visualization mode ImageView
    this->AddControllerToMap(layerId);
    
    // Link to the main view 
    this->LinkToMainDisplay(layerId);

    // Register this as the last layer added
    m_LastLayerAdded = layerId;
    
    // Link this layer to the other layers already added
    this->LinkSubViewers();
    
    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
}

void ViewerEducationController::Brightness(const std::string & layerId)
{
   try
    {
    // Tell the model to build the current visualization mode
    // RenderingModel and to add it to RenderingMap
    m_Model->Brightness(layerId);
    
    // Build a controller to the current visualization mode ImageView
    this->AddControllerToMap(layerId);

    // Link to the main view 
    this->LinkToMainDisplay(layerId);

    // Register this as the last layer added
    m_LastLayerAdded = layerId;
    
    // Link this layer to the other layers already added
    this->LinkSubViewers();

    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
}



void ViewerEducationController::AddControllerToMap(const std::string& layerId)
{
  // Handler instanciation
  ResizingHandlerType::Pointer     resizingHandler        = ResizingHandlerType::New();
  ChangeScaleHandlerType::Pointer  changeScaleHandler     = ChangeScaleHandlerType::New();
  ChangeRegionHandlerType::Pointer changeRegionHandler       = ChangeRegionHandlerType::New();
  ChangeScaledRegionHandlerType::Pointer changeScaledRegionHandler = ChangeScaledRegionHandlerType::New();
  PixelDescriptionActionHandlerType::Pointer pixelDescriptionHandler   = PixelDescriptionActionHandlerType::New();
  pixelDescriptionHandler->SetUsePlaceName(false);
  
  // Add the handlers to the controller
  WidgetsControllerType::Pointer widgetController = WidgetsControllerType::New();
  widgetController->AddActionHandler(resizingHandler);
  widgetController->AddActionHandler(changeScaleHandler);
  widgetController->AddActionHandler(changeRegionHandler);
  widgetController->AddActionHandler(changeScaledRegionHandler);
  widgetController->AddActionHandler(pixelDescriptionHandler);
  
  // Set RenderingModel to the handler
  resizingHandler->SetModel(m_Model->GetRenderingModelMap()[layerId]);
  changeScaleHandler->SetModel(m_Model->GetRenderingModelMap()[layerId]);
  changeRegionHandler->SetModel(m_Model->GetRenderingModelMap()[layerId]);
  changeScaledRegionHandler->SetModel(m_Model->GetRenderingModelMap()[layerId]);
  pixelDescriptionHandler->SetModel(m_Model->GetPixelDescriptionModelMap()[layerId]);

  // Set the imageView to the handlers
  resizingHandler->SetView(m_View->GetImageViewMap()[layerId]);
  changeScaleHandler->SetView(m_View->GetImageViewMap()[layerId]);
  changeRegionHandler->SetView(m_View->GetImageViewMap()[layerId]);
  changeScaledRegionHandler->SetView(m_View->GetImageViewMap()[layerId]);
  pixelDescriptionHandler->SetView(m_View->GetImageViewMap()[layerId]);

  // Add the controller to the map
  m_ControllerMap[layerId] = widgetController;
}

// link displayed widget to the current one
void ViewerEducationController::LinkToMainDisplay(const std::string& layerId)
{
  //Get the controller of the widget to link
  WidgetsControllerType::Pointer controller = m_ControllerMap[layerId];

  //Get the rendering model of the widget to ling
  ViewerEducationModel::VisualizationModelType::Pointer mainRenderingModel = m_Model->GetVisualizationModel();
  ViewerEducationModel::VisualizationModelType::Pointer renderingModel     = m_Model->GetRenderingModelMap()[layerId];

  // Initialize with the same extracted region
  renderingModel->SetExtractRegion(mainRenderingModel->GetExtractRegion());
  renderingModel->SetScaledExtractRegion(mainRenderingModel->GetScaledExtractRegion());
  

  //Get the ImageView
  ViewerEducationView::ImageViewType::Pointer mainImageView = m_View->GetImageView();
  ViewerEducationView::ImageViewType::Pointer imageView     = m_View->GetImageViewMap()[layerId];

  // Get the pixel view
  ViewerEducationModel::PixelDescriptionModelType::Pointer mainPixelModel = m_Model->GetPixelDescriptionModel();
  ViewerEducationModel::PixelDescriptionModelType::Pointer pixelModel    = m_Model->GetPixelDescriptionModelMap()[layerId];

  // -------------  Add the resizing handler
  ResizingHandlerType::Pointer mainResizingHandler = ResizingHandlerType::New();
  mainResizingHandler->SetModel(mainRenderingModel);
  mainResizingHandler->SetView(imageView);

  ResizingHandlerType::Pointer resizingHandler = ResizingHandlerType::New();
  resizingHandler->SetModel(renderingModel);
  resizingHandler->SetView(mainImageView);

  m_WidgetsController->AddActionHandler( resizingHandler);
  controller->AddActionHandler(mainResizingHandler);

  // -------------   Add the changeScaledRegion handler
  ChangeScaledRegionHandlerType::Pointer mainChangeScaledRegionHandler = ChangeScaledRegionHandlerType::New();
  mainChangeScaledRegionHandler->SetModel(mainRenderingModel);
  mainChangeScaledRegionHandler->SetView(imageView);

  ChangeScaledRegionHandlerType::Pointer changeScaledRegionHandler = ChangeScaledRegionHandlerType::New();
  changeScaledRegionHandler->SetModel(renderingModel);
  changeScaledRegionHandler->SetView(mainImageView);

  m_WidgetsController->AddActionHandler(changeScaledRegionHandler);
  controller->AddActionHandler(mainChangeScaledRegionHandler);

  // -------------   Add the ChangeRegionHandler
  ChangeRegionHandlerType::Pointer mainChangeRegionHandler = ChangeRegionHandlerType::New();
  mainChangeRegionHandler->SetModel(mainRenderingModel);
  mainChangeRegionHandler->SetView(imageView);

  ChangeRegionHandlerType::Pointer changeRegionHandler = ChangeRegionHandlerType::New();
  changeRegionHandler->SetModel(renderingModel);
  changeRegionHandler->SetView(mainImageView);

  m_WidgetsController->AddActionHandler(changeRegionHandler );
  controller->AddActionHandler(mainChangeRegionHandler);

  // -------------   Add the changeScale handler
  imageView->GetZoomWidget()->SetIsotropicZoom(mainImageView->GetZoomWidget()->GetIsotropicZoom());

  ChangeScaleHandlerType::Pointer mainChangeScaleHandler = ChangeScaleHandlerType::New();
  mainChangeScaleHandler->SetModel(mainRenderingModel);
  mainChangeScaleHandler->SetView(mainImageView);

  ChangeScaleHandlerType::Pointer changeScaleHandler = ChangeScaleHandlerType::New();
  changeScaleHandler->SetModel(renderingModel);
  changeScaleHandler->SetView(imageView);

  m_WidgetsController->InsertActionHandler(0,  changeScaleHandler);
  controller->InsertActionHandler(0, mainChangeScaleHandler);

  // m_WidgetsController->AddActionHandler(changeScaleHandler );
  // controller->AddActionHandler(mainChangeScaleHandler);
  
  // -------------   Add the pixel description handler
  PixelDescriptionActionHandlerType::Pointer mainPixelHandler = PixelDescriptionActionHandlerType::New();
  mainPixelHandler->SetUsePlaceName(false);
  mainPixelHandler->SetView(imageView);
  mainPixelHandler->SetModel(mainPixelModel);

  PixelDescriptionActionHandlerType::Pointer pixelHandler = PixelDescriptionActionHandlerType::New();
  pixelHandler->SetUsePlaceName(false);
  pixelHandler->SetView(mainImageView);
  pixelHandler->SetModel(pixelModel);

  m_WidgetsController->AddActionHandler(pixelHandler );
  controller->AddActionHandler(mainPixelHandler);
}

// link displayed widget to the current one
void ViewerEducationController::LinkDispalys(const std::string& layerIdReference, 
                                             const std::string& layerIdSecondary)
{
  //Get the controller of the widget to link
  WidgetsControllerType::Pointer refController = m_ControllerMap[layerIdReference];
  WidgetsControllerType::Pointer controller = m_ControllerMap[layerIdSecondary];

  //Get the rendering model of the widget to link
  ViewerEducationModel::VisualizationModelType::Pointer mainRenderingModel = m_Model->GetRenderingModelMap()[layerIdReference];
  ViewerEducationModel::VisualizationModelType::Pointer renderingModel     = m_Model->GetRenderingModelMap()[layerIdSecondary];

  //Get the ImageView
  ViewerEducationView::ImageViewType::Pointer mainImageView = m_View->GetImageViewMap()[layerIdReference];
  ViewerEducationView::ImageViewType::Pointer imageView     = m_View->GetImageViewMap()[layerIdSecondary];

  // Get the pixel view
  ViewerEducationModel::PixelDescriptionModelType::Pointer mainPixelModel = m_Model->GetPixelDescriptionModelMap()[layerIdReference];
  ViewerEducationModel::PixelDescriptionModelType::Pointer pixelModel    = m_Model->GetPixelDescriptionModelMap()[layerIdSecondary];

  // -------------  Add the resizing handler
  ResizingHandlerType::Pointer mainResizingHandler = ResizingHandlerType::New();
  mainResizingHandler->SetModel(mainRenderingModel);
  mainResizingHandler->SetView(imageView);

  ResizingHandlerType::Pointer resizingHandler = ResizingHandlerType::New();
  resizingHandler->SetModel(renderingModel);
  resizingHandler->SetView(mainImageView);

  refController->AddActionHandler( resizingHandler);
  controller->AddActionHandler(mainResizingHandler);

  // -------------   Add the changeScaledRegion handler
  ChangeScaledRegionHandlerType::Pointer mainChangeScaledRegionHandler = ChangeScaledRegionHandlerType::New();
  mainChangeScaledRegionHandler->SetModel(mainRenderingModel);
  mainChangeScaledRegionHandler->SetView(imageView);

  ChangeScaledRegionHandlerType::Pointer changeScaledRegionHandler = ChangeScaledRegionHandlerType::New();
  changeScaledRegionHandler->SetModel(renderingModel);
  changeScaledRegionHandler->SetView(mainImageView);

  refController->AddActionHandler(changeScaledRegionHandler);
  controller->AddActionHandler(mainChangeScaledRegionHandler);

  // -------------   Add the ChangeRegionHandler
  ChangeRegionHandlerType::Pointer mainChangeRegionHandler = ChangeRegionHandlerType::New();
  mainChangeRegionHandler->SetModel(mainRenderingModel);
  mainChangeRegionHandler->SetView(imageView);

  ChangeRegionHandlerType::Pointer changeRegionHandler = ChangeRegionHandlerType::New();
  changeRegionHandler->SetModel(renderingModel);
  changeRegionHandler->SetView(mainImageView);

  refController->AddActionHandler(changeRegionHandler );
  controller->AddActionHandler(mainChangeRegionHandler);

  // -------------   Add the changeScale handler
  imageView->GetZoomWidget()->SetIsotropicZoom(mainImageView->GetZoomWidget()->GetIsotropicZoom());
  
  ChangeScaleHandlerType::Pointer mainChangeScaleHandler = ChangeScaleHandlerType::New();
  mainChangeScaleHandler->SetModel(mainRenderingModel);
  mainChangeScaleHandler->SetView(mainImageView);

  ChangeScaleHandlerType::Pointer changeScaleHandler = ChangeScaleHandlerType::New();
  changeScaleHandler->SetModel(renderingModel);
  changeScaleHandler->SetView(imageView);
  
  refController->InsertActionHandler(0, changeScaleHandler );
  controller->InsertActionHandler(0, mainChangeScaleHandler);

  // refController->AddActionHandler(changeScaleHandler );
  // controller->AddActionHandler(mainChangeScaleHandler);
  
  // -------------   Add the pixel description handler
  PixelDescriptionActionHandlerType::Pointer mainPixelHandler = PixelDescriptionActionHandlerType::New();
  mainPixelHandler->SetUsePlaceName(false);
  mainPixelHandler->SetView(imageView);
  mainPixelHandler->SetModel(mainPixelModel);

  PixelDescriptionActionHandlerType::Pointer pixelHandler = PixelDescriptionActionHandlerType::New();
  pixelHandler->SetUsePlaceName(false);
  pixelHandler->SetView(mainImageView);
  pixelHandler->SetModel(pixelModel);

  refController->AddActionHandler(pixelHandler );
  controller->AddActionHandler(mainPixelHandler);
}

void ViewerEducationController::LinkSubViewers()
{
  // Link the last layer registerd to the other layers
  ControllerMapType::const_iterator  controllerIt = m_ControllerMap.begin();
  while(controllerIt != m_ControllerMap.end())
    {
    if ((*controllerIt).first != m_LastLayerAdded )
      {
      // Link the viewers
      this->LinkDispalys(m_LastLayerAdded, (*controllerIt).first);
      }
    ++controllerIt;
    }
}

void ViewerEducationController::InitializeController()
{
  try
    {
    // Initialize the widgetsController
    this->InitializeWidgetsController();

    // clear maps
    m_ControllerMap.clear();    
    
    m_ResizingHandler->SetView(m_View->GetImageView());
    m_ChangeScaleHandler->SetView(m_View->GetImageView());
    m_ChangeRegionHandler->SetView(m_View->GetImageView());
    m_ChangeScaledRegionHandler->SetView(m_View->GetImageView());
    m_PixelDescriptionHandler->SetView(m_View->GetImageView());
    m_PixelSpectrumController->SetView(m_View->GetPixelSpectrumView());
    
    // Tell the model to clear it maps
    m_Model->ClearMaps();

    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
}


unsigned int 
ViewerEducationController
::GetInputImageMode(const std::string& infname)
{
  unsigned int mode;
  try
    {
    mode  = m_Model->GetInputImageMode(infname);
    }
  catch(itk::ExceptionObject &)
    {
    std::ostringstream oss;
    oss << "Erreur lors de l'ouverture du fichier : " << infname <<  std::endl;
    MsgReporter::GetInstance()->SendError(oss.str().c_str());
    }

  return mode;
}

void
ViewerEducationController
::WriteSpreadsheetFile(const std::string & file)
{
  try
    {
    // const std::string ext = itksys::SystemTools::GetFilenameExtension(file);
    // //TODO : il faut plutot creer une pop up pour dire que l'extension n'est pas valide ...
    // if (ext != ".csv")
    // {
    //    itkGenericExceptionMacro("L'extension du fichier "<< file <<" est incorrecte ! ")
    // }
    m_Model->WriteSpreadsheetFile(file);
    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }  
}

} // end namespace otb

#endif
