/*=========================================================================

Program:   ORFEO Toolbox
Language:  C++
Date:      $Date$
Version:   $Revision$


Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


This software is distributed WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/

#include "otbViewerEducationModel.h"
#include "otbMsgReporter.h"

#include "otbImageMetadataInterfaceBase.h"
#include "otbImageMetadataInterfaceFactory.h"

#include "otbFltkFilterWatcher.h"

namespace otb
{

ViewerEducationModel::ViewerEducationModel() : m_VisualizationModel(),
                                               m_PixelDescriptionModel(),
                                               m_VectorImage(),m_RedChannel(-1),
                                               m_GreenChannel(-1),m_BlueChannel(-1),
                                               m_NIRChannel(-1),m_MIRChannel(-1),
                                               m_UserRedChannel(-1), m_UserGreenChannel(-1),
                                               m_UserBlueChannel(-1), m_UserNIRChannel(-1),
                                               m_UserMIRChannel(-1),
                                               m_ProjectFile(""),
                                               m_ImageOpened(false), m_DetectedMode(Mode_Unknown),
                                               m_GrayColorComposition(false)

{
  m_VisualizationModel    = VisualizationModelType::New();
  m_PixelDescriptionModel = PixelDescriptionModelType::New();
  m_Reader = ReaderType::New();
  m_NDVIFilter = NDVIFilterType::New();
  m_NDVICastFilter = ImageToVectorImageFilterType::New();
  m_BrightnessFilter = IBFilterType::New();
  m_BrightnessCastFilter = ImageToVectorImageFilterType::New();
  
  m_PixelSpectrumModel = PixelSpectrumModelType::New();
  m_PixelSpectrumModel->SetLayers(m_VisualizationModel->GetLayers());

  // initialize the association between spectral bands and their image
  // composant : 4099 is invalid spectral band
  m_AssociatedSpectralBands["Bande spectrale Rouge"] = 4099;
  m_AssociatedSpectralBands["Bande spectrale PIR"] = 4099;
  m_AssociatedSpectralBands["Bande spectrale MIR"] = 4099;
  m_AssociatedSpectralBands["Bande spectrale Bleue"] = 4099;
  m_AssociatedSpectralBands["Bande spectrale Verte"] = 4099;

  // input image label initialization
  m_InputImageLabel = "Image";
  
}

void ViewerEducationModel::OpenImage(const std::string& inputFilename)
{
  m_InputFilename = inputFilename;

  // Get the label to use for the input image
  std::string inputImageLabel = itksys::SystemTools::GetFilenameWithoutExtension(inputFilename);
  if (!inputImageLabel.empty())
    {
    m_InputImageLabel = inputImageLabel;
    }

  m_Reader->SetFileName(inputFilename);
  m_Reader->UpdateOutputInformation();
  m_VectorImage  = m_Reader->GetOutput();

  // Set the number of components per pixel
  m_NumberOfComponentsPerPixel = m_VectorImage->GetNumberOfComponentsPerPixel();
  
  // Generate the layer
  m_ImageGenerator = LayerGeneratorType::New();
  m_ImageGenerator->SetImage(const_cast<VectorImageType *>(m_VectorImage.GetPointer()));
  m_ImageGenerator->GetLayer()->SetName(m_InputImageLabel);
  FltkFilterWatcher qlwatcher(m_ImageGenerator->GetProgressSource(), 
                              0, 0, 300, 20, otbGetTextMacro("chargement en cours ..."));
  m_ImageGenerator->GenerateLayer();
  
  
  //Compute min/max values necessary to the rescale spectrum.
  bool hasQuickLook = m_ImageGenerator->GetLayer()->GetHasQuicklook();
  VectorImageType::ConstPointer tempImage;
  // checking for any quicklook
  if (hasQuickLook)
    {
    tempImage = m_ImageGenerator->GetLayer()->GetQuicklook();
    }
  else
    {
    tempImage = m_VectorImage;
    }
  StatFilterType::Pointer statFilter = StatFilterType::New();
  statFilter->SetInput(const_cast<VectorImageType *>(tempImage.GetPointer()));
  statFilter->Update();

  m_ImageGenerator->GetLayer()->SetMinValues(statFilter->GetMinimum());
  m_ImageGenerator->GetLayer()->SetMaxValues(statFilter->GetMaximum());

  // add the spectral order
  m_ImageGenerator->GetLayer()->UpdateChannelsOrder(m_BlueChannel, m_GreenChannel, 
                                                  m_RedChannel, m_NIRChannel, 
                                                  m_MIRChannel);
    
  // Get the default display order and the sensor id if any
  if (m_NumberOfComponentsPerPixel < 3)
    {
    m_Channels.resize(1);
    m_Channels[0]  = m_RedChannel;
    }
  else
    {
    // Reorder the channels to open the image in FalseColor 
    // composition mode
    m_Channels.resize(3);
    m_Channels[0] = m_NIRChannel;
    m_Channels[1] = m_RedChannel;
    m_Channels[2] = m_GreenChannel;
    }

  // Setup the default display channel order
  StandardRenderingFunctionType::Pointer renderingFunction = StandardRenderingFunctionType::New();
  m_ImageGenerator->GetLayer()->SetRenderingFunction(renderingFunction);
  renderingFunction->SetAutoMinMax(true);
  renderingFunction->SetChannelList(m_Channels);
  
  // Clear previousmy added layers
  m_VisualizationModel->ClearLayers();
  m_PixelDescriptionModel->ClearLayers();
  
  // Add the layer to the models
  m_VisualizationModel->AddLayer(m_ImageGenerator->GetLayer());
  m_PixelDescriptionModel->AddLayer(m_ImageGenerator->GetLayer());

  m_VisualizationModel->Update();

  //Call AddRenderingModelToMap to set the LayerSettings for the main
  //image using this generic method.
  this->AddRenderingModelToMap(m_VectorImage, m_InputImageLabel,"");
  
  //Add img filename to ApplicationSettings
  m_ApplicationSettings.FileName = inputFilename;
  m_ApplicationSettings.NbComponents = m_NumberOfComponentsPerPixel;
  m_ApplicationSettings.LayersMap[m_InputImageLabel].Type = LayerSettings::ColorComposite;
  m_ApplicationSettings.LayersMap[m_InputImageLabel].Mode = LayerSettings::FalseColor;
  
  // Notify the listeners
  m_ImageOpened = true;
  this->NotifyAll();
  m_ImageOpened = false;
}

unsigned int 
ViewerEducationModel::GetInputImageMode(const std::string& infname)
{
  ReaderType::Pointer reader = ReaderType::New();
  reader->SetFileName(infname);
  reader->UpdateOutputInformation();

  // Set the number of components per pixel
  m_NumberOfComponentsPerPixel = reader->GetOutput()->GetNumberOfComponentsPerPixel();

  // image with 2 or 5 components are not handled : throw an exception
  if (m_NumberOfComponentsPerPixel == 2 || m_NumberOfComponentsPerPixel == 5)
    {
    std::ostringstream oss;
    oss << "Les images comportant "<< m_NumberOfComponentsPerPixel<<" bandes ne sont pas gérées" << std::endl;
    itkExceptionMacro(<<oss.str().c_str());
    }

  // Getting informations from the metaDictionnary to setup the
  // default channel order
  typedef otb::ImageMetadataInterfaceBase ImageMetadataInterfaceType;
  ImageMetadataInterfaceType::Pointer metadataInterface = 
    ImageMetadataInterfaceFactory::CreateIMI(reader->GetOutput()->GetMetaDataDictionary());
  m_SensorID = metadataInterface->GetSensorID();
  ChannelListType channels = metadataInterface->GetDefaultDisplay();

  // CASE1 : 3 bands : see if spot5, if not by default put in
  //Unkown_Mode
  
  if (m_NumberOfComponentsPerPixel == 3)
    {
    // set spot5 mode or return as unkown
    if (m_SensorID == "Spot 5")
      {
      this->SetSpot5Mode();
      return Mode_Spot5;
      }
    else
      {
      // if not spot5 mode, set a default channels order
      m_GreenChannel = channels[0];
      m_RedChannel   = channels[1];
      m_NIRChannel   = channels[2];
      m_BlueChannel  = -1;
      m_MIRChannel  = -1;
      this->SetUnknownMode();
      return Mode_Unknown;
      }
    }

  // CASE2 : 4 bands

  // Detect the sensor model 
  if (m_NumberOfComponentsPerPixel == 4)
    {
    if (m_SensorID == "PHR 1A")
      {
      this->SetPleiadesMode();
      return  Mode_Pleiades;
      }
    else if (m_SensorID == "Spot 5")
      {
      this->SetSpot5Mode();
      return Mode_Spot5;
      }
    else if (m_SensorID == "Formosat 2")
      {
      this->SetFormosatMode();
      return Mode_Formosat;
      }
    }

  // Initialize the four channels with the default display set in the
  // metatdataImageInterface
  if (m_NumberOfComponentsPerPixel == 1 )
    {
    m_RedChannel   = 0;//channels[0];
    m_GreenChannel = 0;//channels[0];
    m_BlueChannel  = 0;//channels[0];
    m_NIRChannel   = 0;//channels[0];
    }

  if (m_NumberOfComponentsPerPixel == 3 )
    {
    m_RedChannel   = channels[0];
    m_GreenChannel = channels[1];
    m_BlueChannel  = -1; 
    m_NIRChannel   = channels[2];
    m_MIRChannel   = -1;
    }

  if (m_NumberOfComponentsPerPixel == 4 )
    {
    m_RedChannel   = channels[0];
    m_GreenChannel = channels[1];
    m_BlueChannel  = channels[2];
    m_NIRChannel   = 3;
    m_MIRChannel   = channels[2];
    }

  this->SetUnknownMode();  
  return Mode_Unknown;
}

void
ViewerEducationModel::Save(const std::string & projectFile)
{
  // If no extension, add one
  std::string ext = itksys::SystemTools::GetFilenameExtension(projectFile);
  std::ostringstream fullname;
  if ( ext != ".vep")
    {
    fullname<< projectFile << ".vep";
    }
  else
    {
    fullname << projectFile;
    }
  
  // get the filename without extension
  std::string projectName = itksys::SystemTools::GetFilenameWithoutExtension(projectFile);
  m_ApplicationSettings.ProjectName = projectName;
  save(fullname.str(), m_ApplicationSettings);
}

void
ViewerEducationModel::ExportQGIS(const std::string & qgsFile, const std::string & qgsFile_utf8)
{
  // check if the right extension is used
  std::string ext = itksys::SystemTools::GetFilenameExtension(qgsFile);

  std::ostringstream fullname;
  std::ostringstream fullname_utf8;
  if ( ext != ".qgs")
    {
    fullname<< qgsFile << ".qgs";
    fullname_utf8 << qgsFile_utf8 << ".qgs";
    }
  else
    {
    fullname << qgsFile;
	fullname_utf8 << qgsFile_utf8;
    }

  // export the project
  exportQGIS(fullname.str(), fullname_utf8.str(), m_ApplicationSettings);
}

void
ViewerEducationModel::ExportKMZ(const std::string & kmzFile)
{
  // check if the right extension is used
  std::string ext = itksys::SystemTools::GetFilenameExtension(kmzFile);
  std::ostringstream fullname;
  if ( ext != ".kmz")
    {
    fullname<< kmzFile << ".kmz";
    }
    else
    {
    fullname << kmzFile;
    }

   exportKMZ(fullname.str(), m_ApplicationSettings);
}

void
ViewerEducationModel::SetPleiadesMode()
{
  // Set the detected mode to be pleiades
  m_DetectedMode = Mode_Pleiades;

  // In Pleiades images channels are : 
  m_RedChannel   = 0;
  m_GreenChannel = 1;
  m_BlueChannel  = 2;
  m_NIRChannel   = 3;
  m_MIRChannel   = -1;


  // Set the blue and mir status
  bool blueIsActive = true;
  bool MIRIsActive  = false;
  if (m_MIRChannel != -1)  MIRIsActive = true;
  m_ApplicationSettings.BlueAndMIRStatus = BlueAndMIRSpectralBandStatus(blueIsActive, MIRIsActive);

  m_ApplicationSettings.DefaultChannels = DefaultChannelsSettings(0,1,2,3,-1);
  m_ApplicationSettings.Sensor = ApplicationSettings::PLEIADE;
}

void
ViewerEducationModel::SetFormosatMode()
{
  // Set the detected mode to be formosat
  m_DetectedMode = Mode_Formosat;

  // In Formosat images channels are : 
  m_RedChannel   = 0;
  m_GreenChannel = 1;
  m_BlueChannel  = 2;
  m_NIRChannel   = 3;
  m_MIRChannel   = -1;


  // Set the blue and mir status
  bool blueIsActive = true;
  bool MIRIsActive  = false;
  if (m_MIRChannel != -1)  MIRIsActive = true;
  m_ApplicationSettings.BlueAndMIRStatus = BlueAndMIRSpectralBandStatus(blueIsActive, MIRIsActive);
  

  m_ApplicationSettings.DefaultChannels = DefaultChannelsSettings(0,1,2,3,-1);
  m_ApplicationSettings.Sensor = ApplicationSettings::FORMOSAT;
}

void
ViewerEducationModel::SetSpot5Mode()
{
  // Set the detected mode to be Spot
  m_DetectedMode = Mode_Spot5;

  // In Spot5 images channels are : 
  // Channel means spectral band
  m_RedChannel   = 1;
  m_GreenChannel = 2;
  m_NIRChannel   = 0;

  if (m_NumberOfComponentsPerPixel == 4 )
    {
    m_MIRChannel = 3;
    }
  else
    {
    m_MIRChannel = -1;
    }

  // deactivate the blue channel
  m_BlueChannel  = -1;

  // Set the default channels
  m_ApplicationSettings.DefaultChannels = DefaultChannelsSettings(1,2,-1,0,3);

  // Set the blue and mir status
  bool blueIsActive = false;
  bool MIRIsActive  = false;
  if (m_MIRChannel != -1)  MIRIsActive = true;
  m_ApplicationSettings.BlueAndMIRStatus = BlueAndMIRSpectralBandStatus(blueIsActive, MIRIsActive);

  // Set the sensor type
  m_ApplicationSettings.Sensor = ApplicationSettings::SPOT;
}

void
ViewerEducationModel::UpdateSpectralBandsAssociation()
{
  if (m_GreenChannel!= -1) m_AssociatedSpectralBands["Bande spectrale Verte"] = m_GreenChannel;
  if (m_RedChannel  != -1) m_AssociatedSpectralBands["Bande spectrale Rouge"] = m_RedChannel;
  if (m_NIRChannel  != -1) m_AssociatedSpectralBands["Bande spectrale PIR"] = m_NIRChannel;

  // special treatement for blue and MIR
  if (m_ApplicationSettings.DefaultChannels.BlueBand  != -1) 
    m_AssociatedSpectralBands["Bande spectrale Bleue"] = m_BlueChannel;
  else
    m_AssociatedSpectralBands["Bande spectrale Bleue"] = 4099;

  if (m_ApplicationSettings.DefaultChannels.MIRBand  != -1) 
    m_AssociatedSpectralBands["Bande spectrale MIR"] = m_MIRChannel;
  else
    m_AssociatedSpectralBands["Bande spectrale MIR"] = 4099;
  
}

void
ViewerEducationModel::SetUnknownMode(bool rememberUserChoice )
{
  // Set the detected mode to be unknown
  m_DetectedMode = Mode_Unknown;
  
  // Default channels have been computed the first time the
  // application was launched, use them to fill the choices already
  // set by the user

  // if rememberUserChoice true, it means that the user does not
  // change the input image, he only changes the mode and came back to
  // unknown mode
  if (rememberUserChoice)
    {
    if (m_UserRedChannel == -1 && m_UserGreenChannel == -1 
        && m_UserBlueChannel == -1 && m_UserNIRChannel== -1)
      {
      m_UserRedChannel = m_RedChannel;
      m_UserGreenChannel = m_GreenChannel;
      m_UserBlueChannel = m_BlueChannel; 
      m_UserNIRChannel = m_NIRChannel;
      m_UserMIRChannel = m_MIRChannel;
      }
    else
      {
      m_RedChannel = m_UserRedChannel;
      m_GreenChannel = m_UserGreenChannel;
      m_BlueChannel = m_UserBlueChannel; 
      m_NIRChannel = m_UserNIRChannel;
      m_MIRChannel = m_UserMIRChannel;
      }
    }

  // In  other images, the order is set by the user
  m_ApplicationSettings.DefaultChannels = DefaultChannelsSettings(m_RedChannel, 
                                                                  m_GreenChannel, 
                                                                  m_BlueChannel, 
                                                                  m_NIRChannel,
                                                                  m_MIRChannel);

  // Set the blue and mir status
  bool blueIsActive = false;
  bool MIRIsActive  = false;
  if (m_BlueChannel != -1) blueIsActive = true;
  if (m_MIRChannel != -1)  MIRIsActive  = true;
  m_ApplicationSettings.BlueAndMIRStatus = BlueAndMIRSpectralBandStatus(blueIsActive, MIRIsActive);

  // Set the sensor type
  m_ApplicationSettings.Sensor = ApplicationSettings::UNKNOWN;
}


void 
ViewerEducationModel::UpdateChannelsChoice( int red,
                                            int green, 
                                            int blue, 
                                            int nir,
                                            int mir)
{
  // Unknown mode : user choice
  m_RedChannel   = red;
  m_GreenChannel = green;
  m_BlueChannel  = blue; 
  m_NIRChannel   = nir;
  m_MIRChannel   = mir;

  m_UserRedChannel   = red;
  m_UserGreenChannel = green;
  m_UserBlueChannel  = blue;
  m_UserNIRChannel   = nir;
  m_UserMIRChannel   = mir;

  m_ApplicationSettings.DefaultChannels = DefaultChannelsSettings(red,green,blue,nir, mir);
  //m_ApplicationSettings.Sensor = ApplicationSettings::UNKNOWN;
}


void
ViewerEducationModel::SetTrueColorComposition(const std::string& layerId)
{  
  
  m_Channels.resize(3);
  m_Channels[0] = m_RedChannel;
  m_Channels[1] = m_GreenChannel;
  m_Channels[2] = m_BlueChannel;

  this->AddRenderingModelToMap(m_VectorImage, layerId, "");
  m_ApplicationSettings.LayersMap[layerId].Mode = LayerSettings::TrueColors;
}


void
ViewerEducationModel::SetGrayColorComposition(const std::string& layerId,
                                              const unsigned int& channel, 
                                              const std::string & desc)
{
  m_Channels.resize(3);
  m_Channels[0] = channel;
  m_Channels[1] = channel;
  m_Channels[2] = channel;

  m_GrayColorComposition = true;
  this->AddRenderingModelToMap(m_VectorImage, layerId,"Niveau de gris", true);
  m_GrayColorComposition = false;
  m_ApplicationSettings.LayersMap[layerId].Mode = LayerSettings::GrayLevel;
  m_ApplicationSettings.LayersMap[layerId].GrayBand = channel;
  m_ApplicationSettings.LayersMap[layerId].Desc = desc;
}

void
ViewerEducationModel::KMeans(const std::string& layerId, unsigned int nbClasses)
{  
  // Training sample lists
  ListSampleType::Pointer sampleList = ListSampleType::New();
    
  // Sample dimension and max dimension
  const unsigned int nbComp =  m_NumberOfComponentsPerPixel;
  unsigned int sampleSize = nbComp;
  unsigned int totalSamples = 0;

  EstimatorType::ParametersType initialMeans(nbComp * nbClasses);
  initialMeans.Fill(0);

  // use image and mask shrink
  double theoricNBSamplesForKMeans = 10000;
  const double upperThresholdNBSamplesForKMeans = 1000 * 1000;
  const double actualNBSamplesForKMeans = std::min(theoricNBSamplesForKMeans, upperThresholdNBSamplesForKMeans);

  const double shrinkFactor = vcl_floor(
    vcl_sqrt(
      m_VectorImage->GetLargestPossibleRegion().GetNumberOfPixels()
      / actualNBSamplesForKMeans));

  ImageSamplingFilterType::Pointer imageSampler = ImageSamplingFilterType::New();
  imageSampler->SetInput(const_cast<VectorImageType*>(m_VectorImage.GetPointer()));
  imageSampler->SetShrinkFactor(shrinkFactor);
  imageSampler->Update();

  // Then, build the sample list
  IteratorType it(imageSampler->GetOutput(), 
                  imageSampler->GetOutput()->GetLargestPossibleRegion());

  it.GoToBegin();

  VectorImageType::PixelType min;
  VectorImageType::PixelType max;
  VectorImageType::PixelType sample;
  //first sample

  itk::Statistics::MersenneTwisterRandomVariateGenerator::Pointer 
    randGen =itk::Statistics::MersenneTwisterRandomVariateGenerator::GetInstance();

  //randGen->Initialize();

  min = it.Get();
  max = it.Get();
  sample = it.Get();
  sampleList->PushBack(sample);
  ++it;

  totalSamples = 1;
  while (!it.IsAtEnd())
    {
    totalSamples++;
    sample = it.Get();
    sampleList->PushBack(sample);
      
    for (unsigned int i = 0; i < nbComp; ++i)
      {
      if (min[i] > sample[i])
        {
        min[i] = sample[i];
        }
      if (max[i] < sample[i])
        {
        max[i] = sample[i];
        }
      }
    ++it;
    }

  // Next, initialize centroids by random sampling in the generated
  // list of samples
  for (unsigned int classIndex = 0; classIndex < nbClasses; ++classIndex)
    {
    VectorImageType::PixelType newCentroid = sampleList->GetMeasurementVector(randGen->GetIntegerVariate(sampleList->Size()-1));

    for (unsigned int compIndex = 0; compIndex < sampleSize; ++compIndex)
      {
      initialMeans[compIndex + classIndex * sampleSize] = newCentroid[compIndex];
      }
    }

  /*******************************************/
  /*           Learning                      */
  /*******************************************/
  EstimatorType::Pointer estimator = EstimatorType::New();

  TreeGeneratorType::Pointer treeGenerator = TreeGeneratorType::New();
  treeGenerator->SetSample(sampleList);
  treeGenerator->SetBucketSize(10000);
  treeGenerator->Update();

  estimator->SetParameters(initialMeans);
  estimator->SetKdTree(treeGenerator->GetOutput());
  estimator->SetMaximumIteration(10000);
  //estimator->SetCentroidPositionChangesThreshold(GetParameterFloat("ct"));
  estimator->StartOptimization();

  EstimatorType::ParametersType estimatedMeans = estimator->GetParameters();

  /*******************************************/
  /*           Classification                */
  /*******************************************/

  // Finally, update the KMeans filter
  KMeansFunctorType functor;

  for (unsigned int classIndex = 0; classIndex < nbClasses; ++classIndex)
    {
    VectorImageType::PixelType centroid(sampleSize);

    for (unsigned int compIndex = 0; compIndex < sampleSize; ++compIndex)
      {
      centroid[compIndex] = estimatedMeans[compIndex + classIndex * sampleSize];
      }
    functor.AddCentroid(classIndex, centroid);
    }

  KMeansFilterType::Pointer kMeansFilter = KMeansFilterType::New();
  kMeansFilter->SetFunctor(functor);
  kMeansFilter->SetInput(m_VectorImage);
  m_KMeansFilterMap[layerId] = kMeansFilter;
 // kMeansFilter->Update();
  
  // Colorize this label image
  this->DoColorMapping(kMeansFilter->GetOutput(),layerId);
  
  //Update Settings parameter
  m_ApplicationSettings.LayersMap[layerId].Mode = LayerSettings::ColorRamp;
  m_ApplicationSettings.LayersMap[layerId].Type = LayerSettings::KMeans;
  m_ApplicationSettings.LayersMap[layerId].NbClasses = nbClasses;
}


void
ViewerEducationModel::NDVI(const std::string& layerId)
{  
  // Channels must begin from index 1
  m_NDVIFilter->SetInput(m_VectorImage);
  m_NDVIFilter->SetRedIndex(m_ApplicationSettings.DefaultChannels.RedBand + 1 );
  m_NDVIFilter->SetNIRIndex(m_ApplicationSettings.DefaultChannels.NIRBand + 1 );

  m_NDVICastFilter->SetInput(m_NDVIFilter->GetOutput());
  

  m_Channels.resize(1);
  m_Channels[0] = 0;
  
  this->AddRenderingModelToMap(m_NDVICastFilter->GetOutput(), layerId, "NDVI", true);
  m_ApplicationSettings.LayersMap[layerId].Mode = LayerSettings::GrayLevel;
  m_ApplicationSettings.LayersMap[layerId].GrayBand = 0;
  m_ApplicationSettings.LayersMap[layerId].Type = LayerSettings::NDVI;
}

void
ViewerEducationModel::Brightness(const std::string& layerId)
{  
  // Channels must begin from index 1
  m_BrightnessFilter->SetInput(m_VectorImage);
  m_BrightnessFilter->SetRedIndex(m_ApplicationSettings.DefaultChannels.RedBand + 1 );
  m_BrightnessFilter->SetGreenIndex(m_ApplicationSettings.DefaultChannels.GreenBand + 1);
  //brightness->GetFunctor().SetNIRIndex(m_ApplicationSettings.DefaultChannels.NIRBand);
  //brightness->Update();
  
  m_BrightnessCastFilter->SetInput(m_BrightnessFilter->GetOutput());

  m_Channels.resize(1);
  m_Channels[0] = 0;
  
  this->AddRenderingModelToMap(m_BrightnessCastFilter->GetOutput(), layerId,"Brillance", true);
  m_ApplicationSettings.LayersMap[layerId].Mode = LayerSettings::GrayLevel;
  m_ApplicationSettings.LayersMap[layerId].GrayBand = 0;
  m_ApplicationSettings.LayersMap[layerId].Type = LayerSettings::Brightness;
}

void
ViewerEducationModel::AddRenderingModelToMap(const VectorImageType* image,
                                             const std::string& layerId,
                                             const std::string &pixelValueDescription,
                                             bool hideRGBPixelValue)
{  
  // Generate the layer
  LayerGeneratorType::Pointer imageGenerator = LayerGeneratorType::New();
  imageGenerator->SetImage(const_cast<VectorImageType*>(image));
  imageGenerator->GetLayer()->SetName(layerId);
  imageGenerator->GenerateLayer();
  
  // Simple ImageLayer configuration : labels
  // /imageGenerator->GetLayer()->SetHideRGBPixelValue(hideRGBPixelValue)
  // : always hide RGB images
  if(!pixelValueDescription.empty()) 
    {
    imageGenerator->GetLayer()->SetPixelValueDescriptionLabel(pixelValueDescription);
    }

  // Special case : grayComposition
  if (m_GrayColorComposition)
    {
    imageGenerator->GetLayer()->SetIsGrayComposition(true);
    imageGenerator->GetLayer()->SetGrayChannel(m_Channels[0]);
    }

  // Colormap in case of Kmeans processing
  if ( m_ColorMapRampsContainer.find(layerId) != m_ColorMapRampsContainer.end())
    {
    // it is a kmeans layer : add the ColorMap associated to this
    // image to the ImageLayer
    imageGenerator->GetLayer()->SetColorMapRamp(m_ColorMapRampsContainer[layerId]);
    }

  // add the spectral order
  imageGenerator->GetLayer()->UpdateChannelsOrder(m_BlueChannel, m_GreenChannel, 
                                                m_RedChannel, m_NIRChannel, 
                                                m_MIRChannel);
  

  StandardRenderingFunctionType::Pointer renderingFunction = StandardRenderingFunctionType::New();
  imageGenerator->GetLayer()->SetRenderingFunction(renderingFunction);
  renderingFunction->SetAutoMinMax(true);
  renderingFunction->SetChannelList(m_Channels);
  
  VisualizationModelType::Pointer renderingModel = VisualizationModelType::New();
  renderingModel->AddLayer(imageGenerator->GetLayer());
  renderingModel->Update();

  PixelDescriptionModelType::Pointer pixelDescriptionModel = PixelDescriptionModelType::New();
  pixelDescriptionModel->AddLayer(imageGenerator->GetLayer());

  // Add the rendering model to the map
  m_RenderingModelMap[layerId] = renderingModel;

  // Add the pixel description model to the map
  m_PixelDescriptionModelMap[layerId] = pixelDescriptionModel;
  
  //Add a new layerSettings to the m_ApplicationSettings
  LayerSettings layerSettings;
  layerSettings.Name = layerId;
  layerSettings.GrayBand = m_Channels[0];
  if(m_Channels.size() >= 3)
    {
    layerSettings.RedBand = m_Channels[0];
    layerSettings.GreenBand = m_Channels[1];
    layerSettings.BlueBand = m_Channels[2];
    }
  
  //add histogram values
  RenderingFunctionType::ParametersType  paramsMinMax;
  paramsMinMax = renderingFunction->GetParameters();
  layerSettings.MinMaxHisto.resize(2*image->GetNumberOfComponentsPerPixel(),0.);
  for (unsigned int i = 0; i < paramsMinMax.Size(); i = i + 2)
    {
    unsigned int n = i / 2;
    double min = renderingFunction->GetHistogramList()->GetNthElement(n)->Quantile(0, 2./100.);
    double max = renderingFunction->GetHistogramList()->GetNthElement(n)->Quantile(0, 1. - 2./100.);
      
    //save new min max in application settings
    layerSettings.MinMaxHisto[2*m_Channels[n]] = min;
    layerSettings.MinMaxHisto[2*m_Channels[n]+1] = max;
    }
      
  //save quantiles. Default values to 2.
  //these values are updated when the method SetContrastStretch() is called
  layerSettings.MinQuantile = 2.;
  layerSettings.MaxQuantile = 2.;
   
  m_ApplicationSettings.LayersMap[layerId] = layerSettings;
}

void
ViewerEducationModel::SetContrastStretch(const std::string & layerId, 
                                         bool stretchFull, 
                                         bool stretchZoom, 
                                         double lowerQuantile, 
                                         double upperQuantile)
{
  // Select the current image layer
  LayerType* imageLayer = dynamic_cast<LayerType*>(m_VisualizationModel->GetLayerByName(layerId));

  imageLayer->SetUpdateHisto(true);

  // Get the rendering function and cancel the automatic computation
  // of the extremum
  RenderingFunctionType::Pointer renderer = imageLayer->GetRenderingFunction();
  renderer->SetAutoMinMax(true);

  ChannelListType channels = imageLayer->GetRenderingFunction()->GetChannelList();
   
  if(stretchFull)
    {
    imageLayer->SetComputeHistoOnFullResolution(true);
    imageLayer->SetComputeHistoOnZoomResolution(false);
    }
  else if (stretchZoom)
    {
    imageLayer->SetComputeHistoOnFullResolution(false);
    imageLayer->SetComputeHistoOnZoomResolution(true);
    }
  else
    {
    imageLayer->SetComputeHistoOnFullResolution(false);
    imageLayer->SetComputeHistoOnZoomResolution(false);
    }
  
  // Create a new rendering function 
  StandardRenderingFunctionType::Pointer 
    stdRenderer = StandardRenderingFunctionType::New();
  stdRenderer->SetChannelList(channels);
  stdRenderer->SetAutoMinMax(true);
  imageLayer->SetRenderingFunction(stdRenderer);

  // Compute the histograms with the new rendering function and
  // deactivate the automatic computation by setting AutoMinMax to
  // false
  m_VisualizationModel->Update();
  stdRenderer->SetAutoMinMax(false);

  // Update with quantiles values
  RenderingFunctionType::ParametersType  paramsMinMax;
  paramsMinMax = stdRenderer->GetParameters();

  // Update the parameters
  for (unsigned int i = 0; i < paramsMinMax.Size(); i = i + 2)
    {
    unsigned int n = i / 2;
    double min = stdRenderer->GetHistogramList()->GetNthElement(n)->Quantile(0, lowerQuantile/100.);
    double max = stdRenderer->GetHistogramList()->GetNthElement(n)->Quantile(0, 1. - upperQuantile/100.);
    paramsMinMax.SetElement(i, min);
    paramsMinMax.SetElement(i + 1, max);
    
    //save new min max in application settings
    m_ApplicationSettings.LayersMap[layerId].MinMaxHisto[2*channels[n]] = min;
    m_ApplicationSettings.LayersMap[layerId].MinMaxHisto[2*channels[n]+1] = max;
    }
  stdRenderer->SetParameters(paramsMinMax);
   
  //save quantiles
  m_ApplicationSettings.LayersMap[layerId].MinQuantile = lowerQuantile;
  m_ApplicationSettings.LayersMap[layerId].MaxQuantile = upperQuantile;
  
  // Update the rendering model 
  m_VisualizationModel->Update();
}


void
ViewerEducationModel::DoColorMapping(const LabelImageType* image,
                                     const std::string& layerId)
{
   // Fill the min/max values using the ImageLayer hitograms
  unsigned int colorImageComponents = 3;//std::min(m_VectorImage->GetNumberOfComponentsPerPixel(), (unsigned int)3);
  VectorImageType::PixelType minVal;
  VectorImageType::PixelType maxVal;
  minVal.SetSize(m_NumberOfComponentsPerPixel);
  maxVal.SetSize(m_NumberOfComponentsPerPixel);

  // Since the default support image is in falseColor display mode,
  // colorize the label to fit the rendering of the support image
  // Need to use the {NIR,Red,Green} channels
  std::vector<unsigned int>   rgbIndex;
  rgbIndex.push_back(m_NIRChannel);
  rgbIndex.push_back(m_RedChannel);
  rgbIndex.push_back(m_GreenChannel);
  
  for (unsigned int idx = 0; idx < m_NumberOfComponentsPerPixel; idx++)
    {
    minVal.SetElement(idx, m_ApplicationSettings.LayersMap[m_InputImageLabel].MinMaxHisto[2*idx] );
    maxVal.SetElement(idx, m_ApplicationSettings.LayersMap[m_InputImageLabel].MinMaxHisto[2*idx +1]);
    }

    
  RGBFromImageValueFilterType::Pointer RGBFromImageValueFilter = RGBFromImageValueFilterType::New();
  RGBFromImageValueFilter->SetInput(m_VectorImage);
  RGBFromImageValueFilter->SetInputLabelImage(image);
  RGBFromImageValueFilter->SetMinVal(minVal);
  RGBFromImageValueFilter->SetMaxVal(maxVal);
  RGBFromImageValueFilter->Update();
  
  RGBFromImageValueFilterType::MeanValueMapType labelToMeanIntensityMap 
    = RGBFromImageValueFilter->GetMeanValueMap();
  
  m_RGBFromImageValueFilterMap[layerId] = RGBFromImageValueFilter;

  ChangeLabelFilterType::Pointer RBGFromImageMapper = ChangeLabelFilterType::New();
  RBGFromImageMapper->SetInput(image);
  RBGFromImageMapper->SetNumberOfComponentsPerPixel(colorImageComponents);

  RGBFromImageValueFilterType::MeanValueMapType::const_iterator  mapIt = labelToMeanIntensityMap.begin();
  VectorImageType::PixelType meanValue;
  
  VectorPixelType color(colorImageComponents);
  std::map<unsigned int, std::vector<unsigned int> > colorRampMap;
  ColorMapRampType    imageLayerColorRamp;
  while (mapIt != labelToMeanIntensityMap.end())
    {
    LabelType clabel = mapIt->first;
    meanValue = mapIt->second;
    for (unsigned int currentIndex = 0; currentIndex < colorImageComponents; currentIndex++)
      {
        unsigned int dispIndex = rgbIndex[currentIndex];
        color[currentIndex] = (
          (meanValue.GetElement(dispIndex) - minVal.GetElement(dispIndex)) 
          / (maxVal.GetElement(dispIndex) - minVal.GetElement(dispIndex))) * 255.0;
        
        //QGIS only take unsigned int value for color maping
        colorRampMap[clabel].push_back(static_cast<unsigned int>(color[currentIndex]));
      }
    imageLayerColorRamp[clabel] = color;
    RBGFromImageMapper->SetChange(clabel, color);
    ++mapIt;
    }

  // store the colormap for the current layer
  m_ColorMapRampsContainer[layerId] = imageLayerColorRamp;
  
  // Be sure that the layer is rendered with the channels in the right
  // order
  m_Channels.resize(3);
  m_Channels[0] = 0;
  m_Channels[1] = 1;
  m_Channels[2] = 2;

  // Add it to the map
  m_ChangeLabelFilterMap[layerId] = RBGFromImageMapper;
  this->AddRenderingModelToMap(m_ChangeLabelFilterMap[layerId]->GetOutput(), layerId, "", true);
 
  //Save the map into LayerSettings
  m_ApplicationSettings.LayersMap[layerId].Ramp = colorRampMap;
}

void
ViewerEducationModel::ClearMaps()
{
  // Clear Rendering Maps
  m_RenderingModelMap.clear();
  m_PixelDescriptionModelMap.clear();

  m_ColorMapRampsContainer.clear();

  // Clear Application settings
  ApplicationSettings newApplicationSettings;
  m_ApplicationSettings = newApplicationSettings;
}


void
ViewerEducationModel
::SetSpectrumModelChannelsOrder()
{
  if (m_PixelSpectrumModel.IsNotNull())
    {
    m_PixelSpectrumModel->UpdateChannelsOrder(m_BlueChannel, m_GreenChannel, 
                                              m_RedChannel, m_NIRChannel, 
                                              m_MIRChannel);
    }
}

void
ViewerEducationModel
::WriteSpreadsheetFile(const std::string & file)
{
  // check if the right extension is used
  std::string ext = itksys::SystemTools::GetFilenameExtension(file);
  std::ostringstream fullname;
  if ( ext != ".csv")
    {
    fullname<<file  << ".csv";
    }
  else
    {
    fullname << file;
    }
  
  unsigned int nbSpectrum = m_PixelSpectrumModel->GetSeveralSpectrumData().size();
  std::ostringstream content;
  
  for(unsigned int i = 0; i < nbSpectrum; i++)
  {
      PixelSpectrumModelType::SpectrumAndIdVectorType  spectrumStruct = m_PixelSpectrumModel->GetSeveralSpectrumData()[i];
      
      
      content << spectrumStruct.name <<";Coordonn\351es pixel [" << spectrumStruct.coord[0] << ";" << spectrumStruct.coord[1] <<"] \n";
      
      content << "Bande spectrale; Intensit\351\n";
      
      unsigned int size = spectrumStruct.sprectrum.Size();
      for (unsigned int i = 0; i < size; i++)
         {
         content << i << ";" << spectrumStruct.sprectrum[i] << "\n";
         }
      content << "\n\n";
  }
  /** text format (cvs) export */
  std::ofstream csvFile(fullname.str().c_str(), std::ios::out | std::ios::trunc);

  if (csvFile)
    {
    csvFile << content.str() << std::endl;
    csvFile.close();
    }
  else
    {
    otbMsgDevMacro(<< "The file cannot be opened" << std::endl);
    }
}

}// end namespace

