/*=========================================================================

Program:   ORFEO Toolbox
Language:  C++
Date:      $Date$
Version:   $Revision$


Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


This software is distributed WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbViewerEducationModel_h
#define __otbViewerEducationModel_h

#include "otbMVCModel.h"
#include "otbListenerBase.h"
#include "itkObject.h"
#include "itkObjectFactory.h"

#include "otbVectorImage.h"
#include "otbImage.h"
#include "otbImageFileReader.h"

// Visu
#include "otbImageLayerRenderingModel.h"
#include "otbImageLayerGenerator.h"
#include "otbSimpleImageLayer.h"
#include "otbPixelDescriptionModel.h"

// Rendering Function
#include "otbStandardRenderingFunction.h"

// map
#include <map>

#include "ApplicationSettings.h"
#include "TypeDefinitions.h"
#include "otbPixelSpectrumModel.h"

namespace otb
{

enum
{
  Mode_Pleiades,
  Mode_Spot5,
  Mode_Formosat,
  Mode_Unknown,
};

class ViewerEducationModel
  : public MVCModel<ListenerBase>, public itk::Object
{
public:
  /** Standard class typedefs */
  typedef ViewerEducationModel                           Self;
  typedef itk::Object                                    Superclass;
  typedef itk::SmartPointer<Self>                        Pointer;
  typedef itk::SmartPointer<const Self>                  ConstPointer;

  /** Standard type macro */
  itkTypeMacro(ViewerEducationModel, Object);

  itkNewMacro(Self);


  /** Visualization model */
  typedef itk::RGBPixel<unsigned char>                                   RGBPixelType;
  typedef      Image<RGBPixelType, 2>                                    RGBImageType;
  typedef      SimpleImageLayer<VectorImageType, RGBImageType>           LayerType;
  typedef      LayerType::HistogramType                                  HistogramType;
  typedef      ImageLayerGenerator<LayerType>                            LayerGeneratorType;
  typedef      ImageLayerRenderingModel<RGBImageType>                    VisualizationModelType;
  typedef      PixelDescriptionModel<RGBImageType>                       PixelDescriptionModelType;


  typedef Function::RenderingFunction<VectorImageType::PixelType, RGBPixelType> RenderingFunctionType;
  typedef Function::StandardRenderingFunction<VectorImageType::PixelType,
                                              RGBPixelType>             StandardRenderingFunctionType;

  /** type for channels histogram */
  typedef std::vector<double>                                        ChannelsHistogramType;
  typedef std::vector<unsigned int>                                  ChannelListType;

  /** */
  typedef std::map<std::string, VisualizationModelType::Pointer>                  RenderingModelMapType; 
  typedef std::map<std::string, PixelDescriptionModelType::Pointer>               PixelDescriptionModelMapType; 

  typedef otb::PixelSpectrumModel         PixelSpectrumModelType;
  typedef PixelSpectrumModelType::Pointer PixelSpectrumModelPointerType;

  /** Association map typedef */
  typedef std::map<std::string, unsigned int>                                 AssociationMapType;

  /** the changecolor map */  
  typedef std::map<std::string, ColorMapRampType>                       ColorMapRampContainerType;
   
  /** Open an image with its associated label map */
  void OpenImage(const std::string& inputFilename);

  /** Set the utf8 normalisation of the filename */
  void SetFileNameUTF8(const std::string & fname_utf8)
  {
    m_FileNameUTF8 = fname_utf8;
    m_ApplicationSettings.FileName_utf8 = fname_utf8;
  }
  
  /** Get the visualization models */
  itkGetObjectMacro(VisualizationModel, VisualizationModelType);
  itkGetObjectMacro(PixelDescriptionModel, PixelDescriptionModelType);

  /** Set the order of the channels in the spectrum model */ 
  void SetSpectrumModelChannelsOrder();
  
  /** Get the Spectrum model*/
  itkGetObjectMacro(PixelSpectrumModel, PixelSpectrumModelType);
  itkSetObjectMacro(PixelSpectrumModel, PixelSpectrumModelType);
  
  /** Input Image Accessor*/
  itkGetConstObjectMacro(VectorImage, VectorImageType);
 
  std::vector<unsigned int> GetChannels()
  {
    return m_Channels;
  }

   /** Save project */
   void Save(const std::string &);  
   
   
  /** Export project */
  virtual void ExportQGIS(const std::string &, const std::string &);
  virtual void ExportKMZ(const std::string &);
   
  /** Visualization mode */
  void SetPleiadesMode();
  void SetSpot5Mode();
  void SetFormosatMode();
  void SetUnknownMode(bool rememberUserChoice = true);
  void UpdateChannelsChoice( int red,
                             int green, 
                             int blue, 
                             int nir,
                             int mir);
    
  /** Color composition selected */
  void SetTrueColorComposition(const std::string & layerId);
  void SetFalseColorComposition(const std::string & layerId);
  void SetGrayColorComposition(const std::string & layerId,
                               const unsigned int& channel,
                               const std::string & desc);

  /** Clear all maps */
  void ClearMaps();

  /** KMeans processing */
  void KMeans(const std::string& layerId, unsigned int nbClasses);
  
  /** NDVI selected */
  void NDVI(const std::string & layerId);
  
  /** Brightness processing */
  void Brightness(const std::string & layerId);

  /** method for updating contrast from histogram and quantile values */
  void SetContrastStretch(const std::string & layerId, 
                          bool stretchFull, 
                          bool stretchZoom, 
                          double lowerQuantile, 
                          double upperQuantile);

  /** Detect the input image sensor */
  unsigned int GetInputImageMode(const std::string& infname);

  /*** Coloring label images */
  void DoColorMapping(const LabelImageType* image, const std::string& layerId);
 
  /** Get Models Map (should return const, but in the view when getting
    * this map we must const_cast to not discard consteness qualifier)
    */
  RenderingModelMapType GetRenderingModelMap()
  {
    return m_RenderingModelMap;
  }

  PixelDescriptionModelMapType GetPixelDescriptionModelMap()
  {
    return m_PixelDescriptionModelMap;
  }
  
  /** Get the number of components per pixel of the image */
  itkGetConstMacro(NumberOfComponentsPerPixel, unsigned int);
  
  /** Set the project file name */
  itkSetMacro(ProjectFile, std::string);
  itkGetConstMacro(ProjectFile, std::string);

  /** Get the value of each channel */ 
  itkGetConstMacro(RedChannel,  int );
  itkGetConstMacro(BlueChannel,  int );
  itkGetConstMacro(GreenChannel,  int );
  itkGetConstMacro(NIRChannel,  int );
  itkGetConstMacro(MIRChannel,  int );

  /** Get OpenImage flag */
  itkGetMacro(ImageOpened, bool);

  /** Get the label set for the input image */
  itkGetStringMacro(InputImageLabel);

  ApplicationSettings & GetApplicationSettings()
  {
    return m_ApplicationSettings;
  }

  /** Get the image filename */
  itkGetConstMacro(InputFilename, std::string);
  
  /** Set the qgs file name */
  itkSetMacro(QGSFile, std::string);
  itkGetConstMacro(QGSFile, std::string);
  
  /** Set the kmz file name */
  itkSetMacro(KMZFile, std::string);
  itkGetConstMacro(KMZFile, std::string);
  
  void WriteSpreadsheetFile(const std::string & );

  /** Update the association between spectral bands and the relative
     * image composants in the map
     */
  void UpdateSpectralBandsAssociation();

  /** Get the association map */
  const AssociationMapType&  GetAssociationsMap()
  {
    return m_AssociatedSpectralBands;
  }

  /** Get the detected mode */
  itkGetMacro(DetectedMode, unsigned int);
  
protected:
  /** Constructor */
  ViewerEducationModel();

  /** Destructor */
  ~ViewerEducationModel(){};

  /** Method for rendering and Pixel Description models to their map */
  void AddRenderingModelToMap(const VectorImageType* image,
                              const std::string & layerId,
                              const std::string & pixelValueDescription,
                              bool hideRGBPixelValue = false);

private:
  ViewerEducationModel(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /** Visualization models */
  VisualizationModelType::Pointer    m_VisualizationModel;

  /** Pixel description model */
  PixelDescriptionModelType::Pointer m_PixelDescriptionModel;

  /** Vector reader */
  VectorImageType::ConstPointer      m_VectorImage;
  unsigned int                       m_NumberOfComponentsPerPixel;

  /** Vector Image layer generator */
  LayerGeneratorType::Pointer        m_ImageGenerator;
  
  /** Vector for the channels organisation */
  ChannelListType                    m_Channels;
  
  /** Sensor name */
  std::string                        m_SensorID;

  /** Image Reader */
  ReaderType::Pointer                m_Reader;

  /** Channels specifications */
   int   m_RedChannel;
   int   m_GreenChannel;
   int   m_BlueChannel;
   int   m_NIRChannel;
   int   m_MIRChannel;

  /** Channels specified by the user in unknown mode */
  int   m_UserRedChannel;
  int   m_UserGreenChannel;
  int   m_UserBlueChannel;
  int   m_UserNIRChannel;
  int   m_UserMIRChannel;

  /** Rendering Model map */
  RenderingModelMapType              m_RenderingModelMap;
  PixelDescriptionModelMapType       m_PixelDescriptionModelMap;
  
  /** Application settings, used to save and load project */
  ApplicationSettings m_ApplicationSettings;
  
  /** Project File used to save the project*/
  std::string                        m_ProjectFile;
  
  /** Input Image filename */
  std::string                        m_InputFilename;

  /** utf8 normalisation of the filename */
  std::string                       m_FileNameUTF8;
  
  /** QGS File used to export the project*/
  std::string                        m_QGSFile;
  
  /** KMZ File used to export the project*/
  std::string                        m_KMZFile;
  
  /** Filters pointers */
  NDVIFilterType::Pointer m_NDVIFilter;
  ImageToVectorImageFilterType::Pointer m_NDVICastFilter;
  IBFilterType::Pointer m_BrightnessFilter;
  ImageToVectorImageFilterType::Pointer m_BrightnessCastFilter;
  std::map<std::string, RGBFromImageValueFilterType::Pointer> m_RGBFromImageValueFilterMap;
  std::map<std::string, ChangeLabelFilterType::Pointer> m_ChangeLabelFilterMap;
  std::map<std::string, KMeansFilterType::Pointer> m_KMeansFilterMap;
  
  /** Spectrum */
  PixelSpectrumModelPointerType m_PixelSpectrumModel;


  /** flag when file is opened */
  bool                          m_ImageOpened;

  /** Map associating spectral band description and its value */
  AssociationMapType   m_AssociatedSpectralBands;

  unsigned int                  m_DetectedMode;

  ColorMapRampContainerType     m_ColorMapRampsContainer;
  
  std::string                   m_InputImageLabel;

  bool                         m_GrayColorComposition;

};
}
#endif
