/*=========================================================================

Program:   ORFEO Toolbox
Language:  C++
Date:      $Date$
Version:   $Revision$


Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


This software is distributed WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/

#include "otbViewerEducationView.h"
#include "otbMsgReporter.h"

#include <FL/Fl_File_Chooser.H>
#include <FL/Fl_Native_File_Chooser.H>

#include <FL/Fl_Menu_Item.H>

#include "itksys/SystemTools.hxx"

#include <iostream>

#ifdef WIN32
#include "windows.h"
#include "wchar.h"
#endif

namespace otb
{

ViewerEducationView::ViewerEducationView() : m_Controller(),
                                             m_Model(), m_ImageView(),
                                             m_PixelView(), m_ChoosenFileName(""),
                                             m_ChoosenProjectFileName(""), m_Ylength(0),
                                             m_CurrentComponentsPerPixel(0)
{
  m_ImageView = ImageViewType::New();
  m_PixelView = PixelViewType::New();
  m_PixelSpectrumView = PixelSpectrumViewType::New();

  m_Model     = ViewerEducationModel::New();
  m_Model->RegisterListener(this);
  

  // Build the curve widget for histogram
  m_CurveWidget = CurvesWidgetType::New();
  m_CurveWidget->SetXAxisLabel(otbGetTextMacro("Valeur"));
  m_CurveWidget->SetYAxisLabel(otbGetTextMacro("Nombre"));

  // Color definitions
  m_Red.Fill(0.0);
  m_Green.Fill(0.0);
  m_Blue.Fill(0.0);
  m_Grey.Fill(0.5);
  m_Red[0] = 1.0;
  m_Red[3] = 0.5;
  m_Green[1] = 1.0;
  m_Green[3] = 0.5;
  m_Blue[2] = 1.0;
  m_Blue[3] = 0.5;

  // Set the ScaledextractRegion GlComponent yellow
  m_Yellow.Fill(0.0);
  m_Yellow[0] = 1.0;
  m_Yellow[1] = 1.0;
  m_Yellow[3] = 1.0;
  m_ImageView->SetScaledExtractRegionGlComponent(m_Yellow);
  
  // white for the extractRegion
  m_White.Fill(1.0);
  m_ImageView->SetExtractRegionGlComponentColor(m_White);


  // Main image histogram group components : 
  // - handlers
  // - histogram curves
  // - asymptotes
  m_BlueCurveWidgetGroup = CurvesWidgetType::New();
  m_GreenCurveWidgetGroup = CurvesWidgetType::New();
  m_RedCurveWidgetGroup = CurvesWidgetType::New();

  // Set the right description of the axis
  m_BlueCurveWidgetGroup->SetXAxisLabel("Valeur");
  m_GreenCurveWidgetGroup->SetXAxisLabel("Valeur");
  m_RedCurveWidgetGroup->SetXAxisLabel("Valeur");

  m_BlueCurveWidgetGroup->SetYAxisLabel("Nombre");
  m_GreenCurveWidgetGroup->SetYAxisLabel("Nombre");
  m_RedCurveWidgetGroup->SetYAxisLabel("Nombre");
  

  // Vertical asymptotes for each channel
  m_BlueVaCurveR  = VerticalAsymptoteCurveType::New();
  m_BlueVaCurveL  = VerticalAsymptoteCurveType::New();
  m_GreenVaCurveR = VerticalAsymptoteCurveType::New();
  m_GreenVaCurveL = VerticalAsymptoteCurveType::New();
  m_RedVaCurveR   = VerticalAsymptoteCurveType::New();
  m_RedVaCurveL   = VerticalAsymptoteCurveType::New();

  // Add the blue histogram handler
  m_BlueHistogramHandler = HistogramActionHandlerType::New();
  m_BlueHistogramHandler->SetModel(m_Model->GetVisualizationModel());
  m_BlueHistogramHandler->SetView(m_ImageView);
  m_BlueHistogramHandler->SetCurve(m_BlueCurveWidgetGroup);
  m_BlueHistogramHandler->SetLeftAsymptote(m_BlueVaCurveL);
  m_BlueHistogramHandler->SetRightAsymptote(m_BlueVaCurveR);

  // Add the green histogram handler
  m_GreenHistogramHandler = HistogramActionHandlerType::New();
  m_GreenHistogramHandler->SetModel(m_Model->GetVisualizationModel());
  m_GreenHistogramHandler->SetView(m_ImageView);
  m_GreenHistogramHandler->SetCurve(m_GreenCurveWidgetGroup);
  m_GreenHistogramHandler->SetLeftAsymptote(m_GreenVaCurveL);
  m_GreenHistogramHandler->SetRightAsymptote(m_GreenVaCurveR);

  // Add the red histogram handler
  m_RedHistogramHandler = HistogramActionHandlerType::New();
  m_RedHistogramHandler->SetModel(m_Model->GetVisualizationModel());
  m_RedHistogramHandler->SetView(m_ImageView);
  m_RedHistogramHandler->SetCurve(m_RedCurveWidgetGroup);
  m_RedHistogramHandler->SetLeftAsymptote(m_RedVaCurveL);
  m_RedHistogramHandler->SetRightAsymptote(m_RedVaCurveR);
  
  m_BlueCurveWidgetGroup->SetIdentifier("BlueCurve");
  m_GreenCurveWidgetGroup->SetIdentifier("GreenCurve");
  m_RedCurveWidgetGroup->SetIdentifier("RedCurve");
  
  //Set parameters for localization
  fl_cancel = "Annuler";
  Fl_File_Chooser::favorites_label = "Favoris";
  Fl_File_Chooser::all_files_label = "Tous (*)";
  Fl_File_Chooser::filename_label = "Projet";
  Fl_File_Chooser::new_directory_tooltip = "Creer un nouveau dossier";
  Fl_File_Chooser::new_directory_label = "Nouveau dossier";
  Fl_File_Chooser::save_label = "Ok";
  Fl_File_Chooser::add_favorites_label = "Ajouter aux favoris";
  Fl_File_Chooser::custom_filter_label = "Nouveau Filtre";
  Fl_File_Chooser::filesystems_label = "Fichiers systeme";
  Fl_File_Chooser::manage_favorites_label = "Gestion des favoris";
  Fl_File_Chooser::show_label = "Filtre";
  Fl_File_Chooser::preview_label = "visualisation";
  
  //Pixel action handler for spectrum
  m_PixelActionHandler = PixelSpectrumActionHandlerType::New();
}

ViewerEducationView::~ViewerEducationView()
{
  // // Remove registered visualization components from the interface
  // if (gFullResolution != NULL)
  //   {
  //   gFullResolution->remove(m_ImageView->GetFullWidget());
  //   gScroll->remove(m_ImageView->GetScrollWidget());
  //   gZoom->remove(m_ImageView->GetZoomWidget());
  //   gPixelDescription->remove(m_PixelView->GetPixelDescriptionWidget());
  //   }
}

void 
ViewerEducationView::OpenInputImageFilename()
{
 
  Fl_Native_File_Chooser fnfc;
  fnfc.title("Choisissez une image ...");
  fnfc.type(Fl_Native_File_Chooser::BROWSE_FILE);
  fnfc.filter("*.*");
  fnfc.directory(m_ChoosenFileName.c_str());           // default directory to use
  
  // Show native chooser
  switch ( fnfc.show() ) 
    {
    case -1: 
    {   
    MsgReporter::GetInstance()->SendError("empty selection");
    break;  // ERROR
    }
    case  1: 
    {
    break;  // CANCEL nothing to do
    }
    default: 
    {
    const char * cfname = this->Utf8ToANSICharacters(fnfc.filename()); 
    const char * cfname_utf8 = fnfc.filename();
    m_ChoosenFileName_utf8 = cfname_utf8;
 
    if (cfname != NULL || strlen(cfname)<1 )
      {
      //const std::string imgFile = fileChooser.value();
      const std::string imgFile(cfname);
  
      // Set the filename in the GUI
      gInputFilename->value(imgFile.c_str());
  
      // Guess which mode it is
      try
        {
        this->DetectInputImageMode(imgFile);
        m_ChoosenFileName = imgFile;
        }
      catch(itk::ExceptionObject & err)
        {
        MsgReporter::GetInstance()->SendError(err.GetDescription());
        }
      }
    break;  // FILE CHOSEN
    }
    }
}

void
ViewerEducationView::SetSpot5Mode()
{
  m_Model->SetSpot5Mode();
  this->DeactivateChannelsChoice();
  this->UpdateInputDataChoiceBands();
}

void 
ViewerEducationView::SetPleiadesMode()
{
  m_Model->SetPleiadesMode();
  this->DeactivateChannelsChoice();
  this->UpdateInputDataChoiceBands();
}

void 
ViewerEducationView::SetFormosatMode()
{
  m_Model->SetFormosatMode();
  this->DeactivateChannelsChoice();
  this->UpdateInputDataChoiceBands();
}

void 
ViewerEducationView::SetUnknownMode()
{
  m_Model->SetUnknownMode();
  this->ActivateChannelsChoice();
  this->UpdateInputDataChoiceBands();
}

void 
ViewerEducationView::DetectInputImageMode(const std::string& infname)
{
  // Get the sensor mode 
  unsigned int mode = m_Controller->GetInputImageMode(infname);

  // Get the number ofComponents per pixel
  m_CurrentComponentsPerPixel = m_Model->GetNumberOfComponentsPerPixel();
  
  // Update the inputDataChoice GUI 
  this->UseDetectedSensorToUpdateGUI(mode);
}


void 
ViewerEducationView::UseDetectedSensorToUpdateGUI(const unsigned int& mode)
{
  switch(mode)
    {
    case Mode_Pleiades:
    {
    //this->SetPleiadesMode();
    rPleiadesMode->activate();
    rPleiadesMode->value(1);

    // Deactivate and uncheck all the other round box
    rSpotMode->value(0);
    rFormosatMode->value(0);
    rOtherMode->value(0);
    rSpotMode->deactivate();
    rFormosatMode->deactivate();
    rOtherMode->deactivate();
    this->DeactivateChannelsChoice();
    break;
    }
    case Mode_Spot5:
    {
    //this->SetSpot5Mode();
    rSpotMode->value(1);
    rSpotMode->activate();

    // Deactivate and uncheck all the other round box
    rPleiadesMode->value(0);
    rFormosatMode->value(0);
    rOtherMode->value(0);
    rPleiadesMode->deactivate();
    rFormosatMode->deactivate();
    rOtherMode->deactivate();

    this->DeactivateChannelsChoice();
    break;
    }
    case Mode_Formosat:
    {
    //this->SetFormosatMode();
    rFormosatMode->value(1);
    rFormosatMode->activate();

    // Deactivate and uncheck all the other round box
    rSpotMode->value(0);
    rPleiadesMode->value(0);
    rOtherMode->value(0);
    rSpotMode->deactivate();
    rPleiadesMode->deactivate();
    rOtherMode->deactivate();

    this->DeactivateChannelsChoice();
    break;
    }
    default:
    {
    this->SetChannelsChoiceSpinnerLimits();

    if (m_CurrentComponentsPerPixel >= 4)
      {
      // Initial mode set to unkown
      //m_Model->SetUnknownMode(false);
      rOtherMode->value(1);
      rOtherMode->activate();

      // Activate the other round button
      rSpotMode->value(0);
      rPleiadesMode->value(0);
      rFormosatMode->value(0);
      rSpotMode->activate();
      rPleiadesMode->activate();
      rFormosatMode->activate();

      // Activate the channel choice button
      this->ActivateChannelsChoice();
      }
    else
      {
      //m_Model->SetUnknownMode(false);
      rOtherMode->value(1);
      rOtherMode->activate();
      
      rPleiadesMode->value(0);
      rFormosatMode->value(0);

      rPleiadesMode->deactivate();
      rFormosatMode->deactivate();
      this->ActivateChannelsChoice();

      // leave the spot5 mode activated if 3 bands
      rSpotMode->value(0);
      if (m_CurrentComponentsPerPixel == 3 )
        {
        rSpotMode->activate();
        }
      else
        {
        rSpotMode->deactivate();
        }
      }

    break;
    }
    }

  // Update the bands in the GUI
  this->UpdateInputDataChoiceBands();
}

void 
ViewerEducationView::ChooseBetweenBlueAndMIRSpectralBands()
{
  // activating blue or mir spectral band depends on the
  // sensor detected
  if (m_Model->GetDetectedMode() == Mode_Spot5)
    {
    // activate the mir 
    guiMIRChannelChoice->activate();
    guiMIRSpectralBandCheckBox->value(1);
    guiMIRSpectralBandCheckBox->deactivate();

    // deactivate the blue if it was activated
    guiBlueChannelChoice->deactivate();
    guiBlueSpectralBandCheckBox->value(0);
    guiBlueSpectralBandCheckBox->deactivate();
    }
  else
    {
   if (m_Model->GetDetectedMode() == Mode_Pleiades || m_Model->GetDetectedMode() == Mode_Formosat)
     {
     // activate the blue 
     guiBlueChannelChoice->activate();
     guiBlueSpectralBandCheckBox->value(1);
     guiBlueSpectralBandCheckBox->deactivate();

     // deactivate the mir if it was activated
     guiMIRChannelChoice->deactivate();
     guiMIRSpectralBandCheckBox->value(0);
     guiMIRSpectralBandCheckBox->deactivate();
    }
   else if (m_Model->GetDetectedMode() == Mode_Unknown)
     {
      // activate the blue 
      guiBlueChannelChoice->activate();
      guiBlueSpectralBandCheckBox->value(1);
      guiBlueSpectralBandCheckBox->deactivate();

      // deactivate the mir if it was activated
      guiMIRChannelChoice->deactivate();
      guiMIRSpectralBandCheckBox->value(0);

      // keep the MIR checkbox active to be able to choose if needed
      guiMIRSpectralBandCheckBox->activate();
      }
    }
}

void 
ViewerEducationView::HandleInputChoiceCheckBox()
{
  // activating blue or mir spectral band depends on the
  // sensor detected
  if (m_Model->GetDetectedMode() == Mode_Spot5 )
    {
    // activate the mir cb
    guiMIRSpectralBandCheckBox->value(1);

    // deactivate the blue if it was activated
    guiBlueSpectralBandCheckBox->value(0);
    }
  else
    {
    if (m_Model->GetDetectedMode() == Mode_Pleiades || m_Model->GetDetectedMode() == Mode_Formosat)
      {
      // activate the blue cb
      guiBlueSpectralBandCheckBox->value(1);
      
      // deactivate the mir if it was activated
      guiMIRSpectralBandCheckBox->value(0);
      }
    else if (m_Model->GetDetectedMode() == Mode_Unknown)
      {
      // activate the blue cb
      guiBlueSpectralBandCheckBox->value(1);
      
      // deactivate the mir if it was activated
      guiMIRSpectralBandCheckBox->value(0);

      // In unkown mode, activate the checkbox
      guiBlueSpectralBandCheckBox->activate();
      guiMIRSpectralBandCheckBox->activate();
      }
    }
}

void 
ViewerEducationView::UpdateMIRAndBlueChannels()
{

  int red   = guiRedChannelChoice->value()  -1;
  int green = guiGreenChannelChoice->value()-1;
  int nir   = guiNIRChannelChoice->value()  -1;

  int blue  = -1;
  if (guiBlueSpectralBandCheckBox->value())
    {
    blue  = guiBlueChannelChoice->value() -1;
    }
  
  int mir   = -1;
  if (guiMIRSpectralBandCheckBox->value())
    {
    mir = guiMIRChannelChoice->value()  -1;
    }
  
  // Set the orders
  m_Model->UpdateChannelsChoice(red, green, blue, nir, mir);
}

void 
ViewerEducationView::DeactivateChannelsChoice()
{
  guiRedChannelChoice->deactivate();
  guiGreenChannelChoice->deactivate();
  guiBlueChannelChoice->deactivate();
  guiNIRChannelChoice->deactivate();
  guiMIRChannelChoice->deactivate();
  guiMIRSpectralBandCheckBox->deactivate();
  guiBlueSpectralBandCheckBox->deactivate();
  
  // deactivate the checkbox if 3 components
  
  if (m_CurrentComponentsPerPixel == 3)
    {
    guiMIRSpectralBandCheckBox->value(0);
    guiBlueSpectralBandCheckBox->value(0);
    }
  
  // in 4 bands images : choose between Blue and MIR
  if (m_CurrentComponentsPerPixel == 4)
    {
    this->HandleInputChoiceCheckBox();
    }
}

void 
ViewerEducationView::ActivateChannelsChoice()
{
  guiRedChannelChoice->activate();
  guiGreenChannelChoice->activate();
  guiNIRChannelChoice->activate();
  
  // deactivate the blue and MIR channel selection if 3 channels image
  if (m_CurrentComponentsPerPixel == 3)
    {
    guiBlueChannelChoice->deactivate();
    guiMIRChannelChoice->deactivate();
    guiMIRSpectralBandCheckBox->deactivate();
    guiBlueSpectralBandCheckBox->deactivate();
    guiMIRSpectralBandCheckBox->value(0);
    guiBlueSpectralBandCheckBox->value(0);
    }
  
  // in 4 bands images : choose between Blue and MIR
  if (m_CurrentComponentsPerPixel == 4)
    {
    this->ChooseBetweenBlueAndMIRSpectralBands();
    }
}

void 
ViewerEducationView::SetChannelsChoiceSpinnerLimits()
{
  double  nbComponents = m_CurrentComponentsPerPixel;
  guiRedChannelChoice->range(1,nbComponents);
  guiGreenChannelChoice->range(1,nbComponents);
  guiBlueChannelChoice->range(1,nbComponents);
  guiNIRChannelChoice->range(1,nbComponents);
  guiMIRChannelChoice->range(1,nbComponents);
}

void 
ViewerEducationView::UpdateInputDataChoiceBands()
{
  guiRedChannelChoice->value(m_Model->GetRedChannel()+1);
  guiGreenChannelChoice->value(m_Model->GetGreenChannel()+1);
  guiBlueChannelChoice->value(m_Model->GetBlueChannel()+1);
  guiNIRChannelChoice->value(m_Model->GetNIRChannel()+1);
  guiMIRChannelChoice->value(m_Model->GetMIRChannel()+1);
}

void 
ViewerEducationView::SetChannelsOrderFromGUI()
{
  int red   = guiRedChannelChoice->value()  -1;
  int green = guiGreenChannelChoice->value()-1;
  int blue  = guiBlueChannelChoice->value() -1;
  int nir   = guiNIRChannelChoice->value()  -1;
  int mir   = guiMIRChannelChoice->value()  -1;
  
  m_Model->UpdateChannelsChoice(red, green, blue, nir, mir);
}

void 
ViewerEducationView::UpdateChannelsSelectionInGUI()
{
  Fl_Text_Buffer * buffer = new Fl_Text_Buffer();
  gChannelsDescription->buffer(buffer);

  std::ostringstream channelsDesc;
  channelsDesc << "Plan R <- BS_PIR "<<std::endl;
  gChannelsDescription->insert(channelsDesc.str().c_str());
  channelsDesc.str("");

  channelsDesc << "Plan V <- BS_R "<<std::endl;
  gChannelsDescription->insert(channelsDesc.str().c_str());
  channelsDesc.str("");

  channelsDesc << "Plan B <- BS_V "<<std::endl;
  gChannelsDescription->insert(channelsDesc.str().c_str());
  channelsDesc.str("");

  // channelsDesc << "Bande PIR : "<< guiNIRChannelChoice->value()<<std::endl;
  // gChannelsDescription->insert(channelsDesc.str().c_str());
  // channelsDesc.str("");
}

void 
ViewerEducationView::LaunchApplication()
{
  // deactivate the inputDataChoice while loading
  gInputDataChoice->deactivate();
  
  // Get the filename 
  std::string infname;
  if(rUseFilename->value())
  {
     infname = gInputFilename->value();
  }
  else if (rUseProjectname->value())
  {
     infname = gInputProjectname->value();
  }
  
  if (!infname.empty())
    {
    // Build the GUI  
    this->BuildGUI();
    this->Build();


    //Update the channels order in the model
    this->SetChannelsOrderFromGUI();

    // Update the channels order in the GUI
    this->UpdateChannelsSelectionInGUI();

    // update one last time
    this->UpdateMIRAndBlueChannels();

    // tell the model to update the association map
    m_Model->UpdateSpectralBandsAssociation();

    // Do some checking
    if( rUseFilename->value() )
      {
      try
        {
        // save the choosen file
        m_ChoosenFileName = infname;
        
        // Open the image
        m_Model->OpenImage(infname);

        // Save the utf8 normalisation of the filename
        m_Model->SetFileNameUTF8(m_ChoosenFileName_utf8);
        
        // Close the choice gui
        wInputDataChoice->hide();

        // Set the order of the channels in the spectrum model 
        m_Model->SetSpectrumModelChannelsOrder();
        
        // Show the GUI
        this->Show();

        }
      catch(itk::ExceptionObject & )
        {
        std::ostringstream error;
        error <<"Problème lors de l'ouverture du fichier "<< infname;
        MsgReporter::GetInstance()->SendError(error.str());
        }
      }
    else if( rUseProjectname->value() )
      {      
      // Load project
      this->Load(infname);
      }
    }

  // Activate the ouvrir/quitter button while launching
  gInputDataChoice->activate();
}

void 
ViewerEducationView::ReloadInputGUIInformation()
{

  // image/or project selected
  // Get the filename 
  std::string infname;
  try
    {
    // load the image filename
    if(rUseFilename->value())
      {
      infname = gInputFilename->value();
      if (!infname.empty())
        {
        this->DetectInputImageMode(infname);
        }
      }
    else if (rUseProjectname->value())
      {
      // load the project filename
      infname = gInputProjectname->value();
      if (!infname.empty())
        {
        this->LoadVepFile(m_ChoosenProjectFileName);
        }
      }
    }
  catch(itk::ExceptionObject & )
    {
    std::ostringstream oss;
    oss << "Problème lors de l'ouverture du fichier "<< infname;
    MsgReporter::GetInstance()->SendError(oss.str().c_str());
    }

}

void 
ViewerEducationView::OpenInputProjectFilename()
{
  Fl_Native_File_Chooser fnfc;
  fnfc.title("Choisissez un fichier projet  ...");
  fnfc.type(Fl_Native_File_Chooser::BROWSE_FILE);
  fnfc.filter("*.vep");
  fnfc.directory(m_ChoosenProjectFileName.c_str());           // default directory to use
 
  // Show native chooser
 switch ( fnfc.show() ) 
 {
   case -1: 
   {   
   MsgReporter::GetInstance()->SendError("Aucun fichier selectionn\351");
   break;  // ERROR
   }
   case  1: 
   {
    break;  // CANCEL nothing to do
   }
   default: 
   {
   const char * cfname = this->Utf8ToANSICharacters(fnfc.filename());
 
    if (cfname != NULL || strlen(cfname)<1 )
    {
    const std::string projectFile(cfname);
  
    // Set the filename in the GUI
    gInputProjectname->value(projectFile.c_str());
    m_ChoosenProjectFileName = projectFile; 
    this->LoadVepFile(m_ChoosenProjectFileName);
    }
    break;  // FILE CHOSEN
   }
 }
 
}

void ViewerEducationView::Build()
{
  // Build the visualization part
  m_ImageView->SetModel(m_Model->GetVisualizationModel());
  m_PixelView->SetModel(m_Model->GetPixelDescriptionModel());
  
  if(!m_Controller)
    {
    // Controller is not set, can not build view
    itkExceptionMacro(<<"Une erreur fatale s'est produite");
    }

  if(!m_WidgetsController)
    {
    // WidgetsController is not set, can not build view
    itkExceptionMacro(<<"Une erreur fatale s'est produite");
    }
  
  //Add the pixel action handler and deactivate it
  //it will be activate only when opening Spectral curves
  m_PixelActionHandler->SetIsActive(false);
  m_WidgetsController->AddActionHandler(m_PixelActionHandler);
  
  
  
  // Register controllers
  m_ImageView->SetController(m_WidgetsController);

  // Remove registered visualization components from the interface
  gFullResolution->add(m_ImageView->GetFullWidget());
  gScroll->add(m_ImageView->GetScrollWidget());
  gZoom->add(m_ImageView->GetZoomWidget());
  gPixelDescription->add(m_PixelView->GetPixelDescriptionWidget());
  gMainWindowHistogram->add(m_CurveWidget);
  

  gFullResolution->resizable(m_ImageView->GetFullWidget());
  gScroll->resizable(m_ImageView->GetScrollWidget());
  gZoom->resizable(m_ImageView->GetZoomWidget());
  gPixelDescription->resizable(m_PixelView->GetPixelDescriptionWidget());
  gMainWindowHistogram->resizable(m_CurveWidget);
  
  m_ImageView->GetZoomWidget()->resize(gZoom->x(),
                                       gZoom->y(),
                                       gZoom->w(),
                                       gZoom->h());
  
  m_ImageView->GetFullWidget()->resize(gFullResolution->x(),
                                       gFullResolution->y(),
                                       gFullResolution->w(),
                                       gFullResolution->h());
  
  m_ImageView->GetScrollWidget()->resize(gScroll->x(),
                                         gScroll->y(),
                                         gScroll->w(),
                                         gScroll->h());
  
  m_PixelView->GetPixelDescriptionWidget()->resize(gPixelDescription->x(),
                                                   gPixelDescription->y(),
                                                   gPixelDescription->w(),
                                                   gPixelDescription->h());

  m_CurveWidget->resize(gMainWindowHistogram->x(), gMainWindowHistogram->y(),
                        gMainWindowHistogram->w(), gMainWindowHistogram->h());
}

void ViewerEducationView::ShowInputDataChoice()
{
  // Build the fltk code for Main GUI 
  this->BuildInputDataChoice();

  // Show the input data gui
  wInputDataChoice->show();
}

void ViewerEducationView::SetController(ViewerEducationControllerInterface * controller)
{
  m_Controller = controller;
  m_Controller->SetModel(m_Model);
  m_PixelSpectrumView->SetController(m_Controller->GetPixelSpectrumController());
}

void ViewerEducationView::SetWidgetsController(ImageWidgetController * controller)
{
  m_WidgetsController = controller;
}

void ViewerEducationView::Notify()
{
  // build the main histogram curves
  if (m_Model->GetImageOpened())
    {
    this->BuildHistogramCurves();
    }
}

void ViewerEducationView::Close()
{
  // Close the windows
  this->Exit();

  // Clear all the maps in the View/Controller/Model
  m_ImageViewMap.clear();
  m_PixelViewMap.clear();
  m_SimplePackedWidgetMap.clear();
  m_CurvesWidgetMap.clear();

  // Initialize gui components
  cColorComposition->clear();

  // instanciate new pointers to the Image/Pixel view
  m_ImageView = ImageViewType::New();
  m_PixelView = PixelViewType::New();
  m_PixelSpectrumView = PixelSpectrumViewType::New();
  m_PixelActionHandler = PixelSpectrumActionHandlerType::New();

  
  // Set the extract region Gl Component color to yellow
  m_ImageView->SetExtractRegionGlComponentColor(m_White);
  m_ImageView->SetScaledExtractRegionGlComponent(m_Yellow);
  
  // Initialize the controller, done after the ImageView
  // initialization cause the controller class uses the view
  m_Controller->InitializeController();

  // Initialize the model
  m_Model     = ViewerEducationModel::New();
  m_Model->RegisterListener(this);

  // Set the new model into the controller
  m_Controller->SetModel(m_Model);

  // Add the blue histogram handler
  m_BlueHistogramHandler->SetModel(m_Model->GetVisualizationModel());
  m_BlueHistogramHandler->SetView(m_ImageView);

  // Add the green histogram handler
  m_GreenHistogramHandler->SetModel(m_Model->GetVisualizationModel());
  m_GreenHistogramHandler->SetView(m_ImageView);

  // Add the red histogram handler
  m_RedHistogramHandler->SetModel(m_Model->GetVisualizationModel());
  m_RedHistogramHandler->SetView(m_ImageView);
  
  // Show the inputDataChoice
  this->wInputDataChoice->show();

  // have to detect the input image mode : since the model is reset
  // this information is not available anymore : recompute it again
  std::string oldfname( gInputFilename->value() );
  if (!oldfname.empty() && rUseFilename->value())
    {
    this->DetectInputImageMode(oldfname);
    }

  // reload the project when leaving
  std::string oldprojfname(gInputProjectname->value());
  if ( !oldprojfname.empty() && rUseProjectname->value())
    {
    this->LoadVepFile(oldprojfname);
    }
}

void ViewerEducationView::Exit()
{
  // Close all the standard image viewer
  SimplePackedWidgetMapType::iterator it = m_SimplePackedWidgetMap.begin();
  while(it != m_SimplePackedWidgetMap.end())
    {
    (*it).second->Hide();
    it++;
    }
  
  // Close the other window
  wSpectrumCurves->hide();
  
  // Close the other window
  wMainHistogram->hide();

  // Finally close the main window
  wMainWindow->hide();

  // Close the Msg Reporter
  MsgReporter::GetInstance()->Hide();
}

void ViewerEducationView::SaveAs()
{
  Fl_Native_File_Chooser fnfc;
  fnfc.title("Enregistrer sous ...");
  fnfc.type(Fl_Native_File_Chooser::BROWSE_SAVE_FILE);
  fnfc.filter("*.vep");
  fnfc.directory(m_Model->GetProjectFile().c_str());           // default directory to use
  
 // Show native chooser
 switch ( fnfc.show() ) 
 {
   case -1: 
   {   
   MsgReporter::GetInstance()->SendError("Aucun fichier selectionn\357");
   break;  // ERROR
   }
   case  1: 
   {
    break;  // CANCEL nothing to do
   }
   default: 
   {
   const char * cfname = this->Utf8ToANSICharacters(fnfc.filename());
 
    if (cfname != NULL || strlen(cfname)<1 )
    {
     const std::string projectFile(cfname);
     m_Model->SetProjectFile(projectFile);
     
     this->RemoveClosedLayersFromMap();

     m_Controller->Save(projectFile);
    }
    break;  // FILE CHOSEN
   }
 }
}

void ViewerEducationView::RemoveClosedLayersFromMap()
{
   //remove hide window from layerSettings
   SimplePackedWidgetMapType::const_iterator it;
   for(it = m_SimplePackedWidgetMap.begin(); it != m_SimplePackedWidgetMap.end(); it++ )
   {
      const SimplePackedWidgetManager::Pointer packedWidget = (*it).second;
      if(!m_SimplePackedWidgetMap[(*it).first]->Shown())
      {
         //delete the corresponding layerSettings
         m_Model->GetApplicationSettings().LayersMap.erase((*it).first);
      }
   }
}

void ViewerEducationView::ExportQGIS()
{

  Fl_Native_File_Chooser fnfc;
  fnfc.title("Export du projet ...");
  fnfc.type(Fl_Native_File_Chooser::BROWSE_SAVE_FILE);
  fnfc.filter("*.qgs*");
  fnfc.directory(m_Model->GetQGSFile().c_str());           // default directory to use
  
 // Show native chooser
 switch ( fnfc.show() ) 
 {
   case -1: 
   {   
   MsgReporter::GetInstance()->SendError("Aucun fichier selectionn\357");
   break;  // ERROR
   }
   case  1: 
   {
    break;  // CANCEL nothing to do
   }
   default: 
   {
   const char * cfname = this->Utf8ToANSICharacters(fnfc.filename());
   const char * cfname_utf8 = fnfc.filename();
 
    if (cfname != NULL || strlen(cfname)<1 )
    {
     const std::string qgsFile(cfname);
     const std::string qgsFile_utf8(cfname_utf8);
     
     this->RemoveClosedLayersFromMap();

     m_Controller->ExportQGIS(qgsFile, qgsFile_utf8);
    }
    break;  // FILE CHOSEN
   }
 }
}

void ViewerEducationView::ExportKMZ()
{
  Fl_Native_File_Chooser fnfc;
  fnfc.title("Export du projet ...");
  fnfc.type(Fl_Native_File_Chooser::BROWSE_SAVE_FILE);
  fnfc.filter("*.kmz*");
  fnfc.directory(m_Model->GetQGSFile().c_str());           // default directory to use
  
 // Show native chooser
 switch ( fnfc.show() ) 
 {
   case -1: 
   {   
   MsgReporter::GetInstance()->SendError("Aucun fichier selectionn\357");
   break;  // ERROR
   }
   case  1: 
   {
    break;  // CANCEL nothing to do
   }
   default: 
   {
   const char * cfname = this->Utf8ToANSICharacters(fnfc.filename());
 
    if (cfname != NULL || strlen(cfname)<1 )
    {
     const std::string kmzFile(cfname);
     m_Model->SetKMZFile(kmzFile);
     
     this->RemoveClosedLayersFromMap();

     m_Controller->ExportKMZ(kmzFile);
    }
    break;  // FILE CHOSEN
   }
 }

 
}

void
ViewerEducationView::Show()
{
  try
    {
    // deactivate, add some items in the GUI knowing the image
    this->DoGUIChecking();

    // Rebuild the image view
    this->wMainWindow->show();
    m_ImageView->GetZoomWidget()->show();
    m_ImageView->GetFullWidget()->show();
    m_ImageView->GetScrollWidget()->show();
    m_PixelView->GetPixelDescriptionWidget()->show();
    m_CurveWidget->show();
    }
  catch (itk::ExceptionObject& err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
}

void ViewerEducationView::DoGUIChecking()
{
  // Get the number of components of the image
  unsigned int nbComponents = m_Model->GetNumberOfComponentsPerPixel();

  // if a single channel,  deactivate : 
  // - the true color mode
  // - the false color composition mode
  // - indicies computation
  if (nbComponents == 1 )
    {
    bVegetationIndex->deactivate();
    bBrightnessIndex->deactivate();
    guiSpectralProfile->deactivate();
    }
  else
    {
    // the else if here if the user launch the application the first
    // time with monoband image, closed the app, and then reload it
    // with a multiband image
    bVegetationIndex->activate();
    bBrightnessIndex->activate();
    }

  // Add menuitems to the FL_choice and set callbacks
  if (nbComponents >= 2 )
    {
    
    // label of Couleur naturelles entry
    std::ostringstream oss;
    cColorComposition->add("Couleurs naturelles");
    
    Fl_Menu_Item *  colorCompoItem = 
      const_cast<Fl_Menu_Item *>(cColorComposition->find_item("Couleurs naturelles"));
    colorCompoItem->labelsize(12);

    //  if blue is -1 deactivate the true color composition
    if (m_Model->GetBlueChannel() == -1)
    {
      colorCompoItem->deactivate();
    }
    else
    {
      UserDataStructType* colorCompoUserData = new UserDataStructType;
      colorCompoUserData->pthis = this;
      colorCompoItem->callback(TrueColorCompositionCallback, (void*)colorCompoUserData);
    }

    // Update the Color composition gui with the spectral band values
    //stored in the map
    ViewerEducationModel::AssociationMapType::const_iterator associationIt = m_Model->GetAssociationsMap().begin();
   
    while(associationIt != m_Model->GetAssociationsMap().end())
      {
      // Add the fl_menu_item for valid spectral band
      if ((*associationIt).second != 4099)
        {
        // label of the menu item
        std::ostringstream oss;
        oss <<(*associationIt).first;
        cColorComposition->add(oss.str().c_str());

        Fl_Menu_Item *  item = const_cast<Fl_Menu_Item *>(cColorComposition->find_item(oss.str().c_str()));
        if (item != NULL)
          {
          UserDataStructType* userdata = new UserDataStructType;
          userdata->pthis = this;
          userdata->channelSelected = (*associationIt).second;
          userdata->desc            = oss.str();
          item->callback(GrayScaleColorComposition, (void*)userdata);
          item->labelsize(12);
          }
        }
      associationIt++;
      }//end while
    }
  else
    {
    cColorComposition->deactivate();
    }

  // Add the filename to the wMainWindow as label
  std::ostringstream oss;
  oss<< "Visualisation image : "<< m_Model->GetInputFilename()<< " (composition fausse couleur infrarouge)";
  this->wMainWindow->copy_label(oss.str().c_str());
}

void 
ViewerEducationView::GrayScaleColorComposition(Fl_Widget* w, void* v)
{
  // Cast to UserDataStructType*
  UserDataStructType* userdata = static_cast<UserDataStructType*>(v);
  
  // Get the channel selected
  unsigned int channelSelected = userdata->channelSelected;

  // Get the desc
  std::string  desc            = userdata->desc;
  
  // call the callback 
  (userdata->pthis)->GrayScaleColorComposition(channelSelected, desc);
}

void 
ViewerEducationView::TrueColorCompositionCallback(Fl_Widget* w, void* v)
{
  // Cast to UserDataStructType*
  UserDataStructType* userdata = static_cast<UserDataStructType*>(v);
  
  // call the callback 
  (userdata->pthis)->TrueColorComposition();
}

void
ViewerEducationView::TrueColorComposition()
{
  // Deactivate the object Interface
  gObjectsInterface->deactivate();
  gButtons->deactivate();

  // Build a ImageView for this visualization mode and add it to the
  // ImageView Map
  std::string layerId = "Couleurs naturelles";
  if (m_ImageViewMap.find(layerId) == m_ImageViewMap.end())
    {
    ImageViewType::Pointer currentImageView = ImageViewType::New();
    PixelViewType::Pointer currentPixelView = PixelViewType::New();
    m_ImageViewMap[layerId] = currentImageView;
    m_PixelViewMap[layerId] = currentPixelView;

    m_Controller->SetTrueColorComposition(layerId);
    }
  
  // Display the current image
  this->DisplayWidgetByName(layerId);

  // Set the color composition to the first value
  cColorComposition->value(0);

  // Deactivate the object Interface
  gObjectsInterface->activate();
  gButtons->activate();
}


void
ViewerEducationView::GrayScaleColorComposition(const unsigned int& channel, const std::string & desc )
{
  // Deactivate the object Interface
  gObjectsInterface->deactivate();
  gButtons->deactivate();
  
  // Build a ImageView for this visualization mode and add it to the
  // ImageView Map
  std::string layerId = "Visualisation ";
  std::ostringstream oss;
  oss << layerId<<" "<< desc;
  std::string newlayerId = oss.str();

  if (m_ImageViewMap.find(newlayerId) == m_ImageViewMap.end())
    {

    ImageViewType::Pointer currentImageView = ImageViewType::New();
    PixelViewType::Pointer currentPixelView = PixelViewType::New();
    m_ImageViewMap[newlayerId] = currentImageView;
    m_PixelViewMap[newlayerId] = currentPixelView;

    m_Controller->SetGrayColorComposition(newlayerId, channel,desc) ;
    }

  // Display the current image
  this->DisplayWidgetByName(newlayerId);

  // Set the color composition to the first value
  cColorComposition->value(0);

  // Deactivate the object Interface
  gObjectsInterface->activate();
  gButtons->activate();
}

void
ViewerEducationView::KMeans()
{
  // Deactivate the object Interface
  gObjectsInterface->deactivate();
  gButtons->deactivate();

  // Set the pBar to busy
  std::ostringstream ossBusy;
  ossBusy << "Traitement en cours ...";
  bBusy->copy_label(ossBusy.str().c_str());
  bBusy->show();
  Fl::flush();

  // Build a ImageView for this visualization mode and add it to the
  // ImageView Map
  unsigned int kmeansClasses = (unsigned int)(bKMeansClasses->value());
  std::string layerId = "K-Moyennes, N. classes : ";
  std::ostringstream oss;
  oss << layerId <<  kmeansClasses;
  std::string newLayerId = oss.str();
  if (m_ImageViewMap.find(newLayerId) == m_ImageViewMap.end())
    {
    ImageViewType::Pointer currentImageView = ImageViewType::New();
    PixelViewType::Pointer currentPixelView = PixelViewType::New();
    m_ImageViewMap[newLayerId] = currentImageView;
    m_PixelViewMap[newLayerId] = currentPixelView;

    m_Controller->KMeans(newLayerId, kmeansClasses);
    }
  // Display the current image
  this->DisplayWidgetByName(newLayerId);

  // hide the busy group
  bBusy->hide();
  
  // Deactivate the object Interface
  gObjectsInterface->activate();
  gButtons->activate();
}


void
ViewerEducationView::NDVI()
{
  // Deactivate the object Interface
  gObjectsInterface->deactivate();
  gButtons->deactivate();

  // Set the pBar to busy
  std::ostringstream ossBusy;
  ossBusy << "Traitement en cours ...";
  bBusy->copy_label(ossBusy.str().c_str());
  bBusy->show();
  Fl::flush();
  
  // Build a ImageView for this visualization mode and add it to the
  // ImageView Map
  std::string layerId = "NDVI";
  if (m_ImageViewMap.find(layerId) == m_ImageViewMap.end())
    {
    ImageViewType::Pointer currentImageView = ImageViewType::New();
    PixelViewType::Pointer currentPixelView = PixelViewType::New();
    m_ImageViewMap[layerId] = currentImageView;
    m_PixelViewMap[layerId] = currentPixelView;
    
    m_Controller->NDVI(layerId);
    }
  
  // Display the current image
  this->DisplayWidgetByName(layerId);

  // hide the busy group
  bBusy->hide();

  // Deactivate the object Interface
  gObjectsInterface->activate();
  gButtons->activate();
}

void
ViewerEducationView::Brightness()
{
  // Deactivate the object Interface
  gObjectsInterface->deactivate();
  gButtons->deactivate();
  
  // Set the pBar to busy
  std::ostringstream ossBusy;
  ossBusy << "Traitement en cours ...";
  bBusy->copy_label(ossBusy.str().c_str());
  bBusy->show();
  Fl::flush();
  
  // Build a ImageView for this visualization mode and add it to the
  // ImageView Map
  std::string layerId = "Indice de brillance";
  if (m_ImageViewMap.find(layerId) == m_ImageViewMap.end())
    {
    ImageViewType::Pointer currentImageView = ImageViewType::New();
    PixelViewType::Pointer currentPixelView = PixelViewType::New();
    m_ImageViewMap[layerId] = currentImageView;
    m_PixelViewMap[layerId] = currentPixelView;
    
    m_Controller->Brightness(layerId);
    }
  
  // Display the current image
  this->DisplayWidgetByName(layerId);

  // hide the busy group
  bBusy->hide();

  // Deactivate the object Interface
  gObjectsInterface->activate();
  gButtons->activate();
}

void
ViewerEducationView::DisplayWidgetByName(const std::string & layerId)
{
  // If the display exists in the map, don't recompute its image layer
  if ( m_SimplePackedWidgetMap.find(layerId) != m_SimplePackedWidgetMap.end())
    {
    // If the display is already shown, nothing to do
    if ( !m_SimplePackedWidgetMap[layerId]->Shown())
      {
      this->InternalDisplayBuildAndShow(layerId);
      }
    }
  else
    {
    m_ImageViewMap[layerId]->SetController(m_Controller->GetControllerMap()[layerId]);
    m_ImageViewMap[layerId]->SetModel(m_Model->GetRenderingModelMap()[layerId]);
    m_PixelViewMap[layerId]->SetModel(m_Model->GetPixelDescriptionModelMap()[layerId]);
    
    // set the region color to yellow
    m_ImageViewMap[layerId]->SetExtractRegionGlComponentColor(m_White);
    m_ImageViewMap[layerId]->SetScaledExtractRegionGlComponent(m_Yellow);
    
    this->InternalDisplayBuildAndShow(layerId);
    }
}

void 
ViewerEducationView::InternalDisplayBuildAndShow(const std::string& layerId)
{
  // build the histogram for this layer
  this->BuildHistogramCurvesByName(layerId);

  // build the display
  SimplePackedWidgetManager::Pointer display = SimplePackedWidgetManager::New();
  display->RegisterFullWidget(m_ImageViewMap[layerId]->GetFullWidget());
  display->RegisterScrollWidget(m_ImageViewMap[layerId]->GetScrollWidget());
  display->RegisterZoomWidget(m_ImageViewMap[layerId]->GetZoomWidget());
  display->RegisterHistogramWidget(m_CurvesWidgetMap[layerId]);
  display->RegisterPixelDescriptionWidget(m_PixelViewMap[layerId]->GetPixelDescriptionWidget());
  display->SetLabel(layerId.c_str());
  display->SetResizable(false);

  // Add a callback
  UserDataStructType* packedWidgetUserData = new UserDataStructType;
  packedWidgetUserData->pthis = this;
  packedWidgetUserData->id    = layerId;
  display->AddCallback(&ViewerEducationView::TestCallback, (void*)(packedWidgetUserData));

  // Set the widget labels in french
  display->SetLabelScrollWidget("Carte de navigation");
  display->SetLabelFullWidget("R\351solution maximale");
  display->SetLabelHistogramWidget("Histogrammes");
  display->SetLabelPixelDescriptionWidget("Description des pixels");
  display->SetLabelZoomWidget("Zoom");    

  // Resize 
  display->ResizeFull(gFullResolution->w(),gFullResolution->h());
  display->ResizeZoom(gZoom->w(),gZoom->h());
  display->ResizeScroll(gScroll->w(),gScroll->h());
  display->ResizeHistogramCurve(gMainWindowHistogram->w(),gMainWindowHistogram->h());
  display->ResizePixelDescription(gPixelDescription->w(),gPixelDescription->h());
  display->AdjustWidgets();

  // Add this display to the map
  m_SimplePackedWidgetMap[layerId] = display;    
  display->Show();

  // Set the main window not resizable
  this->SetMainWindowNotResizable();
}

void 
ViewerEducationView::TestCallback(Fl_Widget* w, void* v)
{
  // Cast to UserDataStructType*
  UserDataStructType* userdata = static_cast<UserDataStructType*>(v);

  ViewerEducationView * _this = userdata->pthis;
  std::string         layerId = userdata->id;
  
  // do the callback 
  _this->CloseWindow(layerId);
}

void 
ViewerEducationView::CloseWindow(const std::string& layerId)
{
  if (m_SimplePackedWidgetMap.find(layerId) != m_SimplePackedWidgetMap.end())
    {
    m_SimplePackedWidgetMap[layerId]->Hide();
    m_SimplePackedWidgetMap[layerId]->SetResizable(true);
    }

  // Make the wMainWindow resizable when no sub-viewer is shown
  bool allDisplayClosed = true;
  SimplePackedWidgetMapType::const_iterator itDisplay = m_SimplePackedWidgetMap.begin();  
  while(itDisplay != m_SimplePackedWidgetMap.end())
    {
    if ( (*itDisplay).second->Shown())
      {
      allDisplayClosed = false;
      break;
      }
    ++itDisplay;
    }

  if (allDisplayClosed)
    {
    this->SetMainWindowResizable();
    }
  else
    {
    this->SetMainWindowNotResizable();
    }
}

void
ViewerEducationView::SetMainWindowResizable()
{
  this->wMainWindow->size_range(1,1,
                                10000,10000);
  this->wMainWindow->resizable(this->wMainWindow);
}

void
ViewerEducationView::SetMainWindowNotResizable()
{
  this->wMainWindow->size_range(wMainWindow->w(),wMainWindow->h(),
                                wMainWindow->w(),wMainWindow->h());
  this->wMainWindow->resizable(NULL);
}

void
ViewerEducationView::OpenMainHistogramWidget()
{
  wMainHistogram->show();
  this->BuildSeparateHistogramCurves();
}

void ViewerEducationView::BuildHistogramCurves()
{
  // Clear the widget
  m_CurveWidget->ClearAllCurves();

  // Select the current image layer
  ViewerEducationModel::LayerType* 
    imageLayer = dynamic_cast<ViewerEducationModel::LayerType*>(m_Model->GetVisualizationModel()->GetLayerByName(m_Model->GetInputImageLabel()));

  unsigned int pixelRepresentationSize = imageLayer->GetRenderingFunction()->GetPixelRepresentationSize();

  if (pixelRepresentationSize >= 3)
    {
    HistogramCurveType::Pointer bhistogram = HistogramCurveType::New();
    bhistogram->SetHistogramColor(m_Blue);
    bhistogram->SetLabelColor(m_Blue);
    bhistogram->SetHistogram(imageLayer->GetHistogramList()->GetNthElement(2));
    m_CurveWidget->AddCurve(bhistogram);
    }

  if (pixelRepresentationSize >= 2)
    {
    HistogramCurveType::Pointer ghistogram = HistogramCurveType::New();
    ghistogram->SetHistogramColor(m_Green);
    ghistogram->SetLabelColor(m_Green);
    ghistogram->SetHistogram(imageLayer->GetHistogramList()->GetNthElement(1));
    m_CurveWidget->AddCurve(ghistogram);
    }

  HistogramCurveType::Pointer rhistogram = HistogramCurveType::New();
  if (pixelRepresentationSize == 1)
    {
    rhistogram->SetHistogramColor(m_Grey);
    rhistogram->SetLabelColor(m_Grey);
    }
  else
    {
    rhistogram->SetHistogramColor(m_Red);
    rhistogram->SetLabelColor(m_Red);
    }

  rhistogram->SetHistogram(imageLayer->GetHistogramList()->GetNthElement(0));
  m_CurveWidget->AddCurve(rhistogram);
}

void
ViewerEducationView::BuildHistogramCurvesByName(const std::string& layerId)
{
  CurvesWidgetType::Pointer curveWidget = CurvesWidgetType::New();
  curveWidget->SetXAxisLabel(otbGetTextMacro("Valeur"));
  curveWidget->SetYAxisLabel(otbGetTextMacro("Nombre"));
  
  // Select the current image layer
  ViewerEducationModel::LayerType* imageLayer = 
    dynamic_cast<ViewerEducationModel::LayerType*>(
      m_Model->GetRenderingModelMap()[layerId]->GetLayerByName(layerId));

  unsigned int pixelRepresentationSize = imageLayer->GetRenderingFunction()->GetPixelRepresentationSize();

  if (pixelRepresentationSize >= 3)
    {
    HistogramCurveType::Pointer bhistogram = HistogramCurveType::New();
    bhistogram->SetHistogramColor(m_Blue);
    bhistogram->SetLabelColor(m_Blue);
    bhistogram->SetHistogram(imageLayer->GetHistogramList()->GetNthElement(2));
    curveWidget->AddCurve(bhistogram);
    }

  if (pixelRepresentationSize >= 2)
    {
    HistogramCurveType::Pointer ghistogram = HistogramCurveType::New();
    ghistogram->SetHistogramColor(m_Green);
    ghistogram->SetLabelColor(m_Green);
    ghistogram->SetHistogram(imageLayer->GetHistogramList()->GetNthElement(1));
    curveWidget->AddCurve(ghistogram);
    }

  HistogramCurveType::Pointer rhistogram = HistogramCurveType::New();
  if (pixelRepresentationSize == 1)
    {
    rhistogram->SetHistogramColor(m_Grey);
    rhistogram->SetLabelColor(m_Grey);
    }
  else
    {
    rhistogram->SetHistogramColor(m_Red);
    rhistogram->SetLabelColor(m_Red);
    }

  rhistogram->SetHistogram(imageLayer->GetHistogramList()->GetNthElement(0));
  curveWidget->AddCurve(rhistogram);

  m_CurvesWidgetMap[layerId] = curveWidget;
}

void ViewerEducationView::DeactivateHistogramActionHandlers()
{
  m_RedHistogramHandler->SetIsActive(false);
  m_GreenHistogramHandler->SetIsActive(false);
  m_BlueHistogramHandler->SetIsActive(false);

  // Update the histogram with values set in the quantiles
  this->SetContrastStretch();
}

void ViewerEducationView::ActivateHistogramActionHandlers()
{
  m_RedHistogramHandler->SetIsActive(true);
  m_GreenHistogramHandler->SetIsActive(true);
  m_BlueHistogramHandler->SetIsActive(true);
}

void ViewerEducationView::BuildSeparateHistogramCurves()
{
  // Clear the widget
  m_BlueCurveWidgetGroup->ClearAllCurves();
  m_GreenCurveWidgetGroup->ClearAllCurves();
  m_RedCurveWidgetGroup->ClearAllCurves();
  
  m_BlueCurveWidgetGroup->hide();
  m_GreenCurveWidgetGroup->hide();
  m_RedCurveWidgetGroup->hide();
  
  // Set the widget controller
  m_BlueCurveWidgetGroup->SetController(m_WidgetsController);
  m_GreenCurveWidgetGroup->SetController(m_WidgetsController);
  m_RedCurveWidgetGroup->SetController(m_WidgetsController);

  m_WidgetsController->AddActionHandler(m_RedHistogramHandler);
  m_WidgetsController->AddActionHandler(m_GreenHistogramHandler);
  m_WidgetsController->AddActionHandler(m_BlueHistogramHandler);

  // Get the number of subdivisons in the GUI : 3 blanks, left, middle
  // and right of the histogram widgets
  int blank = 4;
  int width = (int) ((gHistogram->w() - 3*blank) / 2.0);
  int height = (int) ((gHistogram->h() -3* blank) / 2.0);

  // Select the current image layer
  ViewerEducationModel::LayerType* 
    imageLayer = dynamic_cast<ViewerEducationModel::LayerType*>(m_Model->GetVisualizationModel()->GetLayerByName(m_Model->GetInputImageLabel()));

  // Select the current rendering function
  ViewerEducationModel::RenderingFunctionType::Pointer renderer = imageLayer->GetRenderingFunction();

  // Set the rendering to the histogram handlers
  
  m_GreenHistogramHandler->SetRenderingFunction(renderer);
  m_RedHistogramHandler->SetRenderingFunction(renderer);

  
  // Get the association map
  ViewerEducationModel::AssociationMapType asso = 
    const_cast<ViewerEducationModel::AssociationMapType&>(m_Model->GetAssociationsMap());

  /// BLUE band : draw the BLUE band histogram and the asymptote
  if (imageLayer->GetRenderingFunction()->GetHistogramList()->Size() >= 3)
    {
    m_BlueHistogramHandler->SetRenderingFunction(renderer);
    
    HistogramCurveType::Pointer bhistogram = HistogramCurveType::New();
    bhistogram->SetHistogramColor(m_Blue);
    bhistogram->SetLabelColor(m_Blue);
    bhistogram->SetHistogram(imageLayer->GetHistogramList()->GetNthElement(2));
    // Before rendering is called to get the extremum of the histogram
    bhistogram->BeforeRendering();
    m_BlueCurveWidgetGroup->AddCurve(bhistogram);

    // Add to the gHistogram group
    gHistogram->add(m_BlueCurveWidgetGroup);
    gHistogram->resizable(m_BlueCurveWidgetGroup);
    m_BlueCurveWidgetGroup->resize(gHistogram->x() + blank, gHistogram->y() + height + 2*blank, width, height);

    // Left Asymptote
    m_BlueVaCurveL->SetAbcisse(renderer->GetParameters().GetElement(4));
    m_BlueVaCurveL->SetVerticalAsymptoteColor(m_Blue);

    // Right Asymptote
    m_BlueVaCurveR->SetAbcisse(renderer->GetParameters().GetElement(5));
    m_BlueVaCurveR->SetVerticalAsymptoteColor(m_Blue);

    // Add the asymptote to the curve
    m_BlueCurveWidgetGroup->AddCurve(m_BlueVaCurveR);
    m_BlueCurveWidgetGroup->AddCurve(m_BlueVaCurveL);

    // Set the channel to work on and the limits of the asymptotes
    m_BlueHistogramHandler->SetChannel(2);
    m_BlueHistogramHandler->SetMinimumAbcisse(bhistogram->GetMinimum()[0]);
    m_BlueHistogramHandler->SetMaximumAbcisse(bhistogram->GetMaximum()[0]);

    // Set the title of this curve
    std::ostringstream band;
    band << "Plan B : BS_V ";// <- ("<< asso["Bande spectrale Verte"] + 1 <<")";
    m_BlueCurveWidgetGroup->SetTitle(band.str());
    m_BlueCurveWidgetGroup->ZeroCrossingAxisOff();

    m_BlueCurveWidgetGroup->show();
    }
  /// GREEN band : draw the GREEN band histogram and the asymptote
  if (imageLayer->GetRenderingFunction()->GetHistogramList()->Size() >= 2)
    {
    HistogramCurveType::Pointer ghistogram = HistogramCurveType::New();
    ghistogram->SetHistogramColor(m_Green);
    ghistogram->SetLabelColor(m_Green);
    ghistogram->SetHistogram(imageLayer->GetHistogramList()->GetNthElement(1));
    // Before rendering is called to get the extremum of the histogram
    ghistogram->BeforeRendering();
    m_GreenCurveWidgetGroup->AddCurve(ghistogram);

    //Add to the gHistogram group
    gHistogram->add(m_GreenCurveWidgetGroup);
    gHistogram->resizable(m_GreenCurveWidgetGroup);
    m_GreenCurveWidgetGroup->resize(gHistogram->x() + width + 2*blank, gHistogram->y() + blank, width, height);

    // Left Asymptote
    m_GreenVaCurveL->SetAbcisse(renderer->GetParameters().GetElement(2));
    m_GreenVaCurveL->SetVerticalAsymptoteColor(m_Green);

    // Right Asymptote
    m_GreenVaCurveR->SetAbcisse(renderer->GetParameters().GetElement(3));
    m_GreenVaCurveR->SetVerticalAsymptoteColor(m_Green);

    // Add the asymptote to the curve
    m_GreenCurveWidgetGroup->AddCurve(m_GreenVaCurveR);
    m_GreenCurveWidgetGroup->AddCurve(m_GreenVaCurveL);

    // Set the channel to work on and the limits of the asymptotes
    m_GreenHistogramHandler->SetChannel(1);
    m_GreenHistogramHandler->SetMinimumAbcisse(ghistogram->GetMinimum()[0]);
    m_GreenHistogramHandler->SetMaximumAbcisse(ghistogram->GetMaximum()[0]);
    
    // Set the title of this curve
    std::ostringstream band;
    band << "Plan V : BS_R";//  <- ("<< asso["Bande spectrale Rouge"] + 1<<")";
    m_GreenCurveWidgetGroup->SetTitle(band.str());
    m_GreenCurveWidgetGroup->ZeroCrossingAxisOff();

    m_GreenCurveWidgetGroup->show();
    }
  /// RED band : draw the RED band histogram and the asymptote
  HistogramCurveType::Pointer rhistogram = HistogramCurveType::New();
  if (imageLayer->GetRenderingFunction()->GetHistogramList()->Size() == 1)
    {
    rhistogram->SetHistogramColor(m_Grey);
    rhistogram->SetLabelColor(m_Grey);

    //Add to the gHistogram group
    gHistogram->add(m_RedCurveWidgetGroup);
    gHistogram->resizable(m_RedCurveWidgetGroup);
    m_RedCurveWidgetGroup->resize(gHistogram->x() + blank, gHistogram->y()+ blank, width, height);

    // Set the title of this curve
    std::ostringstream band;
    band << "Niveau de gris";
    m_RedCurveWidgetGroup->SetTitle(band.str());
    m_RedCurveWidgetGroup->ZeroCrossingAxisOff();

    }
  else
    {
    rhistogram->SetHistogramColor(m_Red);
    rhistogram->SetLabelColor(m_Red);

    //Add to the gHistogram group
    gHistogram->add(m_RedCurveWidgetGroup);
    gHistogram->resizable(m_RedCurveWidgetGroup);
    m_RedCurveWidgetGroup->resize(gHistogram->x()+ blank, gHistogram->y()+ blank, width, height);
    
    // Set the title of this curve
    std::ostringstream band;
    band << "Plan R : BS_PIR";//  <- ("<< asso["Bande spectrale PIR"] + 1<<")";
    m_RedCurveWidgetGroup->SetTitle(band.str());
    m_RedCurveWidgetGroup->ZeroCrossingAxisOff();
    }

  rhistogram->SetHistogram(imageLayer->GetHistogramList()->GetNthElement(0));
  // Before rendering is called to get the extremum of the histogram
  rhistogram->BeforeRendering();

  m_RedCurveWidgetGroup->AddCurve(rhistogram);
  // Left Asymptote
  m_RedVaCurveL->SetAbcisse(renderer->GetParameters().GetElement(0));
  m_RedVaCurveL->SetVerticalAsymptoteColor(rhistogram->GetLabelColor());

  // Right Asymptote
  m_RedVaCurveR->SetAbcisse(renderer->GetParameters().GetElement(1));
  m_RedVaCurveR->SetVerticalAsymptoteColor(rhistogram->GetLabelColor());

  // Add the asymptote to the curve
  m_RedCurveWidgetGroup->AddCurve(m_RedVaCurveR);
  m_RedCurveWidgetGroup->AddCurve(m_RedVaCurveL);

  // Set the channel to work on and the limits of the asymptotes
  m_RedHistogramHandler->SetChannel(0);
  m_RedHistogramHandler->SetMinimumAbcisse(rhistogram->GetMinimum()[0]);
  m_RedHistogramHandler->SetMaximumAbcisse(rhistogram->GetMaximum()[0]);

  m_RedCurveWidgetGroup->show();
}

void ViewerEducationView::SetContrastStretch()
{
  // Update the contrast knowing the user values for quantiles
  std::string layerId = m_Model->GetInputImageLabel();
  m_Model->SetContrastStretch(layerId, 
                              guiStretchFull->value(),
                              guiStretchZoom->value(),
                              bLowerQuantile->value(),
                              bUpperQuantile->value());
  
  // Update the histogram curves with the new quantiles and the new
  // used rendering function
  this->BuildSeparateHistogramCurves();
}

void
ViewerEducationView::LoadVepFile(const std::string & projectFile)
{
  m_Model->SetProjectFile(projectFile);

  // TODO : should use the controller
  try
    {
    load(projectFile, m_AppSettingsFromProjectFile);
    m_ChoosenProjectFileName = projectFile;
    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }

  //build GUI
  this->BuildGUI();
  this->Build();


  // set the current components per pixel
  m_CurrentComponentsPerPixel = m_AppSettingsFromProjectFile.NbComponents;
    
  // set the limits of the spinner values
  this->SetChannelsChoiceSpinnerLimits();
  
  // Set the mode
  if (m_AppSettingsFromProjectFile.Sensor == ApplicationSettings::PLEIADE)
    {
    m_Model->SetPleiadesMode();
    // Update the inputDataChoice GUI 
    this->UseDetectedSensorToUpdateGUI(Mode_Pleiades);
    }
  else if (m_AppSettingsFromProjectFile.Sensor == ApplicationSettings::SPOT)
    {   
     m_Model->SetSpot5Mode();
    
    // Update the inputDataChoice GUI 
    this->UseDetectedSensorToUpdateGUI(Mode_Spot5);
    }
  else if (m_AppSettingsFromProjectFile.Sensor == ApplicationSettings::UNKNOWN)
    {

     m_Model->SetUnknownMode();
    // Update the inputDataChoice GUI 
    this->UseDetectedSensorToUpdateGUI(Mode_Unknown);

    int red = m_AppSettingsFromProjectFile.DefaultChannels.RedBand;
    int green = m_AppSettingsFromProjectFile.DefaultChannels.GreenBand;
    int blue = m_AppSettingsFromProjectFile.DefaultChannels.BlueBand;
    int nir = m_AppSettingsFromProjectFile.DefaultChannels.NIRBand;
    int mir = m_AppSettingsFromProjectFile.DefaultChannels.MIRBand;
    m_Model->UpdateChannelsChoice(red,green,blue,nir, mir);
    }
  guiRedChannelChoice->value(m_AppSettingsFromProjectFile.DefaultChannels.RedBand + 1);
  guiGreenChannelChoice->value(m_AppSettingsFromProjectFile.DefaultChannels.GreenBand + 1);
  guiBlueChannelChoice->value(m_AppSettingsFromProjectFile.DefaultChannels.BlueBand + 1);
  guiNIRChannelChoice->value(m_AppSettingsFromProjectFile.DefaultChannels.NIRBand + 1);
  guiMIRChannelChoice->value(m_AppSettingsFromProjectFile.DefaultChannels.MIRBand + 1);
  guiBlueSpectralBandCheckBox->value(m_AppSettingsFromProjectFile.BlueAndMIRStatus.BlueStatus);
  guiMIRSpectralBandCheckBox->value(m_AppSettingsFromProjectFile.BlueAndMIRStatus.MIRStatus);
  this->UpdateChannelsSelectionInGUI();
}

void
ViewerEducationView::Load(const std::string & projectFile)
{
  // TODO : should use the controller
  try
    {
    //Open image
    m_Model->OpenImage(m_AppSettingsFromProjectFile.FileName);
    
    // Close the choice gui
    wInputDataChoice->hide();
    
    // Show the GUI
    this->Show();
    }
  catch(itk::ExceptionObject & err)
    {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
    }

  std::map<std::string, LayerSettings>::const_iterator itLayer;
  for(itLayer = m_AppSettingsFromProjectFile.LayersMap.begin(); itLayer != m_AppSettingsFromProjectFile.LayersMap.end(); itLayer++ )
  {
     const LayerSettings::LayerType & type = (*itLayer).second.Type;
     switch (type)
     {
      case LayerSettings::ColorComposite:
      {
         //Check that this is not the main layer
      if( (*itLayer).second.Name != m_Model->GetInputImageLabel())
         {
            if( (*itLayer).second.Mode == LayerSettings::TrueColors)
            {
               this->TrueColorComposition();
            }
            else if ( (*itLayer).second.Mode == LayerSettings::GrayLevel)
            {
            this->GrayScaleColorComposition((*itLayer).second.GrayBand, (*itLayer).second.Desc);
            }
         }
         //else set the contrast values
         else
         {
            bLowerQuantile->value((*itLayer).second.MinQuantile);
            bUpperQuantile->value((*itLayer).second.MaxQuantile);
            this->SetContrastStretch();
         }
         break;
      }
      case LayerSettings::NDVI:
      {
         this->NDVI();
         break;
      }
      case LayerSettings::KMeans:
         //
         bKMeansClasses->value((*itLayer).second.NbClasses);
         this->KMeans();
         break;
      case LayerSettings::Brightness:
         this->Brightness();
         break;
     }
   }
}

void
ViewerEducationView::OpenSpectrumCurves()
{
  m_PixelActionHandler->SetController(m_Controller->GetPixelSpectrumController());
  m_PixelActionHandler->SetView(m_ImageView);
  m_PixelActionHandler->SetIsActive(true);
  m_PixelSpectrumView->SetModel(m_Model->GetPixelSpectrumModel());
  m_PixelSpectrumView->SetController(m_Controller->GetPixelSpectrumController());
  
  gCurves->add(m_PixelSpectrumView->GetCurveWidget());
  gCurves->resizable(m_PixelSpectrumView->GetCurveWidget());
  m_PixelSpectrumView->GetCurveWidget()->resize(gCurves->x(), gCurves->y(), gCurves->w(), gCurves->h());
  
  // Set the Fl_Scroll needed in the SpectralView to add the curves
  m_PixelSpectrumView->SetScrollList(m_ScrollCurveList);
                                                      
  bZoomRoulette->bounds(0, 10000);
  bZoomRoulette->value(m_Controller->GetPixelSpectrumController()->GetYAxisLenghtControl());
   
  bStepRoulette->bounds(1, 800);
  bStepRoulette->value(m_Controller->GetPixelSpectrumController()->GetYGridStep());
  bStepRoulette->step(3);


  // tell the model to update the channels order in the spectrumModel
  m_Model->SetSpectrumModelChannelsOrder();
                                                      
  wSpectrumCurves->show();
  m_PixelSpectrumView->GetCurveWidget()->show();

  //m_Controller->GetPixelSpectrumController()->UpdateWidgetView();
}

void ViewerEducationView::DeactivatePixelActionHandler()
{
  m_PixelActionHandler->SetIsActive(false);
  wSpectrumCurves->hide();
}

void
ViewerEducationView
::StepRouletteCallBack()
{
  m_Controller->GetPixelSpectrumController()->SetAutomaticGridStep(false);
  m_Controller->GetPixelSpectrumController()->SetGridStepY(bStepRoulette->value());
  m_Controller->GetPixelSpectrumController()->UpdateWidgetView();

}

void
ViewerEducationView
::ZoomRouletteCallBack()
{
  float lengthWidget;
  if (m_Ylength == 0)
    {
    bZoomRoulette->maximum(m_Model->GetPixelSpectrumModel()->GetMaxValue());
    m_Ylength = m_PixelSpectrumView->GetCurveWidget()->GetAxisLength()[1];
    }

  if ((m_Ylength - bZoomRoulette->value()) <= bZoomRoulette->minimum())
    {
    lengthWidget = bZoomRoulette->minimum();
    }
  else
    {
    lengthWidget = m_Ylength - bZoomRoulette->value();
    }

  bZoomRoulette->step(lengthWidget / 40);
  m_Controller->GetPixelSpectrumController()->SetYAxisLenghtControl(lengthWidget, true);
}

void
ViewerEducationView
::AutoGridCallBack()
{
  int checkValue = (int) bAutoGridStep->value();

  if (checkValue)
    {
    m_Controller->GetPixelSpectrumController()->SetAutomaticGridStep(true);
    bStepRoulette->deactivate();
    }
  else
    {
    m_Controller->GetPixelSpectrumController()->SetAutomaticGridStep(false);
    bStepRoulette->activate();
    }
}

void
ViewerEducationView
::HideRollOverCurveCallBack()
{

  int checkValue = (int) mHideRollOverCurveButton->value();

  if (checkValue) m_Controller->GetPixelSpectrumController()->SetCurentCurveAlpha(0);
  else m_Controller->GetPixelSpectrumController()->SetCurentCurveAlpha(1);

  m_Controller->GetPixelSpectrumController()->UpdateWidgetView();

}

void
ViewerEducationView
::ExportCSV()
{
   m_PixelSpectrumView->Update();
   /*unsigned int Nb = m_ScrollCurveList->children();
   std::cout<<"NBBBBBBBBB === "<<Nb<<std::endl;
   unsigned int nbCurves = m_PixelSpectrumView->GetLabels().size();
   for (unsigned int c = 0 ; c < nbCurves; c++)
   {
     unsigned int Id = m_PixelSpectrumView->GetLabels()[c]->GetId();
     m_Controller->GetPixelSpectrumController()->SetSpectrumName(Id, )
   }*/
  Fl_Native_File_Chooser fnfc;
  fnfc.title("Enregistrer les spectres ... ...");
  fnfc.type(Fl_Native_File_Chooser::BROWSE_SAVE_FILE);
  fnfc.filter("*.csv");
  fnfc.directory(m_Model->GetProjectFile().c_str());           // default directory to use
  
 // Show native chooser
 switch ( fnfc.show() ) 
 {
   case -1: 
   {   
   MsgReporter::GetInstance()->SendError("Aucun fichier selectionn\357");
   break;  // ERROR
   }
   case  1: 
   {
    break;  // CANCEL nothing to do
   }
   default: 
   {
   const char * cfname = this->Utf8ToANSICharacters(fnfc.filename());
   if (cfname != NULL || strlen(cfname)<1 )
    {
     m_Controller->WriteSpreadsheetFile(std::string(cfname));
    }
    break;  // FILE CHOSEN
   }
 }
}

const char *
ViewerEducationView
::Utf8ToANSICharacters(const char * inputUtf8)
{
#ifdef WIN32
  int length = MultiByteToWideChar(CP_UTF8, 0, (LPCSTR)inputUtf8, -1, NULL, 0);
  if (length > 0)
    {
    wchar_t* wide = new wchar_t[length];
    MultiByteToWideChar(CP_UTF8, 0, (LPCSTR)inputUtf8, -1, wide, length);

    // convert it to ANSI, use setlcale() to set your locale, if not set
    size_t convertedChars = 0;
    char* ansi = new char[length];
    wcstombs_s(&convertedChars, ansi, length, wide, _TRUNCATE);
    
    return ansi;
    }
#endif
  return inputUtf8;
}

} //end namespace otb
