/*=========================================================================

Program:   ORFEO Toolbox
Language:  C++
Date:      $Date$
Version:   $Revision$


Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


This software is distributed WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbViewerEducationView_h
#define __otbViewerEducationView_h


#include "itkObject.h"
#include "itkObjectFactory.h"

#include "otbListenerBase.h"
#include "otbViewerEducationModel.h"
#include "otbViewerEducationControllerInterface.h"
#include "otbViewerEducationGUI.h"

#include "otbImageView.h"
#include "otbPixelDescriptionView.h"
#include "otbVectorDataGlComponent.h"

#include "otbSimplePackedWidgetManager.h"

// histogram
#include "otbHistogramActionHandler.h"
#include "otbCurves2DWidget.h"
#include "otbHistogramCurve.h"

// spectrum
#include "otbPixelSpectrumView.h"
#include "otbPixelSpectrumActionHandler.h"
#include "otbPixelSpectrumController.h"

//TEMP
#include "otbHistogramActionHandler.h"


namespace otb
{
/** \class ViewerEducationView
 *
 */
class ITK_EXPORT ViewerEducationView
  :  public ListenerBase, public ViewerEducationGUI, public itk::Object
{
public:
  /** Standard class typedefs */
  typedef ViewerEducationView  Self;
  typedef itk::Object                    Superclass;
  typedef itk::SmartPointer<Self>        Pointer;
  typedef itk::SmartPointer<const Self>  ConstPointer;

  /** Standard macros */
  itkNewMacro(Self);
  itkTypeMacro(ViewerEducationView, Object);

  /** Visualization typedefs */
  typedef ViewerEducationModel::RGBPixelType              RGBPixelType;
  typedef ViewerEducationModel::VisualizationModelType    VisualizationModelType;
  typedef ViewerEducationModel::PixelDescriptionModelType PixelDescriptionModelType;

  typedef ImageView<VisualizationModelType>                   ImageViewType;
  typedef PixelDescriptionView<PixelDescriptionModelType>     PixelViewType;

  /** Curve widget for histogram */
  typedef Curves2DWidget                                      CurvesWidgetType;
  typedef HistogramCurve<ViewerEducationModel::HistogramType> HistogramCurveType;

  typedef std::map<std::string, ImageViewType::Pointer >      ImageViewMapType;
  typedef std::map<std::string, PixelViewType::Pointer >      PixelViewTypeMapType;
  typedef std::map<std::string, SimplePackedWidgetManager::Pointer> SimplePackedWidgetMapType;
  typedef std::map<std::string, CurvesWidgetType::Pointer>    CurvesWidgetMapType;

  // Histogram curves handler & asymptotes
  typedef VerticalAsymptoteCurve                              VerticalAsymptoteCurveType;
  typedef otb::HistogramActionHandler
    <ViewerEducationModel::VisualizationModelType, 
     ImageViewType, 
     ViewerEducationModel::RenderingFunctionType>             HistogramActionHandlerType;
  
  typedef otb::PixelSpectrumController         PixelSpectrumControllerType;
  typedef otb::PixelSpectrumActionHandler
  <PixelSpectrumControllerType,
   ImageViewType>                                               PixelSpectrumActionHandlerType;
  
  /** Spectrum */
  typedef otb::PixelSpectrumView         PixelSpectrumViewType;
  typedef PixelSpectrumViewType::Pointer PixelSpectrumViewPointerType;
  
  // Struct to pass to a callback
  typedef struct userdataStruct
  {
    ViewerEducationView*  pthis;
    unsigned int          channelSelected;
    std::string           id;
    std::string           desc;
  } UserDataStructType;
  
  /** Event from the model */
  virtual void Notify();

  /** Set the application controller */
  void SetController(ViewerEducationControllerInterface * controller);

  /** Set the widgets controller */
  void SetWidgetsController(ImageWidgetController * controller);

  /** Build the interface */
  virtual void Build();

  /** Show the Gui */
  virtual void Show();
  
  /** Get a pointer to the view parts for the controller */
  itkGetObjectMacro(ImageView, ImageViewType);
  itkGetObjectMacro(PixelView, PixelViewType);
  
  itkGetObjectMacro(PixelSpectrumView, PixelSpectrumViewType);
  itkSetObjectMacro(PixelSpectrumView, PixelSpectrumViewType);

  /** GUI Callbacks */
  virtual void Exit();
  virtual void Close();
  virtual void SaveAs();
  virtual void ExportQGIS();
  virtual void ExportKMZ();
  virtual void TrueColorComposition();
  virtual void GrayScaleColorComposition(const unsigned int& channel, const std::string & desc);
  virtual void KMeans();
  virtual void NDVI();
  virtual void Brightness();
  virtual void OpenMainHistogramWidget();
  virtual void OpenSpectrumCurves();
  virtual void SetContrastStretch();
  virtual void OpenInputImageFilename();
  virtual void OpenInputProjectFilename();
  virtual void SetChannelsOrderFromGUI();
  virtual void LaunchApplication();
  virtual void DeactivateHistogramActionHandlers();
  virtual void ActivateHistogramActionHandlers();

  virtual void SetSpot5Mode();
  virtual void SetPleiadesMode();
  virtual void SetFormosatMode();
  virtual void SetUnknownMode();
  virtual void UpdateMIRAndBlueChannels();

  void BuildSeparateHistogramCurves();
  void DoGUIChecking();
  void ShowInputDataChoice();
  void StepRouletteCallBack();
  void ZoomRouletteCallBack();
  void AutoGridCallBack();
  void HideRollOverCurveCallBack();
  void ExportCSV();
  void DeactivatePixelActionHandler();

  /** Get the Image and pixel view maps */
  ImageViewMapType GetImageViewMap()
  {
    return m_ImageViewMap;
  }
  
  PixelViewTypeMapType GetPixelViewMap()
  {
    return m_PixelViewMap;
  }
  
  /** Open project */
   void Load(const std::string &);
   void LoadVepFile(const std::string &);

   /** Callback associated to each item of the Fl_Choice colorComposition */
   static void GrayScaleColorComposition(Fl_Widget*w , void * v);
   static void TrueColorCompositionCallback(Fl_Widget*w , void * v);
   static void TestCallback(Fl_Widget* w, void* v);


   void CloseWindow(const std::string& layerId);

   /** Callback to reload the informations of the choosen image/project
     * selected 
     */
   virtual void ReloadInputGUIInformation();

protected:
  /** Constructor */
  ViewerEducationView();
  
  /** Destructor */
  ~ViewerEducationView();

  void DisplayWidgetByName(const std::string & layerId);

  void BuildHistogramCurves();
  void BuildHistogramCurvesByName(const std::string& layerId);
  void DetectInputImageMode(const std::string& infname);
  void UseDetectedSensorToUpdateGUI(const unsigned int & mode);
  void UpdateInputDataChoiceBands();
  void SetChannelsChoiceSpinnerLimits();

  void DeactivateChannelsChoice();
  void ActivateChannelsChoice();

  void UpdateChannelsSelectionInGUI();

  void SetMainWindowResizable();
  void SetMainWindowNotResizable();

  void InternalDisplayBuildAndShow(const std::string& layerId);

  // helper function for win32 sys to convert utf8 chars to ansi
  const char *Utf8ToANSICharacters(const char * inputUtf8);

  void ChooseBetweenBlueAndMIRSpectralBands();
  void HandleInputChoiceCheckBox();
  
private:
  ViewerEducationView(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented
  
  
  /** This method erase LayerSettings map item in the 
   * m_model.m_ApplicationSetting.LayersMap map if the layer window is closed .
   */
  void RemoveClosedLayersFromMap();
  
  /** Pointer to the controller */
  ViewerEducationControllerInterface *      m_Controller;
  
  /** Pointer to the widgets controller */
  ImageWidgetController *                   m_WidgetsController;

  /** Pointer to the model */
  ViewerEducationModel::Pointer             m_Model;

  /** Image view */
  ImageViewType::Pointer                    m_ImageView;

  /** Pixel view */
  PixelViewType::Pointer                    m_PixelView;

  /** Image and pixel View maps */
  ImageViewMapType                          m_ImageViewMap;
  PixelViewTypeMapType                      m_PixelViewMap;
  SimplePackedWidgetMapType                 m_SimplePackedWidgetMap;
  
  /** Curve widget */
  CurvesWidgetType::Pointer                 m_CurveWidget;
  CurvesWidgetMapType                       m_CurvesWidgetMap;

  /** Used colors for histograms */
  HistogramCurveType::ColorType             m_Red;
  HistogramCurveType::ColorType             m_Green;
  HistogramCurveType::ColorType             m_Blue;
  HistogramCurveType::ColorType             m_Grey;
  ImageViewType::ColorType                  m_Yellow;
  ImageViewType::ColorType                  m_White;

  // Histograms action handler
  HistogramActionHandlerType::Pointer       m_BlueHistogramHandler;
  HistogramActionHandlerType::Pointer       m_GreenHistogramHandler;
  HistogramActionHandlerType::Pointer       m_RedHistogramHandler;

  /** Curve widget */
  CurvesWidgetType::Pointer                 m_BlueCurveWidgetGroup;
  CurvesWidgetType::Pointer                 m_GreenCurveWidgetGroup;
  CurvesWidgetType::Pointer                 m_RedCurveWidgetGroup;
  /** Vertical Asymptotes*/
  VerticalAsymptoteCurveType::Pointer       m_BlueVaCurveR;
  VerticalAsymptoteCurveType::Pointer       m_BlueVaCurveL;
  VerticalAsymptoteCurveType::Pointer       m_GreenVaCurveR;
  VerticalAsymptoteCurveType::Pointer       m_GreenVaCurveL;
  VerticalAsymptoteCurveType::Pointer       m_RedVaCurveR;
  VerticalAsymptoteCurveType::Pointer       m_RedVaCurveL;

  std::string                               m_ChoosenFileName;
  std::string                               m_ChoosenProjectFileName;
  std::string                               m_ChoosenFileName_utf8;

  /** Spectrum */
  PixelSpectrumViewPointerType              m_PixelSpectrumView;
  float                                     m_Ylength;
  PixelSpectrumActionHandlerType::Pointer   m_PixelActionHandler;


  /** AppSettings from a project file */
  ApplicationSettings                       m_AppSettingsFromProjectFile; 

  /** nbComponents per pixel could change when switching between
    * InputImage fname /Project fname 
    */
  unsigned int                m_CurrentComponentsPerPixel;
};
}//end namespace otb

#endif

