/*=========================================================================

Program:   ORFEO Toolbox
Language:  C++
Date:      $Date$
Version:   $Revision$


Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


This software is distributed WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/

#include "ApplicationSettings.h"
#include <iostream>
#include "otbMacro.h"
#include <boost/lexical_cast.hpp>

#include "itksys/SystemTools.hxx"
#include "otbOGRHelpers.h"
#include "ogr_spatialref.h"
#include "cpl_string.h" // For CPLFree

#include <time.h>


std::string const StringEnumSensorTab[] = {"PLEIADE","SPOT","FORMOSAT","UNKNOWN"};
std::string const StringEnumTypeTab[] = {"ColorComposite","NDVI","KMeans","Brightness"};
std::string const StringEnumDisplayTab[] = {"TrueColor","FalseColor","GrayLevel","ColorRamp"};



void load(const std::string & filename, ApplicationSettings& appSet)
{
   TiXmlDocument doc(filename.c_str());
   if(!doc.LoadFile())
   {
      itkGenericExceptionMacro("Cannot open file "<<filename)
   }
   
   //Create a DOM parser
   TiXmlHandle hDoc(&doc);
   TiXmlElement * pElem;
   pElem = hDoc.FirstChildElement().Element();
   TiXmlHandle hRoot(pElem);
   
   //Get Project Name
   pElem = hRoot.FirstChild("ProjectName").Element();
   appSet.ProjectName = pElem->GetText();
   
   //Get Filename
   pElem = hRoot.FirstChild("FileName").Element();
   appSet.FileName = pElem->GetText();

   //Get nb components
   pElem = hRoot.FirstChild("NbComponents").Element();
   appSet.NbComponents = boost::lexical_cast<unsigned int>(pElem->GetText());
   
   //Get Sensor
   pElem = hRoot.FirstChild("Sensor").Element();
   std::string sensor = pElem->GetText();
   std::string const * wh1 = std::find(boost::begin(StringEnumSensorTab),boost::end(StringEnumSensorTab),sensor);
   if(wh1 == boost::end(StringEnumSensorTab))
   {
      itkGenericExceptionMacro("Sensor Name not recognized !")
   }
   appSet.Sensor = ApplicationSettings::SensorType(std::distance(boost::begin(StringEnumSensorTab),wh1));
   
   //Get Default Channels
   
   std::vector<unsigned int> channels;
   channels.clear();
   for(pElem = hRoot.FirstChild("DefaultChannels").FirstChild().Element();
       pElem;
       pElem=pElem->NextSiblingElement())
   {
      if(pElem->GetText())
      {
         channels.push_back(boost::lexical_cast<unsigned int>(pElem->GetText()));
      }
   }
   if (channels.size() == 5)
   {
   appSet.DefaultChannels = DefaultChannelsSettings(channels[0],channels[1],channels[2],channels[3], channels[4]);
   }

   // Get which spectral band is valid : blue or mir
   pElem = hRoot.FirstChild("BlueAndMIRSpectralBandStatus").FirstChild("BlueStatus").Element();
   appSet.BlueAndMIRStatus.BlueStatus = boost::lexical_cast<bool>(pElem->GetText());

   pElem = hRoot.FirstChild("BlueAndMIRSpectralBandStatus").FirstChild("MIRStatus").Element();
   appSet.BlueAndMIRStatus.MIRStatus = boost::lexical_cast<bool>(pElem->GetText());
   
   //Get List of layers
   appSet.LayersMap.clear();
   
   for(TiXmlElement * pLayer = hRoot.FirstChild("LayersList").FirstChildElement().Element();
       pLayer;
       pLayer=pLayer->NextSiblingElement())
   {
      TiXmlHandle hLayer(pLayer);
      LayerSettings layer;
      std::string name = pLayer->Attribute("name");
      layer.Name = name;
      
      //Get layer type
      pElem = hLayer.FirstChild("Type").Element();
      std::string type = pElem->GetText();
      std::string const * wh2 = std::find(boost::begin(StringEnumTypeTab),boost::end(StringEnumTypeTab),type);
      if(wh2 == boost::end(StringEnumTypeTab))
      {
         itkGenericExceptionMacro("Layer Type not recognized !")
      }
      layer.Type = LayerSettings::LayerType(std::distance(boost::begin(StringEnumTypeTab),wh2));

      if(*wh2 == "KMeans")
      {
         pElem = hLayer.FirstChild("NbClasses").Element();
         layer.NbClasses = boost::lexical_cast<unsigned int>(pElem->GetText());
      } 
 
      //Get Display mode
      pElem = hLayer.FirstChild("DisplayMode").FirstChild().Element();
      std::string mode = pElem->GetText();
      std::string const * wh3 = std::find(boost::begin(StringEnumDisplayTab),boost::end(StringEnumDisplayTab),mode);
      if(wh3 == boost::end(StringEnumDisplayTab))
      {
         itkGenericExceptionMacro("Display Mode not recognized !")
      }
      layer.Mode = LayerSettings::DisplayMode(std::distance(boost::begin(StringEnumDisplayTab),wh3));
      pElem = pElem->NextSiblingElement();
      if (pElem->GetText())
      {
         layer.GrayBand = boost::lexical_cast<unsigned int>(pElem->GetText());
      }

      // Get the description
      if ( mode == "GrayLevel")
        {
        pElem = hLayer.FirstChild("Description").Element();
        if (pElem && pElem->GetText())
          {
          std::string description(pElem->GetText());
          layer.Desc = description;
          }
        }
      
      //Get Quantiles
      pElem = hLayer.FirstChild("ContrastEnhancement").FirstChild("Quantiles").FirstChild().Element();
      layer.MinQuantile = boost::lexical_cast<double>(pElem->GetText());
      pElem = pElem->NextSiblingElement();
      layer.MaxQuantile = boost::lexical_cast<double>(pElem->GetText());
      
      //Get Min Max values for Contrast Enhancement
      
      for(pElem = hLayer.FirstChild("ContrastEnhancement").FirstChild("MinMaxValues").FirstChild().Element();
          pElem;
          pElem=pElem->NextSiblingElement())
      {
         const char * min = pElem->FirstChildElement("Min")->GetText();
         const char * max = pElem->FirstChildElement("Max")->GetText();
         if(min && max)
         {
            layer.MinMaxHisto.push_back(boost::lexical_cast<double>(min));
            layer.MinMaxHisto.push_back(boost::lexical_cast<double>(max));
         }
         else
         {
            layer.MinMaxHisto.push_back(0.);
         }
      }
      appSet.LayersMap[name] = layer;
   }
}

void save(const std::string & filename, const ApplicationSettings& appSet)
{
   TiXmlDocument doc;
   TiXmlElement * root = new TiXmlElement("AppSettings");
   doc.LinkEndChild(root);
   
   //Set Project Name
   TiXmlElement * pName = new TiXmlElement("ProjectName");
   root->LinkEndChild(pName);
   pName->LinkEndChild(new TiXmlText(appSet.ProjectName.c_str()));
   
   //Set FileName
   TiXmlElement * pFileName = new TiXmlElement("FileName");
   root->LinkEndChild(pFileName);
   const std::string & imgFileName = itksys::SystemTools::CollapseFullPath(appSet.FileName.c_str());
   pFileName->LinkEndChild(new TiXmlText(imgFileName.c_str()));

   //Set FileName
   TiXmlElement * pNbComp = new TiXmlElement("NbComponents");
   root->LinkEndChild(pNbComp);
   pNbComp->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>(appSet.NbComponents).c_str()));
   
   //Set Sensor
   TiXmlElement * pSensor = new TiXmlElement("Sensor");
   root->LinkEndChild(pSensor);
   pSensor->LinkEndChild(new TiXmlText(StringEnumSensorTab[appSet.Sensor].c_str()));
   
   //Set Default Color Setting
   TiXmlElement * pChannels = new TiXmlElement("DefaultChannels");
   root->LinkEndChild(pChannels);
   TiXmlElement * pRedBand = new TiXmlElement("RedBand");
   pRedBand->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>(appSet.DefaultChannels.RedBand).c_str()));
   pChannels->LinkEndChild(pRedBand);
   TiXmlElement * pGreenBand = new TiXmlElement("GreenBand");
   pGreenBand->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>(appSet.DefaultChannels.GreenBand).c_str()));
   pChannels->LinkEndChild(pGreenBand);
   TiXmlElement * pBlueBand = new TiXmlElement("BlueBand");
   pBlueBand->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>(appSet.DefaultChannels.BlueBand).c_str()));
   pChannels->LinkEndChild(pBlueBand);
   TiXmlElement * pNIRBand = new TiXmlElement("NIRBand");
   pNIRBand->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>(appSet.DefaultChannels.NIRBand).c_str()));
   pChannels->LinkEndChild(pNIRBand);
   TiXmlElement * pMIRBand = new TiXmlElement("MIRBand");
   pMIRBand->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>(appSet.DefaultChannels.MIRBand).c_str()));
   pChannels->LinkEndChild(pMIRBand);

   // Save the blue and mir spectral band status
   TiXmlElement * pBAndMIRStatus = new TiXmlElement("BlueAndMIRSpectralBandStatus");
   root->LinkEndChild(pBAndMIRStatus);
   //--- Blue status
   TiXmlElement * pBlueStatus = new TiXmlElement("BlueStatus");
   pBlueStatus->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>(appSet.BlueAndMIRStatus.BlueStatus).c_str()));
   pBAndMIRStatus->LinkEndChild(pBlueStatus);
   //--- MIR status
   TiXmlElement * pMIRStatus = new TiXmlElement("MIRStatus");
   pMIRStatus->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>(appSet.BlueAndMIRStatus.MIRStatus).c_str()));
   pBAndMIRStatus->LinkEndChild(pMIRStatus);
   
   //Set Layers list
   TiXmlElement * pLayers = new TiXmlElement("LayersList");
   root->LinkEndChild(pLayers);
   std::map<std::string, LayerSettings>::const_iterator itLayer;
   for(itLayer = appSet.LayersMap.begin(); itLayer != appSet.LayersMap.end(); itLayer++ )
   {
      TiXmlElement * pLayer = new TiXmlElement("Layer");
      pLayers->LinkEndChild(pLayer);
      pLayer->SetAttribute("name",(*itLayer).second.Name.c_str());
      
      //Set Layer Type
      TiXmlElement * pType = new TiXmlElement("Type");
      pLayer->LinkEndChild(pType);
      pType->LinkEndChild(new TiXmlText(StringEnumTypeTab[(*itLayer).second.Type].c_str()));
      
      if ((*itLayer).second.Type == LayerSettings::KMeans)
      {
         TiXmlElement * pNbClasses = new TiXmlElement("NbClasses");
         pLayer->LinkEndChild(pNbClasses);
         pNbClasses->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>((*itLayer).second.NbClasses).c_str()));
      }
      
      //Display Mode
      TiXmlElement * pDisplay = new TiXmlElement("DisplayMode");
      pLayer->LinkEndChild(pDisplay);
      TiXmlElement * pMode = new TiXmlElement("Mode");
      pDisplay->LinkEndChild(pMode);
      pMode->LinkEndChild(new TiXmlText(StringEnumDisplayTab[(*itLayer).second.Mode].c_str()));
      TiXmlElement * pGrayBand = new TiXmlElement("GrayBand");
      pDisplay->LinkEndChild(pGrayBand);
      pGrayBand->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>((*itLayer).second.GrayBand).c_str()));
      TiXmlElement * pDescGrayBand = new TiXmlElement("Description");
      pLayer->LinkEndChild(pDescGrayBand);
      pDescGrayBand->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>((*itLayer).second.Desc).c_str()));
      
      //Contrast enhancement
      TiXmlElement * pContrast = new TiXmlElement("ContrastEnhancement");
      pLayer->LinkEndChild(pContrast);
      //Quantiles
      TiXmlElement * pQuantiles = new TiXmlElement("Quantiles");
      pContrast->LinkEndChild(pQuantiles);
      //minQuantile
      TiXmlElement * pMinQuant = new TiXmlElement("MinQuantile");
      pQuantiles->LinkEndChild(pMinQuant);
      pMinQuant->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>((*itLayer).second.MinQuantile).c_str()));
      //maxQuantile
      TiXmlElement * pMaxQuant = new TiXmlElement("MaxQuantile");
      pQuantiles->LinkEndChild(pMaxQuant);
      pMaxQuant->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>((*itLayer).second.MaxQuantile).c_str()));
      
      
      //MinMaxValues
      TiXmlElement * pMinMaxValues = new TiXmlElement("MinMaxValues");
      pContrast->LinkEndChild(pMinMaxValues);
      
      const unsigned int nbBands = (*itLayer).second.MinMaxHisto.size() / 2;
      for(unsigned int b = 0; b<nbBands; b++)
      {
         TiXmlElement * pMinMax = new TiXmlElement("MinMax");
         pMinMaxValues->LinkEndChild(pMinMax);
         
         //Min
         TiXmlElement * pMin = new TiXmlElement("Min");
         pMinMax->LinkEndChild(pMin);
         pMin->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>((*itLayer).second.MinMaxHisto[2*b]).c_str()));
         
         //Max
         TiXmlElement * pMax = new TiXmlElement("Max");
         pMinMax->LinkEndChild(pMax);
         pMax->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>((*itLayer).second.MinMaxHisto[2*b+1]).c_str()));
      }
   }
   
   //write file
   doc.SaveFile(filename.c_str());
}

void exportQGIS(const std::string & filename, const std::string & filename_utf8, 
                ApplicationSettings& appSet)
{
   //<!DOCTYPE qgis PUBLIC 'http://mrcc.com/qgis.dtd' 'SYSTEM'>
   TiXmlDocument doc;
   TiXmlUnknown * pHeader = new TiXmlUnknown();
   pHeader->SetValue("!DOCTYPE qgis PUBLIC 'http://mrcc.com/qgis.dtd' 'SYSTEM'");
   doc.LinkEndChild(pHeader);
   
   //<qgis projectname="" version="1.7.4-Wroclaw">
   TiXmlElement * root = new TiXmlElement("qgis");
   root->SetAttribute("projectname", appSet.ProjectName.c_str());
   root->SetAttribute("version", "1.7.4-Wroclaw");
   doc.LinkEndChild(root);

   ReaderType::Pointer reader = ReaderType::New();
   reader->SetFileName(appSet.FileName);
   reader->UpdateOutputInformation();
   
   const char * srid = std::string("").c_str();
   const char * n = std::string("").c_str();
   std::string geoFlag("false");
   std::string proj4 = "";
   std::string projAcro = "";
   std::string ellAcro = "";
   const char * desc = std::string("").c_str();
   std::string authid = "";
   
   std::string projectionRefWkt = reader->GetOutput()->GetProjectionRef();
   bool projectionInformationAvailable = !projectionRefWkt.empty();
   OGRSpatialReference * oSRS = NULL;

   if(projectionInformationAvailable)
   {
      oSRS = reinterpret_cast<OGRSpatialReference *>(OSRNewSpatialReference(projectionRefWkt.c_str()));
      
      //srid
      const char * tempSrid = oSRS->GetAuthorityCode(NULL);
      
      //authid
       const char * tempN = oSRS->GetAuthorityName(NULL);

      if (tempSrid != NULL &&  tempN != NULL)
        {
        srid = tempSrid;
        n    = tempN;
        authid = std::string(n)+":"+std::string(srid);
        }

      //Geographic flag
      if(oSRS->IsGeographic())
      {
         geoFlag = "true";
      }

      //proj4
      char *proj4src = NULL;
      oSRS->exportToProj4(&proj4src);
      if (proj4src != NULL)
        {
        proj4 = proj4src;

        //projection acronym
        size_t posProj = proj4.find("+proj=");
        
        if (posProj != std::string::npos)
          {
          unsigned int len = proj4.find(" +", posProj) - posProj;

          if (len > 6)
            {
            projAcro = proj4.substr(posProj+6,len-6);
            }
          }

        //ellipsoid acronym
        size_t posDatum = proj4.find("+datum=");
        if (posDatum != std::string::npos)
          {
          unsigned int len = proj4.find(" +", posDatum) - posDatum;
          if (len > 7)
            {
            ellAcro = proj4.substr(posDatum+7,len-7);
            }
          }

        CPLFree(proj4src);
        }

      //description
      std::string attr("PROJCS");
      const char * tempDesc = oSRS->GetAttrValue(attr.c_str(),0);

      if (tempDesc != NULL)
       {
        desc = tempDesc;
       }
   }
   /*else
   {
      itkGenericExceptionMacro("L'image " << appSet.FileName << " n'est dans aucun système de projection.");
   }*/
   
   
   //<legend>
   TiXmlElement * pLegend = new TiXmlElement("legend");
   root->LinkEndChild(pLegend);
   
   //<projectlayers layercount="2">
   TiXmlElement * pProjLayers = new TiXmlElement("projectlayers");
   root->LinkEndChild(pProjLayers);
   pProjLayers->SetAttribute("layercount",appSet.LayersMap.size());
   
   std::map<std::string, LayerSettings>::const_iterator itLayer;
   for(itLayer = appSet.LayersMap.begin(); itLayer != appSet.LayersMap.end(); itLayer++ )
   {
         const LayerSettings & layer = (*itLayer).second;
          
         // layer id!
         time_t rawtime;
         struct tm * ptm;
         time ( &rawtime );
         ptm = gmtime ( &rawtime );
  
         //layer id !
         std::stringstream ss;
         ss<<1900 + ptm->tm_year << ptm->tm_mon << ptm->tm_mday << ptm->tm_hour << ptm->tm_min << ptm->tm_sec;
         std::string layerId = layer.Name + ss.str();
         
         //<legendlayer open="true" checked="Qt::Checked" name="QB_Toulouse_Ortho_PAN" showFeatureCount="0">
         TiXmlElement * pLegendLayer = new TiXmlElement("legendlayer");
         pLegend->LinkEndChild(pLegendLayer);
         pLegendLayer->SetAttribute("open","true");
         pLegendLayer->SetAttribute("checked","Qt::Checked");
         pLegendLayer->SetAttribute("name",layer.Name.c_str());
         pLegendLayer->SetAttribute("showFeatureCount","0");
         //<filegroup open="true" hidden="false">
         TiXmlElement * pFileGroup = new TiXmlElement("filegroup");
         pLegendLayer->LinkEndChild(pFileGroup);
         pFileGroup->SetAttribute("open","true");
         pFileGroup->SetAttribute("hidden","false");
         //<legendlayerfile isInOverview="0" layerid="QB_Toulouse_Ortho_PAN20120615150451" visible="1"/>
         TiXmlElement * pLegendlayerfile = new TiXmlElement("legendlayerfile");
         pFileGroup->LinkEndChild(pLegendlayerfile);
         pLegendlayerfile->SetAttribute("isInOverview",0);
         pLegendlayerfile->SetAttribute("layerid",layerId.c_str());
         pLegendlayerfile->SetAttribute("visible",1);
         
         
         //<maplayer minimumScale="0" maximumScale="1e+08" type="raster" hasScaleBasedVisibilityFlag="0">
         TiXmlElement * pMapLayer = new TiXmlElement("maplayer");
         pProjLayers->LinkEndChild(pMapLayer);
         pMapLayer->SetAttribute("type","raster");
         
         //<id>QB_Toulouse_Ortho_PAN20120615120435202</id>
         TiXmlElement * pId = new TiXmlElement("id");
         pMapLayer->LinkEndChild(pId);
         pId->LinkEndChild(new TiXmlText(layerId.c_str()));
         
         //<datasource>Z:/ORFEO-TOOLBOX/otb-hg/OTB-Data/Input/QB_Toulouse_Ortho_PAN.tif</datasource>
         //launch process and write output image.
         std::string basePath("");
         basePath = itksys::SystemTools::CollapseFullPath(filename.c_str());
         basePath = itksys::SystemTools::GetFilenamePath(basePath);
         const std::string & imgFileName = itksys::SystemTools::CollapseFullPath(appSet.FileName.c_str());
         
         // for utf8 chars
         std::string basePath_utf8("");
         basePath_utf8 = itksys::SystemTools::CollapseFullPath(filename_utf8.c_str());
         basePath_utf8 = itksys::SystemTools::GetFilenamePath(basePath_utf8);
         const std::string & imgFileName_utf8 = itksys::SystemTools::CollapseFullPath(appSet.FileName_utf8.c_str());
         
         // launch the process 
         std::string filepath_utf8 = launchProcessing(basePath, appSet, layer.Name, imgFileName, 
                                                      basePath_utf8, imgFileName_utf8);
         
         TiXmlElement * pDataSource = new TiXmlElement("datasource");
         pMapLayer->LinkEndChild(pDataSource);
         pDataSource->LinkEndChild(new TiXmlText(filepath_utf8.c_str()));
         
         //<layername>QB_Toulouse_Ortho_PAN</layername>
         TiXmlElement * pLayerName = new TiXmlElement("layername");
         pMapLayer->LinkEndChild(pLayerName);
         pLayerName->LinkEndChild(new TiXmlText(layer.Name.c_str()));
         
         //<srs>
         TiXmlElement * pSRS = new TiXmlElement("srs");
         pMapLayer->LinkEndChild(pSRS);
         //<spatialrefsys>
         TiXmlElement * pSpatialRef = new TiXmlElement("spatialrefsys");
         pSRS->LinkEndChild(pSpatialRef);
         //<proj4>+proj=utm +zone=31 +datum=WGS84 +units=m +no_defs</proj4>
         TiXmlElement * pProj4 = new TiXmlElement("proj4");
         pSpatialRef->LinkEndChild(pProj4);
         pProj4->LinkEndChild(new TiXmlText(proj4.c_str()));
         //<srid>32631</srid>
         TiXmlElement * pSRId = new TiXmlElement("srid");
         pSpatialRef->LinkEndChild(pSRId);
         pSRId->LinkEndChild(new TiXmlText(srid));
         //<authid>EPSG:32631</authid>
         TiXmlElement * pAuthId = new TiXmlElement("authid");
         pSpatialRef->LinkEndChild(pAuthId);
         pAuthId->LinkEndChild(new TiXmlText(authid.c_str()));
         //<description>WGS 84 / UTM zone 31N</description>
         TiXmlElement * pDesc = new TiXmlElement("description");
         pSpatialRef->LinkEndChild(pDesc);
         pDesc->LinkEndChild(new TiXmlText(desc));
         //<projectionacronym>utm</projectionacronym>
         TiXmlElement * pProjAcr = new TiXmlElement("projectionacronym");
         pSpatialRef->LinkEndChild(pProjAcr);
         pProjAcr->LinkEndChild(new TiXmlText(projAcro.c_str()));
         //<ellipsoidacronym>WGS84</ellipsoidacronym>
         TiXmlElement * pEllAcro = new TiXmlElement("ellipsoidacronym");
         pSpatialRef->LinkEndChild(pEllAcro);
         pEllAcro->LinkEndChild(new TiXmlText(ellAcro.c_str()));
         //<geographicflag>false</geographicflag>
         TiXmlElement * pGeoFlag = new TiXmlElement("geographicflag");
         pSpatialRef->LinkEndChild(pGeoFlag);
         pGeoFlag->LinkEndChild(new TiXmlText(geoFlag.c_str()));
         //<provider>gdal</provider>
         TiXmlElement * pProv = new TiXmlElement("provider");
         pMapLayer->LinkEndChild(pProv);
         pProv->LinkEndChild(new TiXmlText("gdal"));
         
         
         //<rasterproperties>
         TiXmlElement * pRasterProp = new TiXmlElement("rasterproperties");
         pMapLayer->LinkEndChild(pRasterProp);
         
         if (layer.Mode == LayerSettings::GrayLevel) //Gray level mode
         {
            //<mDrawingStyle>SingleBandGray</mDrawingStyle>
            TiXmlElement * pDrawingStyle = new TiXmlElement("mDrawingStyle");
            pRasterProp->LinkEndChild(pDrawingStyle);
            pDrawingStyle->LinkEndChild(new TiXmlText("SingleBandGray"));
            
            //<mGrayBandName>Bande 1</mGrayBandName>
            TiXmlElement * pGrayBand = new TiXmlElement("mGrayBandName");
            pRasterProp->LinkEndChild(pGrayBand);
            std::string grayBand = "Bande "+boost::lexical_cast<std::string>(layer.GrayBand + 1);
            pGrayBand->LinkEndChild(new TiXmlText(grayBand.c_str()));
            
            //<mUserDefinedGrayMinimumMaximum boolean="false"/>
            TiXmlElement * pUserGray = new TiXmlElement("mUserDefinedGrayMinimumMaximum");
            pRasterProp->LinkEndChild(pUserGray);
            pUserGray->SetAttribute("boolean","true");
         }
         else if (layer.Mode == LayerSettings::ColorRamp)
         {
            //<mDrawingStyle>SingleBandGray</mDrawingStyle>
            TiXmlElement * pDrawingStyle = new TiXmlElement("mDrawingStyle");
            pRasterProp->LinkEndChild(pDrawingStyle);
            pDrawingStyle->LinkEndChild(new TiXmlText("SingleBandPseudoColor"));
            
            //<mColorShadingAlgorithm>ColorRampShader</mColorShadingAlgorithm>
            TiXmlElement * pColorShading = new TiXmlElement("mColorShadingAlgorithm");
            pRasterProp->LinkEndChild(pColorShading);
            pColorShading->LinkEndChild(new TiXmlText("ColorRampShader"));
            
            //<mGrayBandName>Bande 1</mGrayBandName>
            TiXmlElement * pGrayBand = new TiXmlElement("mGrayBandName");
            pRasterProp->LinkEndChild(pGrayBand);
            std::string grayBand = "Bande "+boost::lexical_cast<std::string>(layer.GrayBand + 1);
            pGrayBand->LinkEndChild(new TiXmlText(grayBand.c_str()));
            
            //<mUserDefinedGrayMinimumMaximum boolean="false"/>
            TiXmlElement * pUserGray = new TiXmlElement("mUserDefinedGrayMinimumMaximum");
            pRasterProp->LinkEndChild(pUserGray);
            pUserGray->SetAttribute("boolean","false");
            
         }
         else
         {
            //<mDrawingStyle>MultiBandColor</mDrawingStyle>
            TiXmlElement * pDrawingStyle = new TiXmlElement("mDrawingStyle");
            pRasterProp->LinkEndChild(pDrawingStyle);
            pDrawingStyle->LinkEndChild(new TiXmlText("MultiBandColor"));
            
            //<mRedBandName>Bande 1</mRedBandName>
            TiXmlElement * pRedBand = new TiXmlElement("mRedBandName");
            pRasterProp->LinkEndChild(pRedBand);
            std::string redBand = "Bande "+boost::lexical_cast<std::string>(layer.RedBand + 1);
            pRedBand->LinkEndChild(new TiXmlText(redBand.c_str()));
            
            //<mGreenBandName>Bande 1</mGreenBandName>
            TiXmlElement * pGreenBand = new TiXmlElement("mGreenBandName");
            pRasterProp->LinkEndChild(pGreenBand);
            std::string greenBand = "Bande "+boost::lexical_cast<std::string>(layer.GreenBand + 1);
            pGreenBand->LinkEndChild(new TiXmlText(greenBand.c_str()));
            
            //<mBlueBandName>Bande 1</mBlueBandName>
            TiXmlElement * pBlueBand = new TiXmlElement("mBlueBandName");
            pRasterProp->LinkEndChild(pBlueBand);
            std::string blueBand = "Bande "+boost::lexical_cast<std::string>(layer.BlueBand + 1);
            pBlueBand->LinkEndChild(new TiXmlText(blueBand.c_str()));
            
            //<mUserDefinedRGBMinimumMaximum boolean="false"/>
            TiXmlElement * pUserRGB = new TiXmlElement("mUserDefinedRGBMinimumMaximum");
            pRasterProp->LinkEndChild(pUserRGB);
            pUserRGB->SetAttribute("boolean","true");
         }
         
         //<mContrastEnhancementAlgorithm>StretchToMinimumMaximum</mContrastEnhancementAlgorithm>
         TiXmlElement * pContrastEnhancement = new TiXmlElement("mContrastEnhancementAlgorithm");
         pRasterProp->LinkEndChild(pContrastEnhancement);
         pContrastEnhancement->LinkEndChild(new TiXmlText("StretchToMinimumMaximum"));
         
         //<contrastEnhancementMinMaxValues>
         TiXmlElement * pContrastValues = new TiXmlElement("contrastEnhancementMinMaxValues");
         pRasterProp->LinkEndChild(pContrastValues);

         const unsigned int nbBands = layer.MinMaxHisto.size() / 2;
         for(unsigned int b = 0; b<nbBands; b++)
         {
            //<minMaxEntry>
            TiXmlElement * pMinMax = new TiXmlElement("minMaxEntry");
            pContrastValues->LinkEndChild(pMinMax);
            //<min>-32768</min>
            TiXmlElement * pMin = new TiXmlElement("min");
            pMinMax->LinkEndChild(pMin);
            pMin->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>(layer.MinMaxHisto[2*b]).c_str()));
 
            //<max>32767</max>
            TiXmlElement * pMax = new TiXmlElement("max");
            pMinMax->LinkEndChild(pMax);
            pMax->LinkEndChild(new TiXmlText(boost::lexical_cast<std::string>(layer.MinMaxHisto[2*b+1]).c_str()));
         }
         
         //Write the color Ramp if any
         if (layer.Mode == LayerSettings::ColorRamp)
         {
            //<customColorRamp>
            TiXmlElement * pCustomRamp = new TiXmlElement("customColorRamp");
            pRasterProp->LinkEndChild(pCustomRamp);
            //<colorRampType>INTERPOLATED</colorRampType>
            TiXmlElement * pRampType = new TiXmlElement("colorRampType");
            pCustomRamp->LinkEndChild(pRampType);
            pRampType->LinkEndChild(new TiXmlText("INTERPOLATED"));
            
            
            std::map<unsigned int, std::vector<unsigned int> >::const_iterator  rampIt;
            for(rampIt = layer.Ramp.begin(); rampIt != layer.Ramp.end(); rampIt++ )
            {
               //<colorRampEntry red="0" blue="255" value="0.000000" green="0" label=""/>
               TiXmlElement * pRampEntry = new TiXmlElement("colorRampEntry");
               pCustomRamp->LinkEndChild(pRampEntry);
               pRampEntry->SetAttribute("red",boost::lexical_cast<std::string>( (*rampIt).second[0] ).c_str());
               pRampEntry->SetAttribute("blue",boost::lexical_cast<std::string>( (*rampIt).second[2] ).c_str());
               pRampEntry->SetAttribute("value",boost::lexical_cast<std::string>( (*rampIt).first ).c_str());
               pRampEntry->SetAttribute("green",boost::lexical_cast<std::string>( (*rampIt).second[1] ).c_str());
               pRampEntry->SetAttribute("label","");
            }
         }
   } //end for
   
   //write file
   doc.SaveFile(filename.c_str());
}


const std::string launchProcessing(const std::string & path, ApplicationSettings & appSet, const std::string & layerId, const std::string & imgfilename, const std::string & path_utf8, const std::string& imgfilename_utf8)
{
   const LayerSettings & layer = appSet.LayersMap[layerId];
   const LayerSettings::LayerType & type = layer.Type;
   ReaderType::Pointer reader = ReaderType::New();
   
   reader->SetFileName(imgfilename);
   reader->UpdateOutputInformation();
   
   std::string outputFileName = path + "/";
   std::string outputFileName_utf8 = path_utf8  + "/";

   switch (type)
   {
      case LayerSettings::ColorComposite:
      {
         outputFileName = imgfilename;
         outputFileName_utf8 = imgfilename_utf8;
         break;
      }
      case LayerSettings::NDVI:
      {
         ImageWriterType::Pointer writer = ImageWriterType::New();
         NDVIFilterType::Pointer ndvi = NDVIFilterType::New();
         ndvi->SetInput(reader->GetOutput());
         ndvi->SetRedIndex(appSet.DefaultChannels.RedBand + 1);
         ndvi->SetNIRIndex(appSet.DefaultChannels.NIRBand + 1);
         outputFileName = outputFileName + layer.Name + ".tif";
         outputFileName_utf8 =  outputFileName_utf8 + layer.Name + ".tif";
         writer->SetInput(ndvi->GetOutput());
         writer->SetFileName(outputFileName);
         writer->Update();
         
         break;
      }
      case LayerSettings::KMeans:
      {
         const unsigned int nbClasses = layer.NbClasses;
         
         // Training sample lists
         ListSampleType::Pointer sampleList = ListSampleType::New();
         
         // Sample dimension and max dimension
         const unsigned int nbComp =  reader->GetOutput()->GetNumberOfComponentsPerPixel();
         unsigned int sampleSize = nbComp;
         unsigned int totalSamples = 0;
         
         EstimatorType::ParametersType initialMeans(nbComp * nbClasses);
         initialMeans.Fill(0);
         
         // use image and mask shrink
         double theoricNBSamplesForKMeans = 10000;
         const double upperThresholdNBSamplesForKMeans = 1000 * 1000;
         const double actualNBSamplesForKMeans = std::min(theoricNBSamplesForKMeans, upperThresholdNBSamplesForKMeans);
         
         const double shrinkFactor = vcl_floor(
            vcl_sqrt(
               reader->GetOutput()->GetLargestPossibleRegion().GetNumberOfPixels()
               / actualNBSamplesForKMeans));
         
         ImageSamplingFilterType::Pointer imageSampler = ImageSamplingFilterType::New();
         imageSampler->SetInput(reader->GetOutput());
         imageSampler->SetShrinkFactor(shrinkFactor);
         imageSampler->Update();
         
         // Then, build the sample list
         IteratorType it(imageSampler->GetOutput(), 
                           imageSampler->GetOutput()->GetLargestPossibleRegion());
         
         it.GoToBegin();
         
         VectorImageType::PixelType min;
         VectorImageType::PixelType max;
         VectorImageType::PixelType sample;
         //first sample
         
         itk::Statistics::MersenneTwisterRandomVariateGenerator::Pointer 
            randGen =itk::Statistics::MersenneTwisterRandomVariateGenerator::GetInstance();
         
         //randGen->Initialize();
         
         min = it.Get();
         max = it.Get();
         sample = it.Get();
         sampleList->PushBack(sample);
         ++it;
         
         totalSamples = 1;
         while (!it.IsAtEnd())
            {
            totalSamples++;
            sample = it.Get();
            sampleList->PushBack(sample);
               
            for (unsigned int i = 0; i < nbComp; ++i)
               {
               if (min[i] > sample[i])
               {
               min[i] = sample[i];
               }
               if (max[i] < sample[i])
               {
               max[i] = sample[i];
               }
               }
            ++it;
            }
         
         // Next, initialize centroids by random sampling in the generated
         // list of samples
         for (unsigned int classIndex = 0; classIndex < nbClasses; ++classIndex)
            {
            VectorImageType::PixelType newCentroid = sampleList->GetMeasurementVector(randGen->GetIntegerVariate(sampleList->Size()-1));
         
            for (unsigned int compIndex = 0; compIndex < sampleSize; ++compIndex)
               {
               initialMeans[compIndex + classIndex * sampleSize] = newCentroid[compIndex];
               }
            }
         
         /*******************************************/
         /*           Learning                      */
         /*******************************************/
         EstimatorType::Pointer estimator = EstimatorType::New();
         
         TreeGeneratorType::Pointer treeGenerator = TreeGeneratorType::New();
         treeGenerator->SetSample(sampleList);
         treeGenerator->SetBucketSize(10000);
         treeGenerator->Update();
         
         estimator->SetParameters(initialMeans);
         estimator->SetKdTree(treeGenerator->GetOutput());
         estimator->SetMaximumIteration(10000);
         //estimator->SetCentroidPositionChangesThreshold(GetParameterFloat("ct"));
         estimator->StartOptimization();
         
         EstimatorType::ParametersType estimatedMeans = estimator->GetParameters();
         
         /*******************************************/
         /*           Classification                */
         /*******************************************/
         
         // Finally, update the KMeans filter
         KMeansFunctorType functor;
         
         for (unsigned int classIndex = 0; classIndex < nbClasses; ++classIndex)
            {
            VectorImageType::PixelType centroid(sampleSize);
         
            for (unsigned int compIndex = 0; compIndex < sampleSize; ++compIndex)
               {
               centroid[compIndex] = estimatedMeans[compIndex + classIndex * sampleSize];
               }
            functor.AddCentroid(classIndex, centroid);
            }
         
         KMeansFilterType::Pointer kMeansFilter = KMeansFilterType::New();
         kMeansFilter->SetFunctor(functor);
         kMeansFilter->SetInput(reader->GetOutput());
         
         LabelImageWriterType::Pointer writer = LabelImageWriterType::New();
         outputFileName = outputFileName + "KMoyennes_nbClasses_" + boost::lexical_cast<std::string>(layer.NbClasses) + ".tif";
         outputFileName_utf8 = outputFileName_utf8 + "KMoyennes_nbClasses_" + boost::lexical_cast<std::string>(layer.NbClasses) + ".tif";
         writer->SetInput(kMeansFilter->GetOutput());
         writer->SetFileName(outputFileName);
         writer->Update();
         
         break;
      }
      case LayerSettings::Brightness:
         IBFilterType::Pointer brightness = IBFilterType::New();
         brightness->SetInput(reader->GetOutput());
         brightness->SetRedIndex(appSet.DefaultChannels.RedBand + 1);
         brightness->SetGreenIndex(appSet.DefaultChannels.GreenBand + 1);
         ImageWriterType::Pointer writer = ImageWriterType::New();
         outputFileName = outputFileName + layer.Name + ".tif";
         outputFileName_utf8 = outputFileName_utf8 + layer.Name + ".tif";
         writer->SetInput(brightness->GetOutput());
         writer->SetFileName(outputFileName);
         writer->Update();
         break;
      
   }
   
   return outputFileName_utf8;
}


void exportKMZ(const std::string & kmzFilename, ApplicationSettings& appSet)
{
   
   
   ReaderListType::Pointer readerList = ReaderListType::New();
   
   KmzProductWriterType::Pointer kmzWriter = KmzProductWriterType::New();
   kmzWriter->SetPath(kmzFilename);
   
   std::map<std::string, LayerSettings>::const_iterator itLayer;
   for(itLayer = appSet.LayersMap.begin(); itLayer != appSet.LayersMap.end(); itLayer++ )
   {
         const LayerSettings & layer = (*itLayer).second;
         
         std::string basePath("");
         basePath = itksys::SystemTools::CollapseFullPath(kmzFilename.c_str());
         basePath = itksys::SystemTools::GetFilenamePath(basePath);
         const std::string & imgFileName = itksys::SystemTools::CollapseFullPath(appSet.FileName.c_str());
         std::string filepath = launchProcessing(basePath, appSet, layer.Name, imgFileName, 
                                                 basePath, imgFileName);
         
         ReaderType::Pointer reader = ReaderType::New();
         reader->SetFileName(filepath);
         reader->UpdateOutputInformation();
         readerList->PushBack(reader);
         
         otb::ProductInformations productInfo;
         productInfo.m_Name = layer.Name;
         otb::ProductInformations::CompositionType channels;
         if(layer.Mode == LayerSettings::GrayLevel)
         {
            channels[0] = layer.GrayBand;
            channels[1] = layer.GrayBand;
            channels[2] = layer.GrayBand;
         }
         //TODO: Process ColorMap !
         else if(layer.Mode == LayerSettings::ColorRamp) 
         {
            channels[0] = 0;
            channels[1] = 0;
            channels[2] = 0;
         }
         else
         {
            channels[0] = layer.RedBand;
            channels[1] = layer.GreenBand;
            channels[2] = layer.BlueBand;
         }
         productInfo.m_Composition = channels;
         kmzWriter->AddInput(reader->GetOutput(), productInfo);
   }
   
   
   //kmzWriter->SetDEMDirectory(demPath);
  
   kmzWriter->Update();
   
}

