/*=========================================================================

Program:   ORFEO Toolbox
Language:  C++
Date:      $Date$
Version:   $Revision$


Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


This software is distributed WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/

#ifndef __ApplicationSettings_h
#define __ApplicationSettings_h

#include "otbTinyXML.h"
#include <string>
#include <map>
#include <vector>

#include "TypeDefinitions.h"

//namespace Settings
//{

struct LayerSettings
{
  enum DisplayMode {TrueColors, FalseColor, GrayLevel, ColorRamp};
  enum LayerType {ColorComposite,NDVI,KMeans,Brightness};
  std::string Name;
  LayerType Type;
   DisplayMode Mode;
   unsigned int GrayBand;
   unsigned int RedBand;
   unsigned int GreenBand;
   unsigned int BlueBand;
   //Kmeans nb classes
   unsigned int NbClasses;
   
   //Quantile for histogram
   double MinQuantile;
   double MaxQuantile;

   // description
   std::string        Desc;
   
   //Contains min, max for each band.
   std::vector<double> MinMaxHisto;
   
   //Color ramp. The key of the map is the label value. 
   //The value associated to a key is a std::vector containing
   //the red, green, blue values (in that order) used for representing this label.
   std::map<unsigned int, std::vector<unsigned int> > Ramp;
   
   //Constructor
   LayerSettings():
   Name(""), Type(ColorComposite), Mode(FalseColor), GrayBand(0),
   RedBand(0), GreenBand(0), BlueBand(0), NbClasses(0),
   MinQuantile(0), MaxQuantile(0), Desc("")
   {
      MinMaxHisto.clear();
   }


}; // End of struct LayerSettings

struct DefaultChannelsSettings
{
  int RedBand;
  int GreenBand;
  int BlueBand;
  int NIRBand;
  int MIRBand;
  
   //Constructor
   DefaultChannelsSettings():
     RedBand(-1), GreenBand(-1), BlueBand(-1), NIRBand(-1), MIRBand(-1)
   {}
  DefaultChannelsSettings( int r, int g, int b, int n,  int m):
    RedBand(r), GreenBand(g), BlueBand(b), NIRBand(n), MIRBand(m)
   {}
};


struct BlueAndMIRSpectralBandStatus
{
  bool BlueStatus;
  bool MIRStatus;
   
   //Constructor
  BlueAndMIRSpectralBandStatus(): 
    BlueStatus(true), MIRStatus(false)
  {}
  BlueAndMIRSpectralBandStatus(bool bStatus, bool mirStatuts):
    BlueStatus(bStatus), MIRStatus(mirStatuts)
  {}
};


struct ApplicationSettings
{
   enum SensorType {PLEIADE,SPOT,FORMOSAT,UNKNOWN};
   
   //attributes
   std::string ProjectName;
   std::string FileName;
   std::string FileName_utf8;  
   SensorType Sensor; 
   DefaultChannelsSettings DefaultChannels;
   BlueAndMIRSpectralBandStatus  BlueAndMIRStatus; 
   std::map<std::string, LayerSettings> LayersMap;
   unsigned int NbComponents;
   
   //Constructor
   ApplicationSettings():
     ProjectName(""), FileName(""), Sensor(UNKNOWN), NbComponents(0)
   {}
   ApplicationSettings(const std::string & p, 
                       const std::string & f, 
                       SensorType s, 
                       const DefaultChannelsSettings & c,
                       const BlueAndMIRSpectralBandStatus& bAndMIR):
     ProjectName(p), FileName(f), Sensor(s), DefaultChannels(c), BlueAndMIRStatus(bAndMIR)
   {}


}; // End of struct ApplicationSettings

void save(const std::string &, const ApplicationSettings&);
void load(const std::string &, ApplicationSettings&);
void exportQGIS(const std::string &, const std::string &, ApplicationSettings&);
const std::string launchProcessing(const std::string &, ApplicationSettings &, const std::string &, const std::string &, 
                                   const std::string &, const std::string &);
void exportKMZ(const std::string &, ApplicationSettings&);

//} //end namespace Settings

#endif
