/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.

Some parts of this code are derived from ITK. See ITKCopyright.txt
for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbRGBFromImageValueFilter_h
#define __otbRGBFromImageValueFilter_h

#include "otbPersistentImageFilter.h"
#include "itkNumericTraits.h"
#include "itkArray.h"
#include "itkSimpleDataObjectDecorator.h"
#include "otbPersistentFilterStreamingDecorator.h"


namespace otb
{

/** \class PersistentRGBFromImageValueFilter
 * \brief Compute mean value of each label in an image using the output requested region.
 *
 *  This filter persists its temporary data. It means that if you Update it n times on n different
 * requested regions, the output statistics will be the statitics of the whole set of n regions.
 *
 * To reset the temporary data, one should call the Reset() function.
 *
 * To get the statistics once the regions have been processed via the pipeline, use the Synthetize() method.
 *
 * \sa PersistentImageFilter
 * \ingroup Streamed
 * \ingroup Multithreaded
 * \ingroup MathematicalStatisticsImageFilters
 */
template<class TInputVectorImage, class TLabelImage>
class ITK_EXPORT PersistentRGBFromImageValueFilter :
  public PersistentImageFilter<TInputVectorImage, TInputVectorImage>
{
public:
  /** Standard Self typedef */
  typedef PersistentRGBFromImageValueFilter               Self;
  typedef PersistentImageFilter<TInputVectorImage, TInputVectorImage> Superclass;
  typedef itk::SmartPointer<Self>                         Pointer;
  typedef itk::SmartPointer<const Self>                   ConstPointer;

  /** Method for creation through the object factory. */
  itkNewMacro(Self);

  /** Runtime information support. */
  itkTypeMacro(PersistentRGBFromImageValueFilter, PersistentImageFilter);

  /** Image related typedefs. */
  typedef TInputVectorImage                   VectorImageType;
  typedef typename TInputVectorImage::Pointer InputVectorImagePointer;
  typedef TLabelImage                         LabelImageType;
  typedef typename TLabelImage::Pointer       LabelImagePointer;

  typedef typename VectorImageType::PixelType         VectorPixelType;
  typedef typename LabelImageType::PixelType          LabelPixelType;
  typedef std::map<LabelPixelType, VectorPixelType>   MeanValueMapType;
  
  itkStaticConstMacro(InputImageDimension, unsigned int,
                      TInputVectorImage::ImageDimension);

  /** Image related typedefs. */
  itkStaticConstMacro(ImageDimension, unsigned int,
                      TInputVectorImage::ImageDimension);

  /** Smart Pointer type to a DataObject. */
  typedef typename itk::DataObject::Pointer DataObjectPointer;

  /** Type of DataObjects used for scalar outputs */
  typedef itk::SimpleDataObjectDecorator<MeanValueMapType>  MeanValueMapObjectType;


  virtual void SetInputLabelImage( const LabelImageType *image);
  virtual const LabelImageType * GetInputLabelImage();
  

  itkSetMacro(MinVal, typename VectorImageType::PixelType)
  itkSetMacro(MaxVal, typename VectorImageType::PixelType)
  
  /** Return the computed Mean. */
  MeanValueMapType GetMeanValueMap() const
  {
    return this->GetMeanIntensity()->Get();
  }
  MeanValueMapObjectType* GetMeanIntensity();
  const MeanValueMapObjectType* GetMeanIntensity() const;

  /** Make a DataObject of the correct type to be used as the specified
   * output. */
  virtual DataObjectPointer MakeOutput(unsigned int idx);

  /** Pass the input through unmodified. Do this by Grafting in the
   *  AllocateOutputs method.
   */
  void AllocateOutputs();
  virtual void GenerateOutputInformation();
  void Synthetize(void);
  void Reset(void);

protected:
  PersistentRGBFromImageValueFilter();
  virtual ~PersistentRGBFromImageValueFilter() {}
  void PrintSelf(std::ostream& os, itk::Indent indent) const;

  /** GenerateData. */
  void  GenerateData();

private:
  PersistentRGBFromImageValueFilter(const Self &); //purposely not implemented
  void operator =(const Self&); //purposely not implemented
  
  MeanValueMapType                  m_MeanValueMap;
  std::map<LabelPixelType, unsigned int>    m_WeigthingMap;

  typename VectorImageType::PixelType m_MinVal;
  typename VectorImageType::PixelType m_MaxVal;

}; // end of class PersistentStatisticsImageFilter

/*===========================================================================*/

/** \class RGBFromImageValueFilter
 * \brief This class streams the whole input image through the PersistentRGBFromImageValueFilter.
 *
 * This way, it allows to compute the first order global statistics of this image. It calls the
 * Reset() method of the PersistentStatisticsImageFilter before streaming the image and the
 * Synthetize() method of the PersistentStatisticsImageFilter after having streamed the image
 * to compute the statistics. The accessor on the results are wrapping the accessors of the
 * internal PersistentStatisticsImageFilter.
 *
 * This filter can be used as:
 * \code
 * typedef otb::RGBFromImageValueFilter<ImageType> StatisticsType;
 * StatisticsType::Pointer statistics = StatisticsType::New();
 * statistics->SetInput(reader->GetOutput());
 * statistics->Update();
 * std::cout << statistics-> GetMaximum() << std::endl;
 * std::cout << statistics-> GetMinimum() << std::endl;
 * \endcode
 *
 * \sa PersistentStatisticsImageFilter
 * \sa PersistentImageFilter
 * \sa PersistentFilterStreamingDecorator
 * \sa StreamingImageVirtualWriter
 * \ingroup Streamed
 * \ingroup Multithreaded
 * \ingroup MathematicalStatisticsImageFilters
 */

template<class TInputVectorImage, class TLabelImage>
class ITK_EXPORT RGBFromImageValueFilter :
  public PersistentFilterStreamingDecorator<PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage> >
{
public:
  /** Standard Self typedef */
  typedef RGBFromImageValueFilter Self;
  typedef PersistentFilterStreamingDecorator
  <PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage> > Superclass;
  typedef itk::SmartPointer<Self>       Pointer;
  typedef itk::SmartPointer<const Self> ConstPointer;

  /** Type macro */
  itkNewMacro(Self);

  /** Creation through object factory macro */
  itkTypeMacro(RGBFromImageValueFilter, PersistentFilterStreamingDecorator);

  typedef TInputVectorImage                   VectorImageType;
  typedef TLabelImage                         LabelImageType;
  
  typedef typename Superclass::FilterType::MeanValueMapType          MeanValueMapType;
  typedef typename Superclass::FilterType::MeanValueMapObjectType    MeanValueMapObjectType;


  void SetInput(const VectorImageType * input)
  {
    this->GetFilter()->SetInput(input);
  }
  const VectorImageType * GetInput()
  {
    return this->GetFilter()->GetInput();
  }
  void SetInputLabelImage(const LabelImageType * input)
  {
    this->GetFilter()->SetInputLabelImage(input);
  }
  const LabelImageType * GetInputLabelImage()
  {
    return this->GetFilter()->GetInputLabelImage();
  }

  void SetMinVal(const typename VectorImageType::PixelType minVal)
  {
      this->GetFilter()->SetMinVal(minVal);
  }
  
  void SetMaxVal(const typename VectorImageType::PixelType maxVal)
  {
      this->GetFilter()->SetMaxVal(maxVal);
  }

  /** Return the computed Mean. */
  MeanValueMapType GetMeanValueMap() const
  {
    return this->GetFilter()->GetMeanIntensity()->Get();
  }
  MeanValueMapObjectType* GetMeanIntensity()
  {
    return this->GetFilter()->GetMeanIntensity();
  }
  const MeanValueMapObjectType* GetMeanIntensity() const
  {
    return this->GetFilter()->GetMeanIntensity();
  }



protected:
  /** Constructor */
  RGBFromImageValueFilter() {};
  /** Destructor */
  virtual ~RGBFromImageValueFilter() {}

private:
  RGBFromImageValueFilter(const Self &); //purposely not implemented
  void operator =(const Self&); //purposely not implemented
};

} // end namespace otb

#ifndef OTB_MANUAL_INSTANTIATION
#include "otbRGBFromImageValueFilter.txx"
#endif

#endif
