/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.

Some parts of this code are derived from ITK. See ITKCopyright.txt
for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbRGBFromImageValueFilter_txx
#define __otbRGBFromImageValueFilter_txx
#include "otbRGBFromImageValueFilter.h"

#include "itkImageRegionIterator.h"
#include "itkImageRegionConstIterator.h"
#include "itkNumericTraits.h"
#include "itkProgressReporter.h"
#include "otbMacro.h"


namespace otb
{

template<class TInputVectorImage, class TLabelImage>
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::PersistentRGBFromImageValueFilter()
{
  // first output is a copy of the image, DataObject created by
  // superclass
  //
  // allocate the data objects for the outputs which are
  // just decorators around pixel types
  typename MeanValueMapObjectType::Pointer output
      = static_cast<MeanValueMapObjectType*>(this->MakeOutput(1).GetPointer());
  this->itk::ProcessObject::SetNthOutput(1, output.GetPointer());

  this->Reset();

}

template<class TInputVectorImage, class TLabelImage>
typename itk::DataObject::Pointer
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::MakeOutput(unsigned int output)
{
  switch (output)
    {
    case 0:
      return static_cast<itk::DataObject*>(TInputVectorImage::New().GetPointer());
      break;
    case 1:
      return static_cast<itk::DataObject*>(MeanValueMapObjectType::New().GetPointer());
      break;
    default:
      // might as well make an image
      return static_cast<itk::DataObject*>(TInputVectorImage::New().GetPointer());
      break;
    }
}

template<class TInputVectorImage, class TLabelImage>
void
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::SetInputLabelImage(const LabelImageType *input)
{
  // Process object is not const-correct so the const_cast is required here
  this->itk::ProcessObject::SetNthInput(1,
                                   const_cast< LabelImageType * >( input ) );

}

template<class TInputVectorImage, class TLabelImage>
const typename PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>::LabelImageType*
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::GetInputLabelImage()
{
  return static_cast< const TLabelImage * >
    (this->itk::ProcessObject::GetInput(1));
}

template<class TInputVectorImage, class TLabelImage>
typename PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>::MeanValueMapObjectType*
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::GetMeanIntensity()
{
  return static_cast<MeanValueMapObjectType*>(this->itk::ProcessObject::GetOutput(1));
}

template<class TInputVectorImage, class TLabelImage>
const typename PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>::MeanValueMapObjectType*
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::GetMeanIntensity() const
{
  return static_cast<const MeanValueMapObjectType*>(this->itk::ProcessObject::GetOutput(1));
}


template<class TInputVectorImage, class TLabelImage>
void
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::GenerateOutputInformation()
{
  Superclass::GenerateOutputInformation();
  
  if (this->GetInput())
    {
    this->GetOutput()->CopyInformation(this->GetInput());
    this->GetOutput()->SetLargestPossibleRegion(this->GetInput()->GetLargestPossibleRegion());

    if (this->GetOutput()->GetRequestedRegion().GetNumberOfPixels() == 0)
      {
      this->GetOutput()->SetRequestedRegion(this->GetOutput()->GetLargestPossibleRegion());
      }
    }
}
template<class TInputVectorImage, class TLabelImage>
void
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::AllocateOutputs()
{
  // This is commented to prevent the streaming of the whole image for the first stream strip
  // It shall not cause any problem because the output image of this filter is not intended to be used.
  //InputImagePointer image = const_cast< TInputImage * >( this->GetInput() );
  //this->GraftOutput( image );
  // Nothing that needs to be allocated for the remaining outputs
}

template<class TInputVectorImage, class TLabelImage>
void
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::Synthetize()
{
   typename MeanValueMapType::iterator it;
   for(it = m_MeanValueMap.begin(); it != m_MeanValueMap.end(); it++)
   {
      it->second = it->second / m_WeigthingMap[it->first];
   }
   this->GetMeanIntensity()->Set(m_MeanValueMap);
}

template<class TInputVectorImage, class TLabelImage>
void
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::Reset()
{
 /* int numberOfThreads = this->GetNumberOfThreads();

  // Resize the thread temporaries
  m_Count.SetSize(numberOfThreads);
  m_SumOfSquares.SetSize(numberOfThreads);
  m_ThreadSum.SetSize(numberOfThreads);
  m_ThreadMin.SetSize(numberOfThreads);
  m_ThreadMax.SetSize(numberOfThreads);

  // Initialize the temporaries
  m_Count.Fill(itk::NumericTraits<long>::Zero);
  m_ThreadSum.Fill(itk::NumericTraits<RealType>::Zero);
  m_SumOfSquares.Fill(itk::NumericTraits<RealType>::Zero);
  m_ThreadMin.Fill(itk::NumericTraits<PixelType>::max());
  m_ThreadMax.Fill(itk::NumericTraits<PixelType>::NonpositiveMin());*/
}

template<class TInputVectorImage, class TLabelImage>
void
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::GenerateData()
{
  /**
   * Grab the input
   */
  InputVectorImagePointer inputPtr =  const_cast<TInputVectorImage *>(this->GetInput());
  LabelImagePointer labelInputPtr =  const_cast<TLabelImage *>(this->GetInputLabelImage());
  
  itk::ImageRegionConstIterator<TInputVectorImage> inIt(inputPtr, inputPtr->GetRequestedRegion());
  itk::ImageRegionConstIterator<TLabelImage> labelIt(labelInputPtr, labelInputPtr->GetRequestedRegion());
  
  inIt.GoToBegin();
  labelIt.GoToBegin();
  typename VectorImageType::PixelType value;
  typename LabelImageType::PixelType label;
  // do the work
  while (!inIt.IsAtEnd() && !labelIt.IsAtEnd())
    {
      value = inIt.Get();
      label = labelIt.Get();
      if (m_MeanValueMap.count(label)<=0) //add new element to the map
      {
         typename VectorImageType::PixelType newValue;
         newValue.SetSize(value.Size());
         for (unsigned int index = 0; index < value.Size(); index++)
         {
            if (value[index] < m_MinVal[index])
               newValue[index] = m_MinVal[index];
            else if (value[index] > m_MaxVal[index])
               newValue[index] = m_MaxVal[index];
            else newValue[index] = value[index];
         }
         m_MeanValueMap[label] = newValue;
         m_WeigthingMap[label] = 1;
      }
      else
      {
         typename VectorImageType::PixelType currentValue = m_MeanValueMap[label];
         m_WeigthingMap[label] = m_WeigthingMap[label] + 1;
         for (unsigned int index = 0; index < value.Size(); index++)
         {
            if (value[index] < m_MinVal[index])
               currentValue[index] += m_MinVal[index];
            else if (value[index] > m_MaxVal[index])
               currentValue[index] += m_MaxVal[index];
            else currentValue[index] += value[index];
         }
         m_MeanValueMap[label] = currentValue;
      }
      ++inIt;
      ++labelIt;
    }
}


template<class TInputVectorImage, class TLabelImage>
void
PersistentRGBFromImageValueFilter<TInputVectorImage, TLabelImage>
::PrintSelf(std::ostream& os, itk::Indent indent) const
{
  Superclass::PrintSelf(os, indent);
}

} // end namespace otb
#endif
