/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbSimpleImageLayer_h
#define __otbSimpleImageLayer_h

#include "otbImageLayer.h"
#include "TypeDefinitions.h"

namespace otb
{
/** \class SimpleImageLayer
*   \brief This class is a layer container.
*   It contains everything related to a layer in the viewer model with
*   a simple pixel description
*
*   \sa ImageViewerModel
*   \sa ImageLayer
*   \sa Layer
*  \ingroup Visualization
 */

template <class TImage, class TOutputImage>
class SimpleImageLayer
  : public ImageLayer<TImage, TOutputImage>
{
public:
  /** Standard class typedefs */
  typedef SimpleImageLayer                            Self;
  typedef ImageLayer<TImage, TOutputImage>        Superclass;
  typedef itk::SmartPointer<Self>                     Pointer;
  typedef itk::SmartPointer<const Self>               ConstPointer;

  /** Method for creation through the object factory */
  itkNewMacro(Self);

  /** Runtime information */
  itkTypeMacro(SimpleImageLayer, SimpleImageLayerBase);

  typedef typename Superclass::IndexType           IndexType;
  typedef typename Superclass::PointType           PointType;

  typedef typename Superclass::RenderingFunctionType  RenderingFunctionType;
  typedef typename RenderingFunctionType::PixelType         PixelType;
  typedef typename RenderingFunctionType::InternalPixelType InternalPixelType;
  typedef typename RenderingFunctionType::OutputPixelType   OutputPixelType;
  //typedef typename RenderingFunctionType::OutputValueType   OutputValueType;

  /** Get the pixel description */
  virtual std::string GetPixelDescription(const IndexType& index, bool getPlaceName = true);

  /** set shown values label */
  itkSetStringMacro(PixelValueDescriptionLabel);
  
  /** flag to show/hide RGB value of a pixel */
  itkSetMacro(HideRGBPixelValue, bool);

  /** flag to show/hide RGB value of a pixel */
  itkSetMacro(IsGrayComposition, bool);
  itkSetMacro(GrayChannel, unsigned int);

  /** Set the colorMap Ramp (for kmeans)*/
  void SetColorMapRamp(const ColorMapRampType & colormap )
  {
    m_ColorMap = colormap;
  }

  /** set the spectral band order */
  void UpdateChannelsOrder(int blue, int green, int red, int pir, int mir)
  {
    // clear it first
    m_ChannelsOrder.clear();
    
    // add the channels
    m_ChannelsOrder.push_back(blue);
    m_ChannelsOrder.push_back(green);
    m_ChannelsOrder.push_back(red);
    m_ChannelsOrder.push_back(pir);
    m_ChannelsOrder.push_back(mir);
  }

protected:
  /** Constructor */
  SimpleImageLayer();
  /** Destructor */
  virtual ~SimpleImageLayer();

  /** Describe the pixel */ 
  const std::string Describe(const typename RenderingFunctionType::PixelType& spixel);

  /** Reorder following the order set in the channel order vector*/
  PixelType  OrderCurrentSpectrum(const PixelType& spixel);

private:
  SimpleImageLayer(const Self&);     // purposely not implemented
  void operator =(const Self&); // purposely not implemented

  std::string      m_PixelValueDescriptionLabel;
  bool             m_HideRGBPixelValue;
  ColorMapRampType m_ColorMap;  
  bool             m_IsGrayComposition;
  unsigned int     m_GrayChannel;

  /** Channels specifications */
  std::vector< int >   m_ChannelsOrder;
  std::vector< std::string >   m_ChannelsName;
  std::vector< std::string >   m_FinalChannelsName;

}; // end class
} // end namespace otb

#ifndef OTB_MANUAL_INSTANTIATION
#include "otbSimpleImageLayer.txx"
#endif

#endif
