/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbSimpleImageLayer_txx
#define __otbSimpleImageLayer_txx

#include "otbSimpleImageLayer.h"

namespace otb
{

template <class TImage, class TOutputImage>
SimpleImageLayer<TImage, TOutputImage>
::SimpleImageLayer()
{
  m_PixelValueDescriptionLabel = "Valeur pixel";
  m_HideRGBPixelValue = true;
  m_ColorMap.clear();
  m_IsGrayComposition = false;

  m_ChannelsName.push_back("BS_Bleue");
  m_ChannelsName.push_back("BS_Verte");
  m_ChannelsName.push_back("BS_Rouge");
  m_ChannelsName.push_back("BS_PIR");
  m_ChannelsName.push_back("BS_MIR");
}

template <class TImage, class TOutputImage>
SimpleImageLayer<TImage, TOutputImage>
::~SimpleImageLayer()
{}

template <class TImage, class TOutputImage>
std::string
SimpleImageLayer<TImage, TOutputImage>
::GetPixelDescription(const IndexType& index, bool getPlaceName)
{
  //FIXME only if necessary
  this->UpdateListSample();

  // The ouptut stringstream
  std::ostringstream oss;
  oss  << "Taille Image" << ": " 
      << this->GetImage()->GetLargestPossibleRegion().GetSize() << std::endl;

  //here, we consider that if the transform is not ready (call to InitTransform)
  //the user of the class don't want to use it
  if (this->GetTransform()->IsUpToDate())
    {
    if (this->GetTransform()->GetTransformAccuracy() != Projection::UNKNOWN)
      {
      // approximated spacing 
      double spacing = std::min(this->GetApproxGroundSpacing()[0], vcl_abs(this->GetApproxGroundSpacing()[1]));

      oss << setiosflags(std::ios::fixed) << std::setprecision(2) 
          << "Taille pixels(en m): (" 
          << spacing << ";" <<  spacing << ")" 
          << std::endl;
      }
    }
  
  // pixel coordinates
  oss << std::endl<<"Coordonnees pixel" << ": " << index << std::endl;
  
  //here, we consider that if the transform is not ready (call to InitTransform)
  //the user of the class don't want to use it
  if (this->GetTransform()->IsUpToDate())
    {
    if (this->GetTransform()->GetTransformAccuracy() != Projection::UNKNOWN)
      {
      PointType point = this->GetPixelLocation (index);

      oss << setiosflags(std::ios::fixed) << std::setprecision(6)
          << "Lon: " << point[0] << " Lat: " << point[1] 
          << std::endl;
      }
    }

  // If we are inside the buffered region
  if (this->GetImage()->GetBufferedRegion().IsInside(index))
    {
    oss << std::endl<<this->Describe(this->GetImage()->GetPixel(index));
    }
  else 
    if (this->GetQuicklook() != NULL)
  // Else we extrapolate the value from the quicklook
    {
    IndexType ssindex = index;
    ssindex[0] /= this->GetQuicklookSubsamplingRate();
    ssindex[1] /= this->GetQuicklookSubsamplingRate();

    if (this->GetQuicklook()->GetBufferedRegion().IsInside(ssindex))
      {
      oss<< this->Describe(this->GetQuicklook()->GetPixel(ssindex));
      }
    }

  return oss.str();
}

template <class TImage, class TOutputImage>
const std::string 
SimpleImageLayer<TImage, TOutputImage>
::Describe(const typename RenderingFunctionType::PixelType& spixel) 
{
  std::ostringstream oss;
  
  // Common to all cases
  InternalPixelType spixelRepresentation = this->GetRenderingFunction()->EvaluatePixelRepresentation(spixel);

  if (m_ColorMap.size() == 0)
    {
    if (m_IsGrayComposition && m_GrayChannel< spixel.Size())
      {
      oss << otbGetTextMacro(m_PixelValueDescriptionLabel) << ":   " <<spixel[m_GrayChannel] << std::endl;
      }
    else
      {

      // Order the pixel
      PixelType  orderedPixel = this->OrderCurrentSpectrum(spixel);

      // Get the pixel description
      oss << otbGetTextMacro(m_PixelValueDescriptionLabel) << ": "<<std::endl;;
      for (unsigned int idx = 0 ; idx < orderedPixel.Size(); idx++)
        {
        oss <<m_FinalChannelsName[idx]<<"     " <<orderedPixel[idx] << std::endl;
        }
      }
    }
  else
    {    
    // Get the label relative to this pixel
    ColorMapRampType::const_iterator it = m_ColorMap.begin();
    while(it != m_ColorMap.end())
      {
      // find the label relative to the pixel value in the colored image
      if (spixelRepresentation == (*it).second)
        oss << "Classe : "<<(*it).first;
      ++it;
      }
    }

  oss << std::endl;
  return oss.str();
}

template <class TImage, class TOutputImage>
typename SimpleImageLayer<TImage, TOutputImage>::PixelType 
SimpleImageLayer<TImage, TOutputImage>
::OrderCurrentSpectrum(const PixelType& spixel)
{
  // Do some checking
  unsigned int nbComponents = spixel.Size();

  // output value
  PixelType orderedSpectrum;
  orderedSpectrum.SetSize(nbComponents);
  
  // How many spectral bands are set
  unsigned int  availableBands = 0;
  for (unsigned int idx = 0; idx < m_ChannelsOrder.size(); idx++)
    {
    if (m_ChannelsOrder[idx] != -1)
      {
      availableBands++;
      }
    }

  // if available bands are equal to spectrum size : reorder, else do
  // nothing 
  if (availableBands != nbComponents)
    {
    orderedSpectrum = spixel;
    m_FinalChannelsName.resize(spixel.Size());
    }
  else
    {
    unsigned int spectrumIndex = 0;
    for (unsigned int cid = 0; cid < m_ChannelsOrder.size(); cid++)
      {
      if (m_ChannelsOrder[cid] != -1)
        {
        orderedSpectrum.SetElement(spectrumIndex, spixel.GetElement( m_ChannelsOrder[cid] ) );
        // reorder the vector containing the spectral bands name
        if (cid < m_ChannelsName.size())
            m_FinalChannelsName.push_back(m_ChannelsName[cid]);
        spectrumIndex++;
        }
      }
    }

  return orderedSpectrum;
}

}

#endif
